﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using System.IO;
using System.Linq;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaLabs.Storage.DataServicesUtils;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests
{
    [TestFixture]
    public class BackupTableGraphBuilderTests
    {
        #region tests data

        private const string BackupTestContainerName = "integrationbackuptests";
        public const string LocalBackupFolder = "local-backup";

        private Guid _customer1Id = Guid.Parse("011dc137-9406-4242-8fad-3f9b667fa4c2");
        private Guid _customer2Id = Guid.Parse("a719ce58-20e7-49f2-a938-b86a0e8af673");

        private Guid _categoryElectronicsId = Guid.Parse("137619d7-1cff-4130-a62a-d69c52e1fd26");
        private Guid _categoryFoodId = Guid.Parse("159ea8bb-6700-4d6e-817f-2681cdbf9ee3");
        private Guid _categorySportId = Guid.Parse("945aed92-9e87-4937-a1cb-41d2e0690e6b");


        [Table(TableName)]
        public class Customer : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsCustomers";

            public Customer()
            {
            }

            public Customer(string customerId)
                : base(customerId, string.Empty)
            {
            }

            public string Name { get; set; }

            public string Address { get; set; }
        }

        [Table(TableName)]
        public class CustomerUser : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsCustomerUsers";

            public CustomerUser()
            {
            }

            public CustomerUser(string customerId, string userId)
                : base(customerId, userId)
            {
            }

            public string Name { get; set; }
        }

        [Table(TableName)]
        public class ProductCategory : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProductCategories";

            public ProductCategory()
            {
            }

            public ProductCategory(string customerId, string categoryId)
                : base(customerId, categoryId)
            {
            }

            public string Name { get; set; }
        }

        [Table(TableName)]
        public class Product : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProducts";

            public Product()
            {
            }

            public Product(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public string CustomerId { get; set; }

            public string Name { get; set; }

            public double PricePerUnit { get; set; }

            public int Quantity { get; set; }
        }

        [Table(TableName)]
        public class ProductCategoryColour : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProductCategoruColours";

            public ProductCategoryColour()
            {
            }

            public ProductCategoryColour(string categoryId, string rowId)
                : base(categoryId, rowId)
            {
            }

            public string CustomerId { get; set; }

            public string Colour { get; set; }
        }

        public interface IContext : ISimpleContextBase
        {
            IDbSet<Customer> Customers { get; set; }
            IDbSet<CustomerUser> CustomerUsers { get; set; }
            IDbSet<ProductCategory> ProductCategories { get; set; }
            IDbSet<ProductCategoryColour> ProductCategoryColours { get; set; }
            IDbSet<Product> Products { get; set; }
        }

        public Customer[] OriginalCustomers { get; set; }
        public CustomerUser[] OriginalCustomerUsers { get; set; }
        public ProductCategory[] OriginalProductCategories { get; set; }
        public Product[] OriginalProducts { get; set; }
        public ProductCategoryColour[] OriginalProductCategoryColours { get; set; }
        public IContext Context { get; set; }

        #endregion test data

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            if (Directory.Exists(LocalBackupFolder))
                Directory.Delete(LocalBackupFolder, true);
        }

        [TearDown]
        public void TearDown()
        {
            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            new BlobStorageProvider().DeleteContainer(BackupTestContainerName);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;

            if (Directory.Exists(LocalBackupFolder))
                Directory.Delete(LocalBackupFolder, true);
        }

        [Test]
        public void BuildAndBackupTest()
        {
            Arrange();

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            var graph = BackupTableGraphBuilder.CreateRoot(blobNameProvider, Customer.TableName)
                            .ChildByPartitionKey(CustomerUser.TableName)
                            .SiblingByPartitionKey(ProductCategory.TableName)
                                .ChildByRowKey(Product.TableName)
                            .BackupRoot;

            graph.Backup(ranges);

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            var blobs = new BlobStorageProvider().FlatBlobList(BackupTestContainerName).ToList();
            Assert.AreEqual(9, blobs.Count());
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestscustomers/a0.to.a8.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestscustomers/null.to.10.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestscustomerusers/011dc137-9406-4242-8fad-3f9b667fa4c2.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestscustomerusers/a719ce58-20e7-49f2-a938-b86a0e8af673.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproductcategories/011dc137-9406-4242-8fad-3f9b667fa4c2.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproductcategories/a719ce58-20e7-49f2-a938-b86a0e8af673.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/137619d7-1cff-4130-a62a-d69c52e1fd26.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/159ea8bb-6700-4d6e-817f-2681cdbf9ee3.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/945aed92-9e87-4937-a1cb-41d2e0690e6b.1");

            new TableRestore(blobNameProvider, Customer.TableName).Restore();
            new TableRestore(blobNameProvider, CustomerUser.TableName).Restore();
            new TableRestore(blobNameProvider, ProductCategory.TableName).Restore();
            new TableRestore(blobNameProvider, Product.TableName).Restore();

            AssertRestore();
        }

        [Test]
        public void BuildAndBackupToLocalFileSystemTest()
        {
            Arrange();

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            var graph = BackupTableGraphBuilder.CreateRoot(blobNameProvider, Customer.TableName, null, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                            .ChildByPartitionKey(CustomerUser.TableName, 1, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                            .SiblingByPartitionKey(ProductCategory.TableName, 1, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                                .ChildByRowKey(Product.TableName, 1, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                            .BackupRoot;

            graph.Backup(ranges);

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            new TableRestore(blobNameProvider, Customer.TableName, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                .Restore();

            new TableRestore(blobNameProvider, CustomerUser.TableName, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                .Restore();

            new TableRestore(blobNameProvider, ProductCategory.TableName, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                .Restore();

            new TableRestore(blobNameProvider, Product.TableName, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                .Restore();

            AssertRestore();
        }

        [Test]
        public void BuildAndBackupTestForOneTableLimitedByPartitionKey()
        {
            Arrange();

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyExactMatchRange.FromPartitionKeyValue(OriginalProductCategories[0].RowKey);

            var graph = BackupTableGraphBuilder.CreateRoot(blobNameProvider, Product.TableName).BackupRoot;

            graph.Backup(ranges);

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            var blobs = new BlobStorageProvider().FlatBlobList(BackupTestContainerName).ToList();
            Assert.AreEqual(1, blobs.Count());
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/137619d7-1cff-4130-a62a-d69c52e1fd26.1");

            new TableRestore(blobNameProvider, Product.TableName).Restore();

            AssertElectrinicsProducts();
        }

        [Test]
        public void BuildAndBackupTestForSiblingsByRowKey()
        {
            Arrange();

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            var graph = BackupTableGraphBuilder.CreateRoot(blobNameProvider, ProductCategory.TableName)
                            .ChildByRowKey(ProductCategoryColour.TableName)
                            .SiblingByRowKey(Product.TableName)
                            .BackupRoot;

            graph.Backup(ranges);

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            var blobs = new BlobStorageProvider().FlatBlobList(BackupTestContainerName).ToList();
            Assert.AreEqual(6, blobs.Count());
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproductcategories/a0.to.a8.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproductcategories/null.to.10.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproductcategorucolours/137619d7-1cff-4130-a62a-d69c52e1fd26.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/137619d7-1cff-4130-a62a-d69c52e1fd26.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/159ea8bb-6700-4d6e-817f-2681cdbf9ee3.1");
            CollectionAssert.Contains(blobs, "integrationbackuptests/integrationbackuptestsproducts/945aed92-9e87-4937-a1cb-41d2e0690e6b.1");

            new TableRestore(blobNameProvider, ProductCategoryColour.TableName).Restore();
            new TableRestore(blobNameProvider, ProductCategory.TableName).Restore();
            new TableRestore(blobNameProvider, Product.TableName).Restore();

            AsserFirstCustomerProductCategories();
            AssertElectrinicsProducts();
            AssertFoodProducts();
            AssertElenctronicsProductCategoryColours();
        }

        private void Arrange()
        {
            CreateOriginalData();

            new BlobStorageProvider().DeleteContainer(BackupTestContainerName);

            Context = SimpleContextBase.CreateInstance<IContext>();

            Context.DeleteTables();

            Context.CreateTablesIfNotExist();

            PopulateTables();
        }

        private void CreateOriginalData()
        {
            OriginalCustomers = new[]
            {
                new Customer(_customer1Id.ToString()) {Name = "Golden Standard", Address = "Somewhere in Ireland"},
                new Customer(_customer2Id.ToString()) {Name = "Silver Standard", Address = "Somewhere in UK"}
            };

            OriginalCustomerUsers = new[]
            {
                new CustomerUser(OriginalCustomers[0].PartitionKey, Guid.NewGuid().ToString()) {Name = "John Smith"}, 
                new CustomerUser(OriginalCustomers[1].PartitionKey, Guid.NewGuid().ToString()) {Name = "Peter Black"}
            };

            OriginalProductCategories = new[]
            {
                new ProductCategory(OriginalCustomers[0].PartitionKey, _categoryElectronicsId.ToString())
                {
                    Name = "Electronics"
                }, 
                new ProductCategory(OriginalCustomers[0].PartitionKey, _categoryFoodId.ToString())
                {
                    Name = "Food"
                },
                new ProductCategory(OriginalCustomers[1].PartitionKey, _categorySportId.ToString())
                {
                    Name = "Sport"
                } 
            };

            OriginalProductCategoryColours = new[]
            {
                new ProductCategoryColour(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[0].PartitionKey,
                    Colour = "Red"
                }, 
                new ProductCategoryColour(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[0].PartitionKey,
                    Colour = "Green"
                }
            };

            OriginalProducts = new[]
            {
                new Product(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[0].PartitionKey,
                    Name = "Fake iPod",
                    PricePerUnit = 89.99,
                    Quantity = 10
                }, 
                new Product(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[0].PartitionKey,
                    Name = "Fake EOS 1Ds Mark I",
                    PricePerUnit = 5589.99,
                    Quantity = 1
                }, 
                new Product(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[0].PartitionKey,
                    Name = "Fake HTC Desire",
                    PricePerUnit = 289.99,
                    Quantity = 5
                }, 
                new Product(OriginalProductCategories[1].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[0].PartitionKey,
                    Name = "Ice cream",
                    PricePerUnit = 1.22,
                    Quantity = 50
                }, 
                new Product(OriginalProductCategories[2].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[1].PartitionKey,
                    Name = "Tennis Racket",
                    PricePerUnit = 289.99,
                    Quantity = 2
                }, 
                new Product(OriginalProductCategories[2].RowKey, Guid.NewGuid().ToString())
                {
                    CustomerId = OriginalCustomers[1].PartitionKey,
                    Name = "Tennis Shoes",
                    PricePerUnit = 59.99,
                    Quantity = 7
                }
            };
        }

        private void PopulateTables()
        {
            foreach (var customer in OriginalCustomers)
            {
                Context.Customers.Add(customer);
            }

            foreach (var user in OriginalCustomerUsers)
            {
                Context.CustomerUsers.Add(user);
            }

            foreach (var category in OriginalProductCategories)
            {
                Context.ProductCategories.Add(category);
            }

            foreach (var colour in OriginalProductCategoryColours)
            {
                Context.ProductCategoryColours.Add(colour);
            }

            foreach (var product in OriginalProducts)
            {
                Context.Products.Add(product);
            }

            Context.SaveChanges();
        }

        private void AssertRestore()
        {
            // normally OriginalProductCategoryColours are not backup

            AssertCustomers();

            AsserCustomerUsers();

            AsserProductCategories();

            AssertProducts();
        }

        private void AssertCustomers()
        {
            var customers = SimpleContextBase.CreateInstance<IContext>().Customers.ToList();

            Assert.IsNotNull(customers);
            Assert.AreEqual(2, customers.Count);
            AssertCustomer(OriginalCustomers[0], customers);
            AssertCustomer(OriginalCustomers[1], customers);
        }

        private void AsserCustomerUsers()
        {
            var customer1Users =
                SimpleContextBase.CreateInstance<IContext>().CustomerUsers.Where(
                    x => x.PartitionKey == OriginalCustomers[0].PartitionKey).ToList();
            Assert.IsNotNull(customer1Users);
            Assert.AreEqual(1, customer1Users.Count);
            AssertCustomerUser(OriginalCustomerUsers[0], customer1Users);

            var customer2Users =
                SimpleContextBase.CreateInstance<IContext>().CustomerUsers.Where(
                    x => x.PartitionKey == OriginalCustomers[1].PartitionKey).ToList();
            Assert.IsNotNull(customer2Users);
            Assert.AreEqual(1, customer2Users.Count);
            AssertCustomerUser(OriginalCustomerUsers[1], customer2Users);
        }

        private void AsserProductCategories()
        {
            AsserFirstCustomerProductCategories();

            var customer2Categories =
                SimpleContextBase.CreateInstance<IContext>().ProductCategories.Where(
                    x => x.PartitionKey == OriginalCustomers[1].PartitionKey).ToList();

            Assert.IsNotNull(customer2Categories);
            Assert.AreEqual(1, customer2Categories.Count);
            AssertProductCategory(OriginalProductCategories[2], customer2Categories);
        }

        private void AsserFirstCustomerProductCategories()
        {
            var customer1Categories =
                SimpleContextBase.CreateInstance<IContext>().ProductCategories.Where(
                    x => x.PartitionKey == OriginalCustomers[0].PartitionKey).ToList();
            
            Assert.IsNotNull(customer1Categories);
            Assert.AreEqual(2, customer1Categories.Count);
            AssertProductCategory(OriginalProductCategories[0], customer1Categories);
            AssertProductCategory(OriginalProductCategories[1], customer1Categories);
        }

        private void AssertProducts()
        {
            AssertElectrinicsProducts();

            AssertFoodProducts();

            var sportProducts =
                SimpleContextBase.CreateInstance<IContext>().Products.Where(
                    x => x.PartitionKey == _categorySportId.ToString()).ToList();

            Assert.IsNotNull(sportProducts);
            Assert.AreEqual(2, sportProducts.Count);
            AssertProduct(OriginalProducts[4], sportProducts);
            AssertProduct(OriginalProducts[5], sportProducts);
        }

        private void AssertElectrinicsProducts()
        {
            var electronicsProducts =
                SimpleContextBase.CreateInstance<IContext>().Products.Where(
                    x => x.PartitionKey == _categoryElectronicsId.ToString()).ToList();

            Assert.IsNotNull(electronicsProducts);
            Assert.AreEqual(3, electronicsProducts.Count);
            AssertProduct(OriginalProducts[0], electronicsProducts);
            AssertProduct(OriginalProducts[1], electronicsProducts);
            AssertProduct(OriginalProducts[2], electronicsProducts);
        }

        private void AssertFoodProducts()
        {
            var foodProducts =
                SimpleContextBase.CreateInstance<IContext>().Products.Where(
                    x => x.PartitionKey == _categoryFoodId.ToString()).ToList();

            Assert.IsNotNull(foodProducts);
            Assert.AreEqual(1, foodProducts.Count);
            AssertProduct(OriginalProducts[3], foodProducts);
        }

        private void AssertElenctronicsProductCategoryColours()
        {
            var colours =
                SimpleContextBase.CreateInstance<IContext>().ProductCategoryColours.Where(
                    x => x.PartitionKey == _categoryElectronicsId.ToString()).ToList();

            Assert.IsNotNull(colours);
            Assert.AreEqual(2, colours.Count);
            AssertProductCategoryColour(OriginalProductCategoryColours[0], colours);
            AssertProductCategoryColour(OriginalProductCategoryColours[1], colours);
        }

        private static void AssertCustomer(Customer expected, IEnumerable<Customer> actualCollection)
        {
            var actual =
                actualCollection.Where(x => x.PartitionKey == expected.PartitionKey && x.RowKey == expected.RowKey).
                    FirstOrDefault();

            Assert.IsNotNull(actual);

            Assert.AreEqual(expected.Name, actual.Name);
            Assert.AreEqual(expected.Address, actual.Address);
        }

        private static void AssertCustomerUser(CustomerUser expected, IEnumerable<CustomerUser> actualCollection)
        {
            var actual =
                actualCollection.Where(x => x.PartitionKey == expected.PartitionKey && x.RowKey == expected.RowKey).
                    FirstOrDefault();

            Assert.IsNotNull(actual);

            Assert.AreEqual(expected.Name, actual.Name);
        }

        private static void AssertProductCategory(ProductCategory expected, IEnumerable<ProductCategory> actualCollection)
        {
            var actual =
                actualCollection.Where(x => x.PartitionKey == expected.PartitionKey && x.RowKey == expected.RowKey).
                    FirstOrDefault();

            Assert.IsNotNull(actual);

            Assert.AreEqual(expected.Name, actual.Name);
        }

        private static void AssertProduct(Product expected, IEnumerable<Product> actualCollection)
        {
            var actual =
                actualCollection.Where(x => x.PartitionKey == expected.PartitionKey && x.RowKey == expected.RowKey).
                    FirstOrDefault();

            Assert.IsNotNull(actual);

            Assert.AreEqual(expected.CustomerId, actual.CustomerId);
            Assert.AreEqual(expected.Name, actual.Name);
            Assert.AreEqual(expected.PricePerUnit, actual.PricePerUnit, double.Epsilon);
            Assert.AreEqual(expected.Quantity, actual.Quantity);
        }

        private static void AssertProductCategoryColour(ProductCategoryColour expected, IEnumerable<ProductCategoryColour> actualCollection)
        {
            var actual =
                actualCollection.Where(x => x.PartitionKey == expected.PartitionKey && x.RowKey == expected.RowKey).
                    FirstOrDefault();

            Assert.IsNotNull(actual);

            Assert.AreEqual(expected.Colour, actual.Colour);
        }
    }
}
