﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Specialized;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using DocaAzure.Storage;
using DocaAzure.Utils;
using DocaLabs.Utils;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests
{
    [TestFixture]
    public class BlobBackupTests
    {
        public const string SampleData = "A user object receives a pointer to another object which implements an algorithm. The first is designated 'element class' and the latter 'the visitor class'.";
        public const string LocalBackupFolder = "folder-for-local-backup-tests";
        public const string BackupToContainer = "tst-bck";

        public class BlobContainerBackupDifferentAccountTestTarget : BlobContainerBackup
        {
            protected override bool IsTheSameAccount()
            {
                return false;
            }   
        }

        public class BlobContainerRestoreDifferentAccountTestTarget : BlobContainerRestore
        {
            protected override bool IsTheSameAccount()
            {
                return false;
            }
        }

        private DateTime _timeMark;
        private string _longBlobName;
        private byte[] _pageContent512;
        private byte[] _pageContent1024;


        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            var provider = new BlobStorageProvider();
            provider.DeleteContainer("backup-source-test-container");
            provider.DeleteContainer(BackupToContainer);

            if (Directory.Exists(LocalBackupFolder))
                Directory.Delete(LocalBackupFolder, true);
        }

        [TearDown]
        public void TearDown()
        {
            var provider = new BlobStorageProvider();
            provider.DeleteContainer("backup-source-test-container");
            provider.DeleteContainer(BackupToContainer);

            if (Directory.Exists(LocalBackupFolder))
                Directory.Delete(LocalBackupFolder, true);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void BackupsAndRestoresInTheSameAccount()
        {
            ArrangeData();

            new BlobContainerBackup().Backup(new [] { "backup-source-test-container" }, BackupToContainer, null, 2);

            new BlobStorageProvider().DeleteContainer("backup-source-test-container");

            new BlobContainerRestore().Restore("backup-source-test-container", BackupToContainer, 2);

            AssertRestore();
        }

        [Test]
        public void BackupsAndRestoresInDifferentAccounts()
        {
            ArrangeData();

            new BlobContainerBackupDifferentAccountTestTarget().Backup(new [] { "backup-source-test-container" }, BackupToContainer, null, 2);

            new BlobStorageProvider().DeleteContainer("backup-source-test-container");

            new BlobContainerRestoreDifferentAccountTestTarget().Restore("backup-source-test-container", BackupToContainer, 2);

            AssertRestore();
        }

        [Test]
        public void BackupsAndRestoresOnlyModifiedSince()
        {
            ArrangeData();

            new BlobContainerBackup().Backup(new [] { "backup-source-test-container" }, BackupToContainer, _timeMark, 2);

            new BlobStorageProvider().DeleteContainer("backup-source-test-container");

            new BlobContainerRestore().Restore("backup-source-test-container", BackupToContainer, 2);

            var provider = new BlobStorageProvider();

            var blobList = provider.FlatBlobList("backup-source-test-container").ToList();

            Assert.AreEqual(3, blobList.Count);

            CollectionAssert.Contains(blobList, "backup-source-test-container/blob3");
            CollectionAssert.Contains(blobList, "backup-source-test-container/sub1/blob4");
            CollectionAssert.Contains(blobList, "backup-source-test-container/sub1/sub2/blob5");

            AssertBlobContent("backup-source-test-container/blob3");
            AssertBlobContent("backup-source-test-container/sub1/blob4");
            AssertBlobContent("backup-source-test-container/sub1/sub2/blob5");
        }

        [Test]
        public void BackupsAndRestoresUsingLocalFileSystem()
        {
            ArrangeData();

            new BlobContainerBackupToLocalFileSystem().Backup("backup-source-test-container", LocalBackupFolder, null/*, 2*/);

            new BlobStorageProvider().DeleteContainer("backup-source-test-container");

            BlobContainerRestoreFromLocalFileSystem.Restore("backup-source-test-container", LocalBackupFolder/*, 2*/);

            AssertRestore();
        }

        private void ArrangeData()
        {
            new BlobStorageProvider().CreateContainer("backup-source-test-container");

            var provider = (IBlobOperations) new BlobStorageProvider();

            var random = new Random();

            // create page blob
            _pageContent512 = new byte[512];
            _pageContent1024 = new byte[1024];

            random.NextBytes(_pageContent512);
            random.NextBytes(_pageContent1024);

            var pageBlob = AzureStorageAccountFactory.GetPageBlobReference("backup-source-test-container/test-page-blob");
            pageBlob.Create(6 * 512);

            using (var stream512 = new MemoryStream(_pageContent512))
            using (var stream1024 = new MemoryStream(_pageContent1024))
            {
                pageBlob.WritePages(stream512, 512);
                pageBlob.WritePages(stream1024, 1536);
            }

            using (var sourceStream = SampleData.StringToStream())
            {

                // create blob with a long name, the maximum length for blob name in devfabric is 256, 
                // to use 225 should be enough for our purpose (the longer would fail on CopyBlob operation)
                var longBlobNameBuilder = new StringBuilder("backup-source-test-container/");

                for (var i = 0; i < 225; ++i)
                    longBlobNameBuilder.Append((char)random.Next('a', 'z'));

                _longBlobName = longBlobNameBuilder.ToString();

                sourceStream.ResetToBegining();
                provider.PutBlob(_longBlobName, sourceStream, null, new NameValueCollection { { "Key1", "Value1" } });

                // create a name that would be invalid for NTFS but still valid blob name
                sourceStream.ResetToBegining();
                provider.PutBlob("backup-source-test-container/blo|b-with-bad-file-name", sourceStream, null, new NameValueCollection { { "Key2", "Value2" } });

                // other blobs
                sourceStream.ResetToBegining();
                provider.PutBlob("backup-source-test-container/blob1", sourceStream, "text/plain");

                sourceStream.ResetToBegining();
                provider.PutBlob("backup-source-test-container/blob2", sourceStream);

                _timeMark = DefaultClock.CurrentTime;
                Thread.Sleep(1000);

                sourceStream.ResetToBegining();
                provider.PutBlob("backup-source-test-container/blob3", sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob("backup-source-test-container/sub1/blob4", sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob("backup-source-test-container/sub1/sub2/blob5", sourceStream);
            }
        }

        private void AssertRestore()
        {
            var provider = new BlobStorageProvider();

            var blobList = provider.FlatBlobList("backup-source-test-container").ToList();

            Assert.AreEqual(8, blobList.Count);

            CollectionAssert.Contains(blobList, "backup-source-test-container/blob1");
            CollectionAssert.Contains(blobList, "backup-source-test-container/blob2");
            CollectionAssert.Contains(blobList, "backup-source-test-container/blob3");
            CollectionAssert.Contains(blobList, "backup-source-test-container/sub1/blob4");
            CollectionAssert.Contains(blobList, "backup-source-test-container/sub1/sub2/blob5");
            CollectionAssert.Contains(blobList, _longBlobName);
            CollectionAssert.Contains(blobList, "backup-source-test-container/blo|b-with-bad-file-name");
            CollectionAssert.Contains(blobList, "backup-source-test-container/test-page-blob");

            AssertBlobContent("backup-source-test-container/blob1", "text/plain");
            AssertBlobContent("backup-source-test-container/blob2");
            AssertBlobContent("backup-source-test-container/blob3");
            AssertBlobContent("backup-source-test-container/sub1/blob4");
            AssertBlobContent("backup-source-test-container/sub1/sub2/blob5");

            AssertBlobContent(_longBlobName, null, new NameValueCollection { { "Key1", "Value1" } });
            AssertBlobContent("backup-source-test-container/blo|b-with-bad-file-name");

            // assert page blob
            var pageBlob = AzureStorageAccountFactory.Create().CreateCloudBlobClient().GetPageBlobReference("backup-source-test-container/test-page-blob");

            var pageRanges = pageBlob.GetPageRanges().ToList();
            Assert.AreEqual(2, pageRanges.Count);
            Assert.AreEqual(512, pageRanges[0].StartOffset);
            Assert.AreEqual(1023, pageRanges[0].EndOffset);
            Assert.AreEqual(1536, pageRanges[1].StartOffset);
            Assert.AreEqual(2559, pageRanges[1].EndOffset);

            using (var stream = pageBlob.OpenRead())
            {
                var reslut512 = new byte[512];
                var reslut1024 = new byte[1024];

                stream.Seek(512, SeekOrigin.Begin);
                stream.Read(reslut512, 0, 512);

                stream.Seek(1536, SeekOrigin.Begin);
                stream.Read(reslut1024, 0, 1024);

                CollectionAssert.AreEqual(_pageContent512, reslut512);
                CollectionAssert.AreEqual(_pageContent1024, reslut1024);
            }
        }

        private static void AssertBlobContent(string blobName, string contentType = null, NameValueCollection metadata = null)
        {
            using (var stream = ((IBlobOperations)new BlobStorageProvider()).GetBlob(blobName))
            {
                Assert.AreEqual(SampleData, stream.StreamToString());

                if (contentType != null)
                {
                    Assert.AreEqual(contentType, stream.GetBlobProperties().ContentType);
                }

                if (metadata == null)
                    return;

                var actualMetadata = stream.GetBlobMetadata();

                foreach (var key in metadata.AllKeys)
                {
                    Assert.AreEqual(metadata[key], actualMetadata[key]);
                }
            }
        }
    }
}
