﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Linq;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaAzure.Utils;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests.Impl
{
    [TestFixture]
    public class AzureTableBackupStorageFactoryTests
    {
        public const string BackupTestContainerName = "integrationbackuptests";
        public const string BlobName = "foo/table1/blob-data";
        public const string BlobName10 = "table1/blob-data";
        public const string BlobName20 = "table2/blob-data";
        public const string TablePrefix = "foo/table1/";

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void TearDown()
        {
            new BlobStorageProvider().DeleteContainer(BackupTestContainerName);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void WriteReadRoundtrip()
        {
            var target = new AzureBackupStorageFactory();

            // writing
            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName))
            {
                for (var i = 0; i < 11; i++)
                {
                    using (var stream = ("Hello World!" + i).StringToStream())
                    {
                        writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                    }
                }

                writer.FinalizeBlob();
            }

            // reading
            using (var reader = target.GetTableRestoreReader(BackupTestContainerName + "/" + BlobName))
            {
                var count = 0;

                foreach (var block in reader.GetBlocks())
                {
                    using (block)
                    {
                        Assert.AreEqual("Hello World!" + count, block.StreamToString());
                        count++;
                    }
                }

                Assert.AreEqual(11, count);
            }
        }

        [Test]
        public void GetTableBlobsReturnsEmptyList()
        {
            Assert.AreEqual(0, new AzureBackupStorageFactory().GetTableBlobs(BackupTestContainerName, TablePrefix).Count());
        }

        [Test]
        public void GetTableBlobsReturnsBlobsNames()
        {
            var target = new AzureBackupStorageFactory();

            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName))
            {
                using (var stream = "Hello World!".StringToStream())
                {
                    writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                }

                writer.FinalizeBlob();
            }

            var result = target.GetTableBlobs(BackupTestContainerName, TablePrefix).ToArray();
            Assert.AreEqual(1, result.Length);
            Assert.AreEqual("integrationbackuptests/foo/table1/blob-data", result[0]);

            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName + "-42"))
            {
                using (var stream = "Hello World!".StringToStream())
                {
                    writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                }

                writer.FinalizeBlob();
            }

            result = target.GetTableBlobs(BackupTestContainerName, TablePrefix).ToArray();
            Array.Sort(result);
            Assert.AreEqual(2, result.Length);
            Assert.AreEqual("integrationbackuptests/foo/table1/blob-data", result[0]);
            Assert.AreEqual("integrationbackuptests/foo/table1/blob-data-42", result[1]);
        }

        [Test]
        public void GetTablesReturnsTableNames()
        {
            var target = new AzureBackupStorageFactory();

            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName10))
            {
                using (var stream = "Hello World!".StringToStream())
                {
                    writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                }

                writer.FinalizeBlob();
            }

            var result = target.GetTables(BackupTestContainerName).ToArray();
            Assert.AreEqual(1, result.Length);
            Assert.AreEqual("table1", result[0]);

            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName20))
            {
                using (var stream = "Hello World!".StringToStream())
                {
                    writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                }

                writer.FinalizeBlob();
            }

            result = target.GetTables(BackupTestContainerName).ToArray();

            Array.Sort(result);
            Assert.AreEqual(2, result.Length);
            Assert.AreEqual("table1", result[0]);
            Assert.AreEqual("table2", result[1]);
        }

        [Test]
        public void GetOriginalContainersReturnsOriginalContainerNames()
        {
            var target = new AzureBackupStorageFactory();

            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName10))
            {
                using (var stream = "Hello World!".StringToStream())
                {
                    writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                }

                writer.FinalizeBlob();
            }

            var result = target.GetOriginalContainers(BackupTestContainerName).ToArray();

            Assert.AreEqual(1, result.Length);
            Assert.AreEqual("table1", result[0]);

            using (var writer = target.GetTableBackupWriter(BackupTestContainerName + "/" + BlobName20))
            {
                using (var stream = "Hello World!".StringToStream())
                {
                    writer.AddBlock(Convert.ToBase64String(Guid.NewGuid().ToByteArray()), stream);
                }

                writer.FinalizeBlob();
            }

            result = target.GetOriginalContainers(BackupTestContainerName).ToArray();

            Array.Sort(result);
            Assert.AreEqual(2, result.Length);
            Assert.AreEqual("table1", result[0]);
            Assert.AreEqual("table2", result[1]);
        }
    }
}
