﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Threading;
using DocaAzure.Backup.Interfaces;
using DocaAzure.Backup.Service;
using DocaAzure.Backup.Service.Blobs;
using DocaAzure.Messaging;
using DocaAzure.Storage;
using DocaAzure.TestUtils;
using DocaAzure.Utils;
using DocaLabs.Utils;
using Moq;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests.Service
{
    [TestFixture]
    public class BlobTriggerServiceTest
    {
        public const string BackupTestContainerName = "blobbackuptests";
        public const string InitiatedTime = "2011-02-12T00:00:05.345Z";
        public const string SampleData = "A user object receives a pointer to another object which implements an algorithm. The first is designated 'element class' and the latter 'the visitor class'.";

        public class RestoreDefaultProgressMonitorTestTarget : RestoreDefaultProgressMonitor
        {
            public static bool RestoreSuccessful { get; set; }

            protected override MonitoringResults OnSuccess(string restoreId, MonitoringResults result)
            {
                RestoreSuccessful = true;
                return base.OnSuccess(restoreId, result);
            }
        }

        private DateTime _timeMark;

        private static TimeSpan OriginalBackupTimeout { get; set; }
        private static TimeSpan OriginalPollingTimeout { get; set; }
        private static TimeSpan OriginalRestoreTimeout { get; set; }
        private static TimeSpan OriginalRestorePollingTimeout { get; set; }
        private static Type OriginalRestoreProgressMonitorType { get; set; }

        private static IClock _originalClock;
        private static FixedClock _fixedClock;

        [SetUp]
        public void Setup()
        {
            _originalClock = DefaultClock.Current;
            DefaultClock.Current = _fixedClock = new FixedClock();

            OriginalBackupTimeout = BackupDefaultProgressMonitor.BackupTimeout;
            OriginalPollingTimeout = BackupDefaultProgressMonitor.PollingTimeout;

            OriginalRestoreTimeout = RestoreDefaultProgressMonitor.RestoreTimeout;
            OriginalRestorePollingTimeout = RestoreDefaultProgressMonitor.PollingTimeout;

            BackupDefaultProgressMonitor.BackupTimeout = TimeSpan.FromMinutes(30);
            BackupDefaultProgressMonitor.PollingTimeout = TimeSpan.FromMilliseconds(100);

            RestoreDefaultProgressMonitor.RestoreTimeout = TimeSpan.FromMinutes(30);
            RestoreDefaultProgressMonitor.PollingTimeout = TimeSpan.FromMilliseconds(100);

            OriginalRestoreProgressMonitorType = DefaultBackupServiceFactory.RestoreProgressMonitorType;
            DefaultBackupServiceFactory.RestoreProgressMonitorType = typeof(RestoreDefaultProgressMonitorTestTarget);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            CleanTestData();
        }

        [TearDown]
        public void TearDown()
        {
            DefaultClock.Current = _originalClock;

            CleanTestData();

            DefaultConfigurationManager.Default = null;

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;

            BackupDefaultProgressMonitor.BackupTimeout = OriginalBackupTimeout;
            BackupDefaultProgressMonitor.PollingTimeout = OriginalPollingTimeout;

            RestoreDefaultProgressMonitor.RestoreTimeout = OriginalRestoreTimeout;
            RestoreDefaultProgressMonitor.PollingTimeout = OriginalRestorePollingTimeout;

            DefaultBackupServiceFactory.RestoreProgressMonitorType = OriginalRestoreProgressMonitorType;
        }

        [Test]
        public void BackupsAndRestores()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new BlobContainerBackupTriggerMessage
            {
                BackupConfiguration = "blobAdHocBackup",
                OnlyIfModifiedSince = _timeMark
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new BlobContainerBackupTriggerMessageHandler(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages);

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new BlobContainerBackupMessageHandler().Handle((BlobContainerBackupMessage)backupMessage);
            }

            var provider = new BlobStorageProvider();
            provider.DeleteContainer("tests-b1-container");
            provider.DeleteContainer("tests-b1-test-container");
            provider.DeleteContainer("tests-b2-container");

            AssertBackup();

            // trigger the restore
            var restoreProgressMonitor = BlobContainerRestoreTrigger.Trigger(
                "blobbackuptests", DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), 4, mockBus.Object,
                "tests-b1-container", "tests-b1-test-container", "tests-b1-test42-container");

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new BlobContainerRestoreMessageHandler().Handle((BlobContainerRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestore();

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true);

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void BackupTypeInTriggerMessageOverridesBackupTypeInConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new BlobContainerBackupTriggerMessage
            {
                BackupConfiguration = "blobAdHocBackup",
                OnlyIfModifiedSince = _timeMark,
                BackupType = BackupType.Daily
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new BlobContainerBackupTriggerMessageHandler(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages, "Blob 2011-02-12 Daily");

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new BlobContainerBackupMessageHandler().Handle((BlobContainerBackupMessage)backupMessage);
            }

            var provider = new BlobStorageProvider();
            provider.DeleteContainer("tests-b1-container");
            provider.DeleteContainer("tests-b1-test-container");
            provider.DeleteContainer("tests-b2-container");

            AssertBackup("Blob 2011-02-12 Daily");

            // trigger the restore
            var restoreProgressMonitor = BlobContainerRestoreTrigger.Trigger(
                "blobbackuptests", DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), 4, mockBus.Object,
                "tests-b1-container", "tests-b1-test-container", "tests-b1-test42-container");

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new BlobContainerRestoreMessageHandler().Handle((BlobContainerRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestore();

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true, "Blob 2011-02-12 Daily");

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void BackupsAndRestoresUsingDiscoveringRestore()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new BlobContainerBackupTriggerMessage
            {
                BackupConfiguration = "blobAdHocBackup",
                OnlyIfModifiedSince = _timeMark
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new BlobContainerBackupTriggerMessageHandler(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages);

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new BlobContainerBackupMessageHandler().Handle((BlobContainerBackupMessage)backupMessage);
            }

            var provider = new BlobStorageProvider();
            provider.DeleteContainer("tests-b1-container");
            provider.DeleteContainer("tests-b1-test-container");
            provider.DeleteContainer("tests-b2-container");

            AssertBackup();

            // trigger the restore
            var restoreProgressMonitor = BlobContainerRestoreTrigger.TriggerForAllTables(
                "blobbackuptests", DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), mockBus.Object);

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new BlobContainerRestoreMessageHandler().Handle((BlobContainerRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestore();

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true);

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void BackupTriggerMessageHandlerFailsProgressRecordsWhenExceptionIsThrownFromSend()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new BlobContainerBackupTriggerMessage
            {
                BackupConfiguration = "blobAdHocBackup"
            };

            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Throws(new Exception());

            var triggerHandler = new BlobContainerBackupTriggerMessageHandler(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            var testExceptionThrown = false;

            try
            {
                triggerHandler.Handle(triggerMessage);
            }
            catch (DontBotherAgainException)
            {
                testExceptionThrown = true;
            }

            // assert
            Assert.IsTrue(testExceptionThrown, "Expected the DontBotherAgainException exception");

            mockBus.Verify(x => x.Send(It.IsAny<IMessage[]>()), Times.AtLeastOnce());

            // trigger record
            var triggerRecord = BackupTriggerRecord.CreateContext().BackupTriggerRecords.ToArray();

            Assert.AreEqual(1, triggerRecord.Length);

            Assert.AreEqual("blobAdHocBackup", triggerRecord[0].PartitionKey);
            Assert.AreEqual(634330656053450000, DateTime.Parse(triggerRecord[0].RowKey).Ticks);
            Assert.AreEqual("Blob 2011-02-12 00:00:05 AdHoc", triggerRecord[0].BackupId);
            Assert.AreEqual("blobbackuptests", triggerRecord[0].BackupContainer);
            Assert.AreEqual(0, triggerRecord[0].ProgressRecordsNumber);

            // progress records
            var progressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(2, progressRecords.Length);

            Assert.AreEqual("Blob 2011-02-12 00:00:05 AdHoc", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsFalse(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[0].FailureReason);
            StringAssert.Contains("tests-b1-container", progressRecords[0].Name);

            Assert.AreEqual("Blob 2011-02-12 00:00:05 AdHoc", progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsFalse(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[1].FailureReason);
            StringAssert.Contains("tests-b1-test-container", progressRecords[1].Name);
        }

        [Test]
        public void TableRestoreTriggerFailsProgressRecordsWhenExceptionIsThrownFromSend()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Throws(new Exception());

            var testExceptionThrown = false;

            try
            {
                BlobContainerRestoreTrigger.Trigger(
                    "blobbackuptests", DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), 4, mockBus.Object, "tests-b1-container");
            }
            catch (DontBotherAgainException)
            {
                testExceptionThrown = true;
            }

            Assert.IsTrue(testExceptionThrown, "Expected the DontBotherAgainException exception");

            mockBus.Verify(x => x.Send(It.IsAny<IMessage[]>()), Times.AtLeastOnce());

            var progressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray();

            Assert.AreEqual(1, progressRecords.Length);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsFalse(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[0].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[0].ContainerName);
            Assert.AreEqual("tests-b1-container", progressRecords[0].Name);
        }

        private void ArrangeData()
        {
            var provider = (IBlobOperations)new BlobStorageProvider();

            using (var sourceStream = SampleData.StringToStream())
            {
                sourceStream.ResetToBegining();
                provider.PutBlob("tests-b1-container/blob1", sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob("tests-b2-container/blob2", sourceStream);

                _timeMark = DateTime.UtcNow; // don't use DefaultClock.CurrentTime here
                Thread.Sleep(1000);

                sourceStream.ResetToBegining();
                provider.PutBlob("tests-b1-test-container/blob3", sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob("tests-b2-container/sub1/blob4", sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob("tests-b1-container/sub1/sub2/blob5", sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob("tests-b1-test42-container/blob6", sourceStream);
            }
        }

        private void AssertBackupTriggerResults(IList<IMessage> backupMessages, string backupId = "Blob 2011-02-12 00:00:05 AdHoc")
        {
            // messages
            Assert.IsNotNull(backupMessages);
            Assert.AreEqual(2, backupMessages.Count);

            var backupMessage = backupMessages[0];

            Assert.IsInstanceOf<BlobContainerBackupMessage>(backupMessage);
            Assert.AreEqual(backupId, ((BlobContainerBackupMessage)backupMessage).BackupId);
            Assert.AreEqual("blobbackuptests", ((BlobContainerBackupMessage)backupMessage).DestinationContainerName);
            Assert.AreEqual("00634330656053450000:000001", ((BlobContainerBackupMessage)backupMessage).JobId);
            Assert.AreEqual(4, ((BlobContainerBackupMessage)backupMessage).MaxDegreeOfParallelism);
            Assert.AreEqual(_timeMark, ((BlobContainerBackupMessage)backupMessage).OnlyIfModifiedSince);

            Assert.AreEqual(2, ((BlobContainerBackupMessage)backupMessage).SourceContainerNames.Length);
            CollectionAssert.Contains(((BlobContainerBackupMessage)backupMessage).SourceContainerNames, "tests-b1-container");
            CollectionAssert.Contains(((BlobContainerBackupMessage)backupMessage).SourceContainerNames, "tests-b1-test42-container");

            backupMessage = backupMessages[1];

            Assert.IsInstanceOf<BlobContainerBackupMessage>(backupMessage);

            Assert.AreEqual(backupId, ((BlobContainerBackupMessage)backupMessage).BackupId);
            Assert.AreEqual("blobbackuptests", ((BlobContainerBackupMessage)backupMessage).DestinationContainerName);
            Assert.AreEqual("00634330656053450000:000002", ((BlobContainerBackupMessage)backupMessage).JobId);
            Assert.AreEqual(4, ((BlobContainerBackupMessage)backupMessage).MaxDegreeOfParallelism);
            Assert.AreEqual(_timeMark, ((BlobContainerBackupMessage)backupMessage).OnlyIfModifiedSince);

            Assert.AreEqual(1, ((BlobContainerBackupMessage)backupMessage).SourceContainerNames.Length);
            CollectionAssert.Contains(((BlobContainerBackupMessage)backupMessage).SourceContainerNames, "tests-b1-test-container");

            // trigger record
            AssertBackupTriggerRecord(false, backupId);

            // progress records
            var progressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(2, progressRecords.Length);

            Assert.AreEqual(backupId, progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNull(progressRecords[0].Completed);
            Assert.IsNull(progressRecords[0].Successful);
            Assert.IsNull(progressRecords[0].FailureReason);
            Assert.AreEqual("tests-b1-container;tests-b1-test42-container", progressRecords[0].Name);
            Assert.AreEqual("blobbackuptests", progressRecords[0].ContainerName);

            Assert.AreEqual(backupId, progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNull(progressRecords[1].Completed);
            Assert.IsNull(progressRecords[1].Successful);
            Assert.IsNull(progressRecords[1].FailureReason);
            Assert.AreEqual("tests-b1-test-container", progressRecords[1].Name);
            Assert.AreEqual("blobbackuptests", progressRecords[1].ContainerName);
        }

        private static void AssertBackupTriggerRecord(bool completed, string backupId = "Blob 2011-02-12 00:00:05 AdHoc")
        {
            var triggerRecord = BackupTriggerRecord.CreateContext().BackupTriggerRecords.ToArray();

            Assert.AreEqual(1, triggerRecord.Length);

            Assert.AreEqual("blobAdHocBackup", triggerRecord[0].PartitionKey);
            Assert.AreEqual(634330656053450000, DateTime.Parse(triggerRecord[0].RowKey).Ticks);
            Assert.AreEqual(backupId, triggerRecord[0].BackupId);
            Assert.AreEqual("blobbackuptests", triggerRecord[0].BackupContainer);
            Assert.AreEqual(2, triggerRecord[0].ProgressRecordsNumber);

            if (completed)
            {
                Assert.IsNotNull(triggerRecord[0].Successful);
                Assert.IsTrue(triggerRecord[0].Successful.GetValueOrDefault());
                Assert.IsNotNull(triggerRecord[0].Completed);
            }
            else
            {
                Assert.IsNull(triggerRecord[0].Successful);
                Assert.IsNull(triggerRecord[0].Completed);
            }
        }

        private static void AssertBackup(string backupId = "Blob 2011-02-12 00:00:05 AdHoc")
        {
            var provider = new BlobStorageProvider();

            var blobList = provider.FlatBlobList("blobbackuptests").ToList();
            Assert.AreEqual(3, blobList.Count);
            CollectionAssert.Contains(blobList, "blobbackuptests/tests-b1-test42-container/blob6");
            CollectionAssert.Contains(blobList, "blobbackuptests/tests-b1-container/sub1/sub2/blob5");
            CollectionAssert.Contains(blobList, "blobbackuptests/tests-b1-test-container/blob3");

            AssertBlobContent("blobbackuptests/tests-b1-test42-container/blob6");
            AssertBlobContent("blobbackuptests/tests-b1-container/sub1/sub2/blob5");
            AssertBlobContent("blobbackuptests/tests-b1-test-container/blob3");

            AssertCompletedBackupProgress(backupId);
        }

        private static void AssertCompletedBackupProgress(string backupId = "Blob 2011-02-12 00:00:05 AdHoc")
        {
            var progressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(2, progressRecords.Length);

            Assert.AreEqual(backupId, progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsTrue(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[0].FailureReason);
            Assert.AreEqual("tests-b1-container;tests-b1-test42-container", progressRecords[0].Name);
            Assert.AreEqual("blobbackuptests", progressRecords[0].ContainerName);

            Assert.AreEqual(backupId, progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsTrue(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[1].FailureReason);
            Assert.AreEqual("tests-b1-test-container", progressRecords[1].Name);
            Assert.AreEqual("blobbackuptests", progressRecords[1].ContainerName);
        }

        private static void AssertRestoreTriggerResults(IList<IMessage> restoreMessages)
        {
            Assert.IsNotNull(restoreMessages);
            Assert.AreEqual(3, restoreMessages.Count);

            Assert.IsInstanceOf<BlobContainerRestoreMessage>(restoreMessages[0]);
            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", ((BlobContainerRestoreMessage)restoreMessages[0]).RestoreId);
            Assert.AreEqual("blobbackuptests", ((BlobContainerRestoreMessage)restoreMessages[0]).BaseContainerName);
            Assert.AreEqual("00634330656053450000:000001", ((BlobContainerRestoreMessage)restoreMessages[0]).JobId);
            Assert.AreEqual(4, ((BlobContainerRestoreMessage)restoreMessages[0]).MaxDegreeOfParallelism);

            Assert.IsInstanceOf<BlobContainerRestoreMessage>(restoreMessages[1]);
            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", ((BlobContainerRestoreMessage)restoreMessages[1]).RestoreId);
            Assert.AreEqual("blobbackuptests", ((BlobContainerRestoreMessage)restoreMessages[1]).BaseContainerName);
            Assert.AreEqual("00634330656053450000:000002", ((BlobContainerRestoreMessage)restoreMessages[1]).JobId);
            Assert.AreEqual(4, ((BlobContainerRestoreMessage)restoreMessages[1]).MaxDegreeOfParallelism);

            Assert.IsInstanceOf<BlobContainerRestoreMessage>(restoreMessages[2]);
            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", ((BlobContainerRestoreMessage)restoreMessages[2]).RestoreId);
            Assert.AreEqual("blobbackuptests", ((BlobContainerRestoreMessage)restoreMessages[2]).BaseContainerName);
            Assert.AreEqual("00634330656053450000:000003", ((BlobContainerRestoreMessage)restoreMessages[2]).JobId);
            Assert.AreEqual(4, ((BlobContainerRestoreMessage)restoreMessages[2]).MaxDegreeOfParallelism);

            var progressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray().OrderBy(x => x.Name).ToArray();

            Assert.AreEqual(3, progressRecords.Length);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNull(progressRecords[0].Completed);
            Assert.IsNull(progressRecords[0].Successful);
            Assert.IsNull(progressRecords[0].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[0].ContainerName);
            Assert.AreEqual("tests-b1-container", progressRecords[0].Name);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000003", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNull(progressRecords[1].Completed);
            Assert.IsNull(progressRecords[1].Successful);
            Assert.IsNull(progressRecords[1].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[1].ContainerName);
            Assert.AreEqual("tests-b1-test42-container", progressRecords[1].Name);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[2].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[2].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[2].Initiated.Ticks);
            Assert.IsNull(progressRecords[2].Started);
            Assert.IsNull(progressRecords[2].Completed);
            Assert.IsNull(progressRecords[2].Successful);
            Assert.IsNull(progressRecords[2].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[2].ContainerName);
            Assert.AreEqual("tests-b1-test-container", progressRecords[2].Name);
        }

        private static void AssertRestore()
        {
            var provider = new BlobStorageProvider();

            var blobList = provider.FlatBlobList("tests-b1-container").ToList();
            Assert.AreEqual(1, blobList.Count);
            CollectionAssert.Contains(blobList, "tests-b1-container/sub1/sub2/blob5");

            blobList = provider.FlatBlobList("tests-b1-test-container").ToList();
            Assert.AreEqual(1, blobList.Count);
            CollectionAssert.Contains(blobList, "tests-b1-test-container/blob3");

            blobList = provider.FlatBlobList("tests-b2-container").ToList();
            Assert.AreEqual(0, blobList.Count);

            AssertBlobContent("tests-b1-container/sub1/sub2/blob5");
            AssertBlobContent("tests-b1-test-container/blob3");

            AssertCompletedRestoreProgress();
        }

        private static void AssertCompletedRestoreProgress()
        {
            var progressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray().OrderBy(x => x.Name).ToArray();

            Assert.AreEqual(3, progressRecords.Length);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[0].PartitionKey);
            StringAssert.StartsWith("00634330656053450000:00000", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsTrue(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[0].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[0].ContainerName);
            Assert.AreEqual("tests-b1-container", progressRecords[0].Name);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[1].PartitionKey);
            StringAssert.StartsWith("00634330656053450000:00000", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsTrue(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[1].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[1].ContainerName);
            Assert.AreEqual("tests-b1-test42-container", progressRecords[1].Name);

            Assert.AreEqual("Blob 2011-02-12T00:00:05.345Z", progressRecords[2].PartitionKey);
            StringAssert.StartsWith("00634330656053450000:00000", progressRecords[2].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[2].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[2].Started);
            Assert.IsNotNull(progressRecords[2].Completed);
            Assert.IsNotNull(progressRecords[2].Successful);
            Assert.IsTrue(progressRecords[2].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[2].FailureReason);
            Assert.AreEqual("blobbackuptests", progressRecords[2].ContainerName);
            Assert.AreEqual("tests-b1-test-container", progressRecords[2].Name);
        }

        private static void AssertBlobContent(string blobName)
        {
            using (var stream = ((IBlobOperations)new BlobStorageProvider()).GetBlob(blobName))
            {
                Assert.AreEqual(SampleData, stream.StreamToString());
            }
        }

        private static void CleanTestData()
        {
            BackupTriggerRecord.CreateContext().DeleteTables();

            BackupProgressRecord.CreateContext().DeleteTables();

            RestoreProgressRecord.CreateContext().DeleteTables();

            var provider = new BlobStorageProvider();

            provider.DeleteContainer("tests-b1-container");
            provider.DeleteContainer("tests-b1-test-container");
            provider.DeleteContainer("tests-b1-test42-container");
            provider.DeleteContainer("tests-b2-container");
            provider.DeleteContainer("blobbackuptests");
        }
    }
}
