﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using System.Globalization;
using System.Linq;
using System.Threading;
using DocaAzure.Backup.Interfaces;
using DocaAzure.Backup.Service;
using DocaAzure.Backup.Service.Tables;
using DocaAzure.Messaging;
using DocaAzure.Storage;
using DocaAzure.TestUtils;
using DocaLabs.Storage.DataServicesUtils;
using DocaLabs.Utils;
using Microsoft.WindowsAzure.StorageClient;
using Moq;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests.Service
{
    [TestFixture]
    public class TableTriggerServiceTest
    {
        public const string InitiatedTime = "2011-02-12T00:00:05.345Z";

        #region tests classes

        [Table(TableName)]
        public class Product : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProducts";

            public Product()
            {
            }

            public Product(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public string Name { get; set; }
        }

        [Table(TableName)]
        public class Category : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsCategories";

            public Category()
            {
            }

            public Category(string customerId, string categoryId)
                : base(customerId, categoryId)
            {
            }

            public string Name { get; set; }
        }

        public interface IContext : ISimpleContextBase
        {
            IDbSet<Product> Products { get; set; }
            IDbSet<Category> Categories { get; set; }
        }

        public class TableBackupTriggerMessageHandlerTestTarget : TableBackupTriggerMessageHandler
        {
            public TableBackupTriggerMessageHandlerTestTarget(IBus bus)
                : base(bus)
            {
            }
        }

        #endregion tests classes

        public class RestoreDefaultProgressMonitorTestTarget : RestoreDefaultProgressMonitor
        {
            public static bool RestoreSuccessful { get; set; }

            protected override MonitoringResults OnSuccess(string restoreId, MonitoringResults result)
            {
                RestoreSuccessful = true;
                return base.OnSuccess(restoreId, result);
            }
        }

        private IClock _originalClock;
        private FixedClock _fixedClock;

        private Dictionary<string, Product> _originalProducts;
        private Dictionary<string, Category> _originalCategories;
        private DateTime _halfProductEntitiesMark;

        private static TimeSpan OriginalBackupTimeout { get; set; }
        private static TimeSpan OriginalBackupPollingTimeout { get; set; }
        private static TimeSpan OriginalRestoreTimeout { get; set; }
        private static TimeSpan OriginalRestorePollingTimeout { get; set; }
        private static Type OriginalRestoreProgressMonitorType { get; set; }

        [SetUp]
        public void Setup()
        {
            RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful = false;

            OriginalBackupTimeout = BackupDefaultProgressMonitor.BackupTimeout;
            OriginalBackupPollingTimeout = BackupDefaultProgressMonitor.PollingTimeout;

            OriginalRestoreTimeout = RestoreDefaultProgressMonitor.RestoreTimeout;
            OriginalRestorePollingTimeout = RestoreDefaultProgressMonitor.PollingTimeout;

            BackupDefaultProgressMonitor.BackupTimeout = TimeSpan.FromMinutes(30);
            BackupDefaultProgressMonitor.PollingTimeout = TimeSpan.FromMilliseconds(100);

            RestoreDefaultProgressMonitor.RestoreTimeout = TimeSpan.FromMinutes(30);
            RestoreDefaultProgressMonitor.PollingTimeout = TimeSpan.FromMilliseconds(100);

            OriginalRestoreProgressMonitorType = DefaultBackupServiceFactory.RestoreProgressMonitorType;
            DefaultBackupServiceFactory.RestoreProgressMonitorType = typeof (RestoreDefaultProgressMonitorTestTarget);

            _originalClock = DefaultClock.Current;
            DefaultClock.Current = _fixedClock = new FixedClock();

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            ClearAllPossibleTestData();
        }

        [TearDown]
        public void TearDown()
        {
            ClearAllPossibleTestData();

            DefaultClock.Current = _originalClock;

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;

            DefaultConfigurationManager.Default = null;

            BackupDefaultProgressMonitor.BackupTimeout = OriginalBackupTimeout;
            BackupDefaultProgressMonitor.PollingTimeout = OriginalBackupPollingTimeout;

            RestoreDefaultProgressMonitor.RestoreTimeout = OriginalRestoreTimeout;
            RestoreDefaultProgressMonitor.PollingTimeout = OriginalRestorePollingTimeout;

            DefaultBackupServiceFactory.RestoreProgressMonitorType = OriginalRestoreProgressMonitorType;
        }

        [Test]
        public void RollingWeekDayBackupAndRestoreUsingDiscoveringBackupConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableDiscoveringWeekDayBackup"
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);


            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages, "Table 2011-02-12 RollingWeekDay", "backup-table-rolling-weekday-saturday", "tableDiscoveringWeekDayBackup");

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new TableBackupMessageHandler().Handle((TableBackupMessage)backupMessage);
            }

            RecreateBlankTables();

            AssertCompletedBackupProgress();

            // trigger the restore
            var restoreProgressMonitor = TableRestoreTrigger.Trigger("backup-table-rolling-weekday-saturday", DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), mockBus.Object,
                new[] 
                {
                    new TableRestoreJob
                    {
                        TableName = "IntegrationBackupTestsCategories",
                        MaxDegreeOfParallelism = 2
                    },
                    new TableRestoreJob
                    {
                        TableName = "IntegrationBackupTestsProducts",
                        MaxDegreeOfParallelism = 2
                    }
                }
            );

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new TableRestoreMessageHandler().Handle((TableRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestoredData(_originalProducts.Count);

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true, "Table 2011-02-12 RollingWeekDay", "backup-table-rolling-weekday-saturday", "tableDiscoveringWeekDayBackup");

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void RollingWeekDayBackupAndRestoreUsingDiscoveringBackupAndRestoreConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableDiscoveringWeekDayBackup"
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages, "Table 2011-02-12 RollingWeekDay", "backup-table-rolling-weekday-saturday", "tableDiscoveringWeekDayBackup");

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new TableBackupMessageHandler().Handle((TableBackupMessage)backupMessage);
            }

            RecreateBlankTables();

            AssertCompletedBackupProgress();

            // trigger the restore
            var restoreProgressMonitor = TableRestoreTrigger.TriggerForAllTables(
                "backup-table-rolling-weekday-saturday", DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), mockBus.Object, 2);

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new TableRestoreMessageHandler().Handle((TableRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestoredData(_originalProducts.Count);

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true, "Table 2011-02-12 RollingWeekDay", "backup-table-rolling-weekday-saturday", "tableDiscoveringWeekDayBackup");

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void RollingWeekDayBackupAndRestoreUsingExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableWeekDayBackup"
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages);

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new TableBackupMessageHandler().Handle((TableBackupMessage)backupMessage);
            }

            RecreateBlankTables();

            AssertCompletedBackupProgress();

            // trigger the restore
            var restoreProgressMonitor = TableRestoreTrigger.Trigger("backup-table-rolling-weekday-saturday",
                DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal),
                mockBus.Object,
                new[] {
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsCategories",
                            MaxDegreeOfParallelism = 2
                        },
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsProducts",
                            MaxDegreeOfParallelism = 2
                        }});

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new TableRestoreMessageHandler().Handle((TableRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestoredData(_originalProducts.Count);

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true);

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void BackupTypeInTriggerMessageOverridesBackupTypeInConfigurationUsingExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            var initiatedTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            var containerName = TableBackupTrigger.GetContainerName(initiatedTime, BackupType.Daily, null);

            ArrangeData();

            var triggerMessage = new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableWeekDayBackup",
                BackupType = BackupType.Daily
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = initiatedTime;

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages, "Table 2011-02-12 Daily", containerName);

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new TableBackupMessageHandler().Handle((TableBackupMessage)backupMessage);
            }

            RecreateBlankTables();

            AssertCompletedBackupProgress("Table 2011-02-12 Daily", containerName);

            // trigger the restore
            var restoreProgressMonitor = TableRestoreTrigger.Trigger(containerName,
                initiatedTime,
                mockBus.Object,
                new[] {
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsCategories",
                            MaxDegreeOfParallelism = 2
                        },
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsProducts",
                            MaxDegreeOfParallelism = 2
                        }});

            AssertRestoreTriggerResults(sendMessages, containerName);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new TableRestoreMessageHandler().Handle((TableRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestoredData(_originalProducts.Count, containerName);

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true, "Table 2011-02-12 Daily", containerName);

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        [Test]
        public void TableRestoreTriggerFailsProgressRecordsWhenExceptionIsThrownFromSendAndUsingExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Throws(new Exception());

            var testExceptionThrown = false;

            try
            {
                TableRestoreTrigger.Trigger("backup-table-rolling-weekday-saturday",
                    DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal),
                    mockBus.Object,
                    new[] {
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsCategories",
                            MaxDegreeOfParallelism = 2
                        },
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsProducts",
                            MaxDegreeOfParallelism = 2
                        }});
            }
            catch (InvalidOperationException)
            {
                testExceptionThrown = true;
            }

            Assert.IsTrue(testExceptionThrown, "Expected the DontBotherAgainException exception");

            mockBus.Verify(x => x.Send(It.IsAny<IMessage[]>()), Times.AtLeastOnce());

            var progressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray();

            Assert.AreEqual(2, progressRecords.Length);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsFalse(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[0].FailureReason);
            Assert.AreEqual("IntegrationBackupTestsCategories", progressRecords[0].Name);
            Assert.AreEqual("backup-table-rolling-weekday-saturday", progressRecords[0].ContainerName);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsFalse(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[1].FailureReason);
            Assert.AreEqual("IntegrationBackupTestsProducts", progressRecords[1].Name);
            Assert.AreEqual("backup-table-rolling-weekday-saturday", progressRecords[1].ContainerName);
        }

        [Test]
        public void TableBackupTriggerFailsProgressRecordsWhenExceptionIsThrownFromSendAndUsingExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            var triggerMessage = new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableWeekDayBackup"
            };

            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Throws(new Exception());

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            var testExceptionThrown = false;

            try
            {
                triggerHandler.Handle(triggerMessage);
            }
            catch (DontBotherAgainException)
            {
                testExceptionThrown = true;
            }

            // assert
            Assert.IsTrue(testExceptionThrown, "Expected the DontBotherAgainException exception");

            mockBus.Verify(x => x.Send(It.IsAny<IMessage[]>()), Times.AtLeastOnce());

            // assert trigger record
            var triggerRecord = BackupTriggerRecord.CreateContext().BackupTriggerRecords.ToArray();

            Assert.AreEqual(1, triggerRecord.Length);

            Assert.AreEqual("tableWeekDayBackup", triggerRecord[0].PartitionKey);
            Assert.AreEqual(634330656053450000, DateTime.Parse(triggerRecord[0].RowKey).Ticks);
            Assert.AreEqual("Table 2011-02-12 RollingWeekDay", triggerRecord[0].BackupId);
            Assert.AreEqual("backup-table-rolling-weekday-saturday", triggerRecord[0].BackupContainer);
            Assert.AreEqual(0, triggerRecord[0].ProgressRecordsNumber);

            // assert progress records
            var progressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(3, progressRecords.Length);

            Assert.AreEqual("Table 2011-02-12 RollingWeekDay", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsFalse(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[0].FailureReason);
            StringAssert.Contains("IntegrationBackupTestsCategories", progressRecords[0].Name);

            Assert.AreEqual("Table 2011-02-12 RollingWeekDay", progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsFalse(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[1].FailureReason);
            StringAssert.Contains("IntegrationBackupTestsProducts", progressRecords[1].Name);

            Assert.AreEqual("Table 2011-02-12 RollingWeekDay", progressRecords[2].PartitionKey);
            Assert.AreEqual("00634330656053450000:000003", progressRecords[2].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[2].Initiated.Ticks);
            Assert.IsNull(progressRecords[2].Started);
            Assert.IsNotNull(progressRecords[2].Completed);
            Assert.IsNotNull(progressRecords[2].Successful);
            Assert.IsFalse(progressRecords[2].Successful.GetValueOrDefault());
            Assert.IsNotNull(progressRecords[2].FailureReason);
            StringAssert.Contains("IntegrationBackupTestsProducts", progressRecords[2].Name);
        }

        [Test]
        public void BackupAndRestoreWithOnlyIfModifiedSinceUsingQueryAndExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            BackupAndRestoreWithOnlyIfModifiedSinceTestRunner(new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableWeekDayBackup",
                OnlyIfModifiedSince = _halfProductEntitiesMark // ArrangeData must be called beforehand
            });
        }

        [Test]
        public void BackupAndRestoreWithOnlyIfModifiedSinceUsingFilterAndExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            BackupAndRestoreWithOnlyIfModifiedSinceTestRunner(new TableBackupTriggerMessage
            {
                BackupConfiguration = "tableWeekDayBackup",
                OnlyIfModifiedSince = _halfProductEntitiesMark,
                UseFilterForIfModifiedSince = true  // ArrangeData must be called beforehand
            });
        }

        [Test]
        public void NothingIsBackedUpButReportedProperlyIfTableDoesNotExistAndUsingExactConfiguration()
        {
            DefaultConfigurationManager.Default = DefaultConfigurationManager.GetExternal("DocaAzure.Backup.IntegrationTests.config");

            ArrangeData();

            var triggerMessage = new TableBackupTriggerMessage
            {
                BackupConfiguration = "unknownTableBackupTest"
            };

            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the messages
            Assert.IsNotNull(sendMessages);
            Assert.AreEqual(1, sendMessages.Length);

            Assert.IsInstanceOf<TableBackupMessage>(sendMessages[0]);
            Assert.AreEqual("Table 2011-02-12 Daily", ((TableBackupMessage)sendMessages[0]).BackupId);
            Assert.AreEqual("backup-table-daily-2011-02-12", ((TableBackupMessage)sendMessages[0]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000001", ((TableBackupMessage)sendMessages[0]).JobId);
            Assert.AreEqual(100, ((TableBackupMessage)sendMessages[0]).JobDescription.MaxBlocksInBlobs);
            Assert.AreEqual(3984588, ((TableBackupMessage)sendMessages[0]).JobDescription.MaxBlockSize);
            Assert.AreEqual(4, ((TableBackupMessage)sendMessages[0]).JobDescription.MaxDegreeOfParallelism);
            Assert.AreEqual("SomeVeryUnknownTestTable", ((TableBackupMessage)sendMessages[0]).JobDescription.TableName);
            Assert.AreEqual(16, ((TableBackupMessage)sendMessages[0]).JobDescription.Ranges.Count());
            Assert.IsNull(((PartitionKeyRange)((TableBackupMessage)sendMessages[0]).JobDescription.Ranges.ElementAt(0)).Min);
            Assert.AreEqual("1", ((PartitionKeyRange)((TableBackupMessage)sendMessages[0]).JobDescription.Ranges.ElementAt(0)).Max);
            Assert.AreEqual("f", ((PartitionKeyRange)((TableBackupMessage)sendMessages[0]).JobDescription.Ranges.ElementAt(15)).Min);
            Assert.IsNull(((PartitionKeyRange)((TableBackupMessage)sendMessages[0]).JobDescription.Ranges.ElementAt(15)).Max);

            // assert the trigger record
            var triggerRecord = BackupTriggerRecord.CreateContext().BackupTriggerRecords.ToArray();

            Assert.AreEqual(1, triggerRecord.Length);

            Assert.AreEqual("unknownTableBackupTest", triggerRecord[0].PartitionKey);
            Assert.AreEqual(634330656053450000, DateTime.Parse(triggerRecord[0].RowKey).Ticks);
            Assert.AreEqual("Table 2011-02-12 Daily", triggerRecord[0].BackupId);
            Assert.AreEqual("backup-table-daily-2011-02-12", triggerRecord[0].BackupContainer);
            Assert.AreEqual(1, triggerRecord[0].ProgressRecordsNumber);

            // assert the progress records
            var backupProgressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(1, backupProgressRecords.Length);

            Assert.AreEqual("Table 2011-02-12 Daily", backupProgressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", backupProgressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, backupProgressRecords[0].Initiated.Ticks);
            Assert.IsNull(backupProgressRecords[0].Started);
            Assert.IsNull(backupProgressRecords[0].Completed);
            Assert.IsNull(backupProgressRecords[0].Successful);
            Assert.IsNull(backupProgressRecords[0].FailureReason);
            StringAssert.Contains("SomeVeryUnknownTestTable", backupProgressRecords[0].Name);
            Assert.AreEqual("backup-table-daily-2011-02-12", backupProgressRecords[0].ContainerName);
            /**/

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new TableBackupMessageHandler().Handle((TableBackupMessage)backupMessage);
            }

            RecreateBlankTables();

            //AssertCompletedBackupProgress
            backupProgressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(1, backupProgressRecords.Length);

            Assert.AreEqual("Table 2011-02-12 Daily", backupProgressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", backupProgressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, backupProgressRecords[0].Initiated.Ticks);
            Assert.IsNotNull(backupProgressRecords[0].Started);
            Assert.IsNotNull(backupProgressRecords[0].Completed);
            Assert.IsNotNull(backupProgressRecords[0].Successful);
            Assert.IsTrue(backupProgressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNull(backupProgressRecords[0].FailureReason);
            StringAssert.Contains("SomeVeryUnknownTestTable", backupProgressRecords[0].Name);
            Assert.AreEqual("backup-table-daily-2011-02-12", backupProgressRecords[0].ContainerName);
            /**/

            // trigger the restore
            TableRestoreTrigger.Trigger("backup-table-daily-2011-02-12",
                DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal),
                mockBus.Object,
                new[] {
                        new TableRestoreJob
                        {
                            TableName = "SomeVeryUnknownTestTable",
                            MaxDegreeOfParallelism = 2
                        }});

            // AssertRestoreTriggerResults
            Assert.IsNotNull(sendMessages);
            Assert.AreEqual(1, sendMessages.Length);

            Assert.IsInstanceOf<TableRestoreMessage>(sendMessages[0]);
            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", ((TableRestoreMessage)sendMessages[0]).RestoreId);
            Assert.AreEqual("backup-table-daily-2011-02-12", ((TableRestoreMessage)sendMessages[0]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000001", ((TableRestoreMessage)sendMessages[0]).JobId);
            Assert.AreEqual(2, ((TableRestoreMessage)sendMessages[0]).JobDescription.MaxDegreeOfParallelism);
            Assert.AreEqual("SomeVeryUnknownTestTable", ((TableRestoreMessage)sendMessages[0]).JobDescription.TableName);


            var restoreProgressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray();

            Assert.AreEqual(1, restoreProgressRecords.Length);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", restoreProgressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", restoreProgressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, restoreProgressRecords[0].Initiated.Ticks);
            Assert.IsNull(restoreProgressRecords[0].Started);
            Assert.IsNull(restoreProgressRecords[0].Completed);
            Assert.IsNull(restoreProgressRecords[0].Successful);
            Assert.IsNull(restoreProgressRecords[0].FailureReason);
            Assert.AreEqual("SomeVeryUnknownTestTable", restoreProgressRecords[0].Name);
            Assert.AreEqual("backup-table-daily-2011-02-12", restoreProgressRecords[0].ContainerName);
            /**/

            // restore
            foreach (var backupMessage in sendMessages)
            {
                new TableRestoreMessageHandler().Handle((TableRestoreMessage)backupMessage);
            }

            // assert restore
            var tableClient = AzureStorageAccountFactory.Create().CreateCloudTableClient();
            Assert.IsFalse(tableClient.DoesTableExist("SomeVeryUnknownTestTable"));

            restoreProgressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray();

            Assert.AreEqual(1, restoreProgressRecords.Length);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", restoreProgressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", restoreProgressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, restoreProgressRecords[0].Initiated.Ticks);
            Assert.IsNotNull(restoreProgressRecords[0].Started);
            Assert.IsNotNull(restoreProgressRecords[0].Completed);
            Assert.IsNotNull(restoreProgressRecords[0].Successful);
            Assert.IsTrue(restoreProgressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNull(restoreProgressRecords[0].FailureReason);
            Assert.AreEqual("SomeVeryUnknownTestTable", restoreProgressRecords[0].Name);
            Assert.AreEqual("backup-table-daily-2011-02-12", restoreProgressRecords[0].ContainerName);
        }

        private void BackupAndRestoreWithOnlyIfModifiedSinceTestRunner(TableBackupTriggerMessage triggerMessage)
        {
            IMessage[] sendMessages = null;
            var mockBus = new Mock<IBus>();
            mockBus.Setup(x => x.Send(It.IsAny<IMessage[]>())).Callback((IMessage[] m) => { sendMessages = m; });

            var triggerHandler = new TableBackupTriggerMessageHandlerTestTarget(mockBus.Object);

            _fixedClock.ClockTime = DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal);

            // trigger the backup
            triggerHandler.Handle(triggerMessage);


            // assert the trigger results
            AssertBackupTriggerResults(sendMessages);

            // backup
            foreach (var backupMessage in sendMessages)
            {
                new TableBackupMessageHandler().Handle((TableBackupMessage)backupMessage);
            }

            RecreateBlankTables();

            AssertCompletedBackupProgress();

            // trigger the restore
            var restoreProgressMonitor = TableRestoreTrigger.Trigger("backup-table-rolling-weekday-saturday",
                DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal),
                mockBus.Object,
                new[] {
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsCategories",
                            MaxDegreeOfParallelism = 2
                        },
                        new TableRestoreJob
                        {
                            TableName = "IntegrationBackupTestsProducts",
                            MaxDegreeOfParallelism = 2
                        }});

            AssertRestoreTriggerResults(sendMessages);

            // restore
            foreach (var backupMessage in sendMessages)
            {
                Assert.IsFalse(restoreProgressMonitor.IsCompleted);
                new TableRestoreMessageHandler().Handle((TableRestoreMessage)backupMessage);
            }

            // assert restore
            AssertRestoredData(_originalProducts.Count / 2);

            // assert here in order to get chance for the monitor routine to catch up
            AssertBackupTriggerRecord(true);

            // assert here in order to get chance for the monitor routine to catch up
            Assert.AreEqual(MonitoringResults.Result.Success, restoreProgressMonitor.Result.RunResult);
            Assert.IsTrue(restoreProgressMonitor.IsCompleted);
            Assert.IsTrue(RestoreDefaultProgressMonitorTestTarget.RestoreSuccessful);
        }

        private void ArrangeData()
        {
            var backupContext = BackupProgressRecord.CreateContext();
            backupContext.DeleteTables();
            backupContext.CreateTablesIfNotExist();

            var restoreContext = RestoreProgressRecord.CreateContext();
            restoreContext.DeleteTables();
            restoreContext.CreateTablesIfNotExist();

            RecreateBlankTables();

            var context = SimpleContextBase.CreateInstance<IContext>();

            _originalProducts = new Dictionary<string, Product>();
            for (var i = 0; i < 100; i++)
            {
                var product = new Product("56fc6a50-d937-433f-a3d9-cd566aefdca1", Guid.NewGuid().ToString())
                {
                    Name = "Some product " + Guid.NewGuid()
                };

                _originalProducts.Add(product.RowKey, product);
                context.Products.Add(product);

                if (i == 49)
                {
                    context.SaveBatchWithRetries();
                    _halfProductEntitiesMark = DateTime.UtcNow; // don't use DefaultClock.CurrentTime here
                    Thread.Sleep(1000);
                }
            }
            context.SaveBatchWithRetries();

            _originalCategories = new Dictionary<string, Category>();
            for (var i = 0; i < 50; i++)
            {
                var category = new Category("1b5639d0-d250-4d5f-b955-8b1273dddd20", Guid.NewGuid().ToString())
                {
                    Name = "Some category " + Guid.NewGuid()
                };

                _originalCategories.Add(category.RowKey, category);
                context.Categories.Add(category);
            }
            context.SaveBatchWithRetries();
        }

        private static void AssertBackupTriggerResults(IList<IMessage> backupMessages, string expectedPartitionKey = "Table 2011-02-12 RollingWeekDay", string expectedContainerName = "backup-table-rolling-weekday-saturday", string triggerRecordPartitionKey = "tableWeekDayBackup")
        {
            // messages
            Assert.IsNotNull(backupMessages);
            Assert.AreEqual(3, backupMessages.Count);

            Assert.IsInstanceOf<TableBackupMessage>(backupMessages[0]);
            Assert.AreEqual(expectedPartitionKey, ((TableBackupMessage)backupMessages[0]).BackupId);
            Assert.AreEqual(expectedContainerName, ((TableBackupMessage)backupMessages[0]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000001", ((TableBackupMessage)backupMessages[0]).JobId);
            Assert.AreEqual(100, ((TableBackupMessage)backupMessages[0]).JobDescription.MaxBlocksInBlobs);
            Assert.AreEqual(3984588, ((TableBackupMessage)backupMessages[0]).JobDescription.MaxBlockSize);
            Assert.AreEqual(1, ((TableBackupMessage)backupMessages[0]).JobDescription.MaxDegreeOfParallelism);
            Assert.AreEqual("IntegrationBackupTestsCategories", ((TableBackupMessage)backupMessages[0]).JobDescription.TableName);
            Assert.AreEqual(16, ((TableBackupMessage)backupMessages[0]).JobDescription.Ranges.Count());
            Assert.IsNull(((PartitionKeyRange)((TableBackupMessage)backupMessages[0]).JobDescription.Ranges.ElementAt(0)).Min);
            Assert.AreEqual("1", ((PartitionKeyRange)((TableBackupMessage)backupMessages[0]).JobDescription.Ranges.ElementAt(0)).Max);
            Assert.AreEqual("f", ((PartitionKeyRange)((TableBackupMessage)backupMessages[0]).JobDescription.Ranges.ElementAt(15)).Min);
            Assert.IsNull(((PartitionKeyRange)((TableBackupMessage)backupMessages[0]).JobDescription.Ranges.ElementAt(15)).Max);

            Assert.IsInstanceOf<TableBackupMessage>(backupMessages[1]);
            Assert.AreEqual(expectedPartitionKey, ((TableBackupMessage)backupMessages[1]).BackupId);
            Assert.AreEqual(expectedContainerName, ((TableBackupMessage)backupMessages[1]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000002", ((TableBackupMessage)backupMessages[1]).JobId);
            Assert.AreEqual(100, ((TableBackupMessage)backupMessages[1]).JobDescription.MaxBlocksInBlobs);
            Assert.AreEqual(3984588, ((TableBackupMessage)backupMessages[1]).JobDescription.MaxBlockSize);
            Assert.AreEqual(4, ((TableBackupMessage)backupMessages[1]).JobDescription.MaxDegreeOfParallelism);
            Assert.AreEqual("IntegrationBackupTestsProducts", ((TableBackupMessage)backupMessages[1]).JobDescription.TableName);
            Assert.AreEqual(16, ((TableBackupMessage)backupMessages[1]).JobDescription.Ranges.Count());
            Assert.IsNull(((PartitionKeyRange)((TableBackupMessage)backupMessages[1]).JobDescription.Ranges.ElementAt(0)).Min);
            Assert.AreEqual("08", ((PartitionKeyRange)((TableBackupMessage)backupMessages[1]).JobDescription.Ranges.ElementAt(0)).Max);
            Assert.AreEqual("78", ((PartitionKeyRange)((TableBackupMessage)backupMessages[1]).JobDescription.Ranges.ElementAt(15)).Min);
            Assert.AreEqual("80", ((PartitionKeyRange)((TableBackupMessage)backupMessages[1]).JobDescription.Ranges.ElementAt(15)).Max);

            Assert.IsInstanceOf<TableBackupMessage>(backupMessages[2]);
            Assert.AreEqual(expectedPartitionKey, ((TableBackupMessage)backupMessages[2]).BackupId);
            Assert.AreEqual(expectedContainerName, ((TableBackupMessage)backupMessages[2]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000003", ((TableBackupMessage)backupMessages[2]).JobId);
            Assert.AreEqual(100, ((TableBackupMessage)backupMessages[2]).JobDescription.MaxBlocksInBlobs);
            Assert.AreEqual(3984588, ((TableBackupMessage)backupMessages[2]).JobDescription.MaxBlockSize);
            Assert.AreEqual(4, ((TableBackupMessage)backupMessages[2]).JobDescription.MaxDegreeOfParallelism);
            Assert.AreEqual("IntegrationBackupTestsProducts", ((TableBackupMessage)backupMessages[2]).JobDescription.TableName);
            Assert.AreEqual(16, ((TableBackupMessage)backupMessages[2]).JobDescription.Ranges.Count());
            Assert.AreEqual("80", ((PartitionKeyRange)((TableBackupMessage)backupMessages[2]).JobDescription.Ranges.ElementAt(0)).Min);
            Assert.AreEqual("88", ((PartitionKeyRange)((TableBackupMessage)backupMessages[2]).JobDescription.Ranges.ElementAt(0)).Max);
            Assert.AreEqual("f8", ((PartitionKeyRange)((TableBackupMessage)backupMessages[2]).JobDescription.Ranges.ElementAt(15)).Min);
            Assert.IsNull(((PartitionKeyRange)((TableBackupMessage)backupMessages[2]).JobDescription.Ranges.ElementAt(15)).Max);

            // trigger record
            AssertBackupTriggerRecord(false, expectedPartitionKey, expectedContainerName, triggerRecordPartitionKey);

            // progress records
            var progressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(3, progressRecords.Length);

            Assert.AreEqual(expectedPartitionKey, progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNull(progressRecords[0].Completed);
            Assert.IsNull(progressRecords[0].Successful);
            Assert.IsNull(progressRecords[0].FailureReason);
            StringAssert.Contains("IntegrationBackupTestsCategories", progressRecords[0].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[0].ContainerName);

            Assert.AreEqual(expectedPartitionKey, progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNull(progressRecords[1].Completed);
            Assert.IsNull(progressRecords[1].Successful);
            Assert.IsNull(progressRecords[1].FailureReason);
            StringAssert.Contains("IntegrationBackupTestsProducts", progressRecords[1].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[1].ContainerName);

            Assert.AreEqual(expectedPartitionKey, progressRecords[2].PartitionKey);
            Assert.AreEqual("00634330656053450000:000003", progressRecords[2].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[2].Initiated.Ticks);
            Assert.IsNull(progressRecords[2].Started);
            Assert.IsNull(progressRecords[2].Completed);
            Assert.IsNull(progressRecords[2].Successful);
            Assert.IsNull(progressRecords[2].FailureReason);
            StringAssert.Contains("IntegrationBackupTestsProducts", progressRecords[2].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[2].ContainerName);
        }

        private static void AssertBackupTriggerRecord(bool completed, string expectedPartitionKey = "Table 2011-02-12 RollingWeekDay", string expectedContainerName = "backup-table-rolling-weekday-saturday", string triggerRecordPartitionKey = "tableWeekDayBackup")
        {
            // trigger record
            var triggerRecord = BackupTriggerRecord.CreateContext().BackupTriggerRecords.ToArray();

            Assert.AreEqual(1, triggerRecord.Length);

            Assert.AreEqual(triggerRecordPartitionKey, triggerRecord[0].PartitionKey);
            Assert.AreEqual(634330656053450000, DateTime.Parse(triggerRecord[0].RowKey).Ticks);
            Assert.AreEqual(expectedPartitionKey, triggerRecord[0].BackupId);
            Assert.AreEqual(expectedContainerName, triggerRecord[0].BackupContainer);
            Assert.AreEqual(3, triggerRecord[0].ProgressRecordsNumber);

            if(completed)
            {
                Assert.IsNotNull(triggerRecord[0].Successful);
                Assert.IsTrue(triggerRecord[0].Successful.GetValueOrDefault());
                Assert.IsNotNull(triggerRecord[0].Completed);
            }
            else
            {
                Assert.IsNull(triggerRecord[0].Successful);
                Assert.IsNull(triggerRecord[0].Completed);
            }
        }

        private static void AssertCompletedBackupProgress(string expectedPartitionKey = "Table 2011-02-12 RollingWeekDay", string expectedContainerName = "backup-table-rolling-weekday-saturday")
        {
            var progressRecords = BackupProgressRecord.CreateContext().BackupProgressRecords.ToArray();

            Assert.AreEqual(3, progressRecords.Length);

            Assert.AreEqual(expectedPartitionKey, progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsTrue(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[0].FailureReason);
            StringAssert.Contains("integrationbackuptestscategories", progressRecords[0].Name.ToLower());
            Assert.AreEqual(expectedContainerName, progressRecords[0].ContainerName);

            Assert.AreEqual(expectedPartitionKey, progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsTrue(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[1].FailureReason);
            StringAssert.Contains("integrationbackuptestsproducts", progressRecords[1].Name.ToLower());
            Assert.AreEqual(expectedContainerName, progressRecords[1].ContainerName);

            Assert.AreEqual(expectedPartitionKey, progressRecords[2].PartitionKey);
            Assert.AreEqual("00634330656053450000:000003", progressRecords[2].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[2].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[2].Started);
            Assert.IsNotNull(progressRecords[2].Completed);
            Assert.IsNotNull(progressRecords[2].Successful);
            Assert.IsTrue(progressRecords[2].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[2].FailureReason);
            StringAssert.Contains("integrationbackuptestsproducts", progressRecords[2].Name.ToLower());
            Assert.AreEqual(expectedContainerName, progressRecords[2].ContainerName);
        }

        private static void AssertRestoreTriggerResults(IList<IMessage> restoreMessages, string expectedContainerName = "backup-table-rolling-weekday-saturday")
        {
            Assert.IsNotNull(restoreMessages);
            Assert.AreEqual(2, restoreMessages.Count);

            Assert.IsInstanceOf<TableRestoreMessage>(restoreMessages[0]);
            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", ((TableRestoreMessage)restoreMessages[0]).RestoreId);
            Assert.AreEqual(expectedContainerName, ((TableRestoreMessage)restoreMessages[0]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000001", ((TableRestoreMessage)restoreMessages[0]).JobId);
            Assert.AreEqual(2, ((TableRestoreMessage)restoreMessages[0]).JobDescription.MaxDegreeOfParallelism);
            StringAssert.AreEqualIgnoringCase("IntegrationBackupTestsCategories", ((TableRestoreMessage)restoreMessages[0]).JobDescription.TableName);

            Assert.IsInstanceOf<TableRestoreMessage>(restoreMessages[1]);
            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", ((TableRestoreMessage)restoreMessages[1]).RestoreId);
            Assert.AreEqual(expectedContainerName, ((TableRestoreMessage)restoreMessages[1]).BackupContainerName);
            Assert.AreEqual("00634330656053450000:000002", ((TableRestoreMessage)restoreMessages[1]).JobId);
            Assert.AreEqual(2, ((TableRestoreMessage)restoreMessages[1]).JobDescription.MaxDegreeOfParallelism);
            StringAssert.AreEqualIgnoringCase("IntegrationBackupTestsProducts", ((TableRestoreMessage)restoreMessages[1]).JobDescription.TableName);


            var progressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray();

            Assert.AreEqual(2, progressRecords.Length);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNull(progressRecords[0].Started);
            Assert.IsNull(progressRecords[0].Completed);
            Assert.IsNull(progressRecords[0].Successful);
            Assert.IsNull(progressRecords[0].FailureReason);
            StringAssert.AreEqualIgnoringCase("IntegrationBackupTestsCategories", progressRecords[0].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[0].ContainerName);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNull(progressRecords[1].Started);
            Assert.IsNull(progressRecords[1].Completed);
            Assert.IsNull(progressRecords[1].Successful);
            Assert.IsNull(progressRecords[1].FailureReason);
            StringAssert.AreEqualIgnoringCase("IntegrationBackupTestsProducts", progressRecords[1].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[1].ContainerName);
        }

        private void AssertRestoredData(int expectedProductCount, string expectedContainerName = "backup-table-rolling-weekday-saturday")
        {
            var context = SimpleContextBase.CreateInstance<IContext>();

            var products = context.Products.ToList();
            Assert.AreEqual(expectedProductCount, products.Count);
            foreach (var actualProduct in products)
            {
                Product originalProduct;
                if(!_originalProducts.TryGetValue(actualProduct.RowKey, out originalProduct))
                    Assert.Fail("Original product for {0} is not found", actualProduct.RowKey);

                Assert.AreEqual(originalProduct.PartitionKey, actualProduct.PartitionKey);
                Assert.AreEqual(originalProduct.RowKey, actualProduct.RowKey);
                Assert.AreEqual(originalProduct.Name, actualProduct.Name);
            }

            var categories = context.Categories.ToList();
            Assert.AreEqual(_originalCategories.Count, categories.Count);
            foreach (var actualCategory in categories)
            {
                Category originalCategory;
                if (!_originalCategories.TryGetValue(actualCategory.RowKey, out originalCategory))
                    Assert.Fail("Original category for {0} is not found", actualCategory.RowKey);

                Assert.AreEqual(originalCategory.PartitionKey, actualCategory.PartitionKey);
                Assert.AreEqual(originalCategory.RowKey, actualCategory.RowKey);
                Assert.AreEqual(originalCategory.Name, actualCategory.Name);
            }

            AssertCompletedRestoreProgress(expectedContainerName);
        }

        private static void AssertCompletedRestoreProgress(string expectedContainerName = "backup-table-rolling-weekday-saturday")
        {
            var progressRecords = RestoreProgressRecord.CreateContext().RestoreProgressRecords.ToArray();

            Assert.AreEqual(2, progressRecords.Length);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", progressRecords[0].PartitionKey);
            Assert.AreEqual("00634330656053450000:000001", progressRecords[0].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[0].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[0].Started);
            Assert.IsNotNull(progressRecords[0].Completed);
            Assert.IsNotNull(progressRecords[0].Successful);
            Assert.IsTrue(progressRecords[0].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[0].FailureReason);
            StringAssert.AreEqualIgnoringCase("IntegrationBackupTestsCategories", progressRecords[0].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[0].ContainerName);

            Assert.AreEqual("Table 2011-02-12T00:00:05.345Z", progressRecords[1].PartitionKey);
            Assert.AreEqual("00634330656053450000:000002", progressRecords[1].RowKey);
            Assert.AreEqual(634330656053450000, progressRecords[1].Initiated.Ticks);
            Assert.IsNotNull(progressRecords[1].Started);
            Assert.IsNotNull(progressRecords[1].Completed);
            Assert.IsNotNull(progressRecords[1].Successful);
            Assert.IsTrue(progressRecords[1].Successful.GetValueOrDefault());
            Assert.IsNull(progressRecords[1].FailureReason);
            StringAssert.AreEqualIgnoringCase("IntegrationBackupTestsProducts", progressRecords[1].Name);
            Assert.AreEqual(expectedContainerName, progressRecords[1].ContainerName);
        }

        private static void ClearAllPossibleTestData()
        {
            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            BackupTriggerRecord.CreateContext().DeleteTables();

            BackupProgressRecord.CreateContext().DeleteTables();

            RestoreProgressRecord.CreateContext().DeleteTables();

            new BlobStorageProvider().DeleteContainer(TableBackupTrigger.GetContainerName(DateTime.Parse(InitiatedTime, null, DateTimeStyles.AdjustToUniversal), BackupType.Daily, null));

            new BlobStorageProvider().DeleteContainer("backup-table-rolling-weekday-saturday");

            new BlobStorageProvider().DeleteContainer("backup-table-daily-2011-02-12");

            AzureStorageAccountFactory.Create().CreateCloudTableClient().DeleteTableIfExist("SomeTestTableToBeExcluded");
        }

        private static void RecreateBlankTables()
        {
            var context = SimpleContextBase.CreateInstance<IContext>();
            context.DeleteTables();
            context.CreateTablesIfNotExist();

            AzureStorageAccountFactory.Create().CreateCloudTableClient().CreateTableIfNotExist("SomeTestTableToBeExcluded");
        }
    }
}
