﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using System.IO;
using System.Linq;
using System.Text;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaLabs.Storage.DataServicesUtils;
using DocaLabs.Utils;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests
{
    [TestFixture]
    public class TableBackupTests
    {
        public const string BackupTestContainerName = "integrationbackuptests";
        public const string LocalBackupFolder = "local-backup";
        public const string StringSample =
                "A user object receives a pointer to another object which implements an algorithm. The first is designated 'element class' and the latter 'the visitor class'." +
                " The idea is to use a structure of element classes, each of which has an accept() method taking a visitor object for an argument. visitor is an interface having a visit()" +
                " method for each element class. The accept() method of an element class calls back the visit() method for its class. Separate concrete visitor classes can then be written" +
                " to perform some particular operations, by implementing these operations in their respective visit() methods." +
                "One of these visit() methods of a concrete visitor can be thought of as a method not of a single class, but rather a method of a pair of classes: the concrete visitor" +
                " and the particular element class. Thus the visitor pattern simulates double dispatch in a conventional single-dispatch object-oriented language such as Java, Smalltalk," +
                " and C++. For an explanation of how double dispatch differs from function overloading, see Double dispatch is more than function overloading in the double dispatch article." +
                " In the Java language, two techniques have been documented that use reflection to simplify the mechanics of double dispatch simulation in the visitor pattern: getting" +
                " rid of accept() methods (the Walkabout variation), and getting rid of extra visit() methods." +
                "The visitor pattern also specifies how iteration occurs over the object structure. In the simplest version, where each algorithm needs to iterate in the same way," +
                " the accept() method of a container element, in addition to calling back the visit() method of the visitor, also passes the visitor object to the accept() method of all" +
                " its constituent child elements." +
                "Because the visitor object has one principal function (manifested in a plurality of specialized methods) and that function is called visit(), the visitor can be readily" +
                " identified as a potential function object. Likewise, the accept() function can be identified as a function applicator, a mapper, which knows how to traverse a particular" +
                " type of object and apply a function to its elements. Lisp's object system with its multiple dispatch does not replace the Visitor pattern, but merely provides a more" +
                " concise implementation of it in which the pattern all but disappears.";

        [Table(TableName)]
        public class Product : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProducts";

            public Product()
            {
            }

            public Product(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public string CustomerId { get; set; }

            public string Name { get; set; }

            public double PricePerUnit { get; set; }

            public int Quantity { get; set; }

            public string NullOrNotNull { get; set; }

            public byte[] LongData { get; set; }

            public byte[] P01 { get; set; }
            public byte[] P02 { get; set; }
            public byte[] P03 { get; set; }
            public byte[] P04 { get; set; }
            public byte[] P05 { get; set; }
            public byte[] P06 { get; set; }
        }

        public interface IContext : ISimpleContextBase
        {
            IDbSet<Product> Products { get; set; }
        }

        public Guid[] ProductCategories { get; set; }
        public Guid[] Customers { get; set; }
        public Dictionary<Guid, Product> OriginalProducts { get; set; }

        private Type _originalBufferType;

        [SetUp]
        public void Setup()
        {
            _originalBufferType = DefaultBackupServiceFactory.BufferType;
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            if (Directory.Exists(LocalBackupFolder))
                Directory.Delete(LocalBackupFolder, true);
        }

        [TearDown]
        public void TearDown()
        {
            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            new BlobStorageProvider().DeleteContainer(BackupTestContainerName);

            if (Directory.Exists(LocalBackupFolder))
                Directory.Delete(LocalBackupFolder, true);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;

            DefaultBackupServiceFactory.BufferType = _originalBufferType;
        }

        [Test]
        [Category("LongRunninTest")]
        public void FullBackupAndRestoreUsingSharpZlib()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(5500, 1984, 6);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new []
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();
            
            AssertBlockNumberInBlobs();

            RecreateBlankTables();

            new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);

            AssertProducts(OriginalProducts.Count);
        }

        [Test]
        [Category("LongRunninTest")]
        public void FullBackupAndRestoreUsingLocalFileSystem()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(5500, 1984, 6);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(
                blobNameProvider, Product.TableName, null, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                {
                    MaxBlocksInBlobs = 10, 
                    CleanHeapAfterEachBlock = true
                }.Backup(ranges, 4);

            RecreateBlankTables();

            new TableRestore(
                blobNameProvider, Product.TableName, new LocalFileSystemBackupStorageFactory(LocalBackupFolder))
                {
                    CleanHeapAfterEachBlock = true
                }.Restore(4);

            AssertProducts(OriginalProducts.Count);
        }

        [Test]
        public void ShortBackupAndRestoreUsingSharpZlib()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(50, 25, 6);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            AssertBlockNumberInBlobs();

            RecreateBlankTables();

            new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);

            AssertProducts(OriginalProducts.Count);
        }

        [Test]
        public void ShortBackupAndRestoreUsingStandardDeflate()
        {
            DefaultBackupServiceFactory.BufferType = typeof(StandardDeflateBuffer);

            Arrange(50, 25, 20);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            AssertBlockNumberInBlobs();

            RecreateBlankTables();

            new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);

            AssertProducts(OriginalProducts.Count);
        }

        [Test]
        public void ShortBackupAndRestoreWithoutCompression()
        {
            DefaultBackupServiceFactory.BufferType = typeof(UncompressedBuffer);

            Arrange(50, 25, 15);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            AssertBlockNumberInBlobs();

            RecreateBlankTables();

            new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);

            AssertProducts(OriginalProducts.Count);
        }

        [Test]
        public void RestoreCreatesTableWhenItDoesnotExist()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(1, 1, 0);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            AssertBlockNumberInBlobs();

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            try
            {
                new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);
            }
            catch (Exception e)
            {
                Console.WriteLine(@"Ignore exception as it most likely development fabric problem");
                Console.WriteLine(e);
            }

            var cloudClient = AzureStorageAccountFactory.Create().CreateCloudTableClient();
            Assert.IsTrue(cloudClient.DoesTableExist(Product.TableName));
        }

        [Test]
        public void BackupsAndRestoreSingleEntityOnly()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(1, 0, 0);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            AssertBlockNumberInBlobs();

            RecreateBlankTables();

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            try
            {
                new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);
            }
            catch (Exception e)
            {
                Console.WriteLine(@"Ignore exception as it most likely development fabric problem");
                Console.WriteLine(e);
            }

            var cloudClient = AzureStorageAccountFactory.Create().CreateCloudTableClient();
            Assert.IsTrue(cloudClient.DoesTableExist(Product.TableName));

            AssertProducts(OriginalProducts.Count);
        }

        [Test]
        public void NothingHappensWhenThereIsNoData()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(0, 0, 0);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            var account = DefaultBackupServiceFactory.Default.GetStorageAccount();
            Assert.AreEqual(0, new BlobStorageProvider(account).FlatBlobList(BackupTestContainerName).Count());

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            try
            {
                new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);
            }
            catch (Exception e)
            {
                Console.WriteLine(@"Ignore exception as it most likely development fabric problem");
                Console.WriteLine(e);
            }

            var cloudClient = AzureStorageAccountFactory.Create().CreateCloudTableClient();
            Assert.IsFalse(cloudClient.DoesTableExist(Product.TableName));
        }

        [Test]
        public void NothingHappensWhenThereIsNoTargetTable()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(0, 0, 0);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, "someunknowntesttable") { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            var account = DefaultBackupServiceFactory.Default.GetStorageAccount();
            Assert.AreEqual(0, new BlobStorageProvider(account).FlatBlobList(BackupTestContainerName).Count());

            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            try
            {
                new TableRestore(blobNameProvider, "someunknowntesttable") { CleanHeapAfterEachBlock = true }.Restore(4);
            }
            catch (Exception e)
            {
                Console.WriteLine(@"Ignore exception as it most likely development fabric problem");
                Console.WriteLine(e);
            }

            var cloudClient = AzureStorageAccountFactory.Create().CreateCloudTableClient();
            Assert.IsFalse(cloudClient.DoesTableExist(Product.TableName));
        }

        [Test]
        public void BackupAndRestoreIntoSomeAlreadyExisitngEntities()
        {
            DefaultBackupServiceFactory.BufferType = typeof(SharpZipLibBuffer);

            Arrange(50, 25, 6);

            var blobNameProvider = new TableBackupBlobNameProvider(BackupTestContainerName);

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            new TableBackup(blobNameProvider, Product.TableName) { MaxBlocksInBlobs = 10, CleanHeapAfterEachBlock = true }.Backup(ranges, 4);

            PrintBlobInfo();

            AssertBlockNumberInBlobs();

            DeleteOnlySomeProducts();

            new TableRestore(blobNameProvider, Product.TableName) { CleanHeapAfterEachBlock = true }.Restore(4);

            AssertProducts(OriginalProducts.Count);
        }

        private void Arrange(int entitiesPerShortCategories, int entitiesPerLongCategories, int extraLargeDataCount)
        {
            CreateOriginalData(entitiesPerShortCategories, entitiesPerLongCategories, extraLargeDataCount);

            new BlobStorageProvider().DeleteContainer(BackupTestContainerName);

            RecreateBlankTables();

            PopulateTables();
        }

        private void CreateOriginalData(int entitiesPerShortCategories, int entitiesPerLongCategories, int extraLargeDataCount)
        {
            ProductCategories = new[]
            {
                new Guid("56fc6a50-d937-433f-a3d9-cd566aefdca1"),
                new Guid("1b5639d0-d250-4d5f-b955-8b1273dddd20"), 
                new Guid("fc87bb62-32c9-4279-b7c0-2d09d1a5c0e3"), 
                new Guid("3f8b3c6e-0a3b-4219-8830-410b0d2a0844") 
            };

            Customers = new[]
            {
                new Guid("ae1ae27f-f26b-41cc-a657-c7547a62188f"), 
                new Guid("83150a70-1669-477a-ac6b-63270ac6e0fe") 
            };

            OriginalProducts = new Dictionary<Guid, Product>();

            var random = new Random(42);

            var count = 0;

            // first three categories (without long data)
            for (var idx = 0; idx < (ProductCategories.Length - 1); idx++)
            {
                var productCategory = ProductCategories[idx];

                for (var i = 0; i < entitiesPerShortCategories; ++i)
                {
                    ++count;

                    var productId = Guid.NewGuid();

                    var longData = new byte[random.Next(327, 959)];
                    random.NextBytes(longData);

                    var product = new Product(productCategory.ToString(), productId.ToString())
                                      {
                                          CustomerId = Customers[productCategory.GetHashCode() % 2].ToString(),
                                          Name = "Product " + productId + " generated at " + DefaultClock.CurrentTime,
                                          PricePerUnit = random.Next(30, 99),
                                          Quantity = random.Next(1, 20),
                                          NullOrNotNull = i % 2 == 0 ? GenerateRandomString(random) : null,
                                          LongData = longData
                                      };

                    OriginalProducts.Add(productId, product);
                }

                if(entitiesPerShortCategories == 1 && entitiesPerLongCategories == 0 && extraLargeDataCount == 0)
                    break;
            }
            /**/

            // the last category (with long data)
            var lastCategory = ProductCategories[ProductCategories.Length - 1];

            for (var i = 0; i < entitiesPerLongCategories; ++i)
            {
                ++count;

                var productId = Guid.NewGuid();

                var longData = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData);

                var product = new Product(lastCategory.ToString(), productId.ToString())
                {
                    CustomerId = Customers[lastCategory.GetHashCode() % 2].ToString(),
                    Name = "Product " + productId + " generated at " + DefaultClock.CurrentTime,
                    PricePerUnit = random.Next(30, 99),
                    Quantity = random.Next(1, 20),
                    NullOrNotNull = i % 2 == 0 ? GenerateRandomString(random) : null,
                    LongData = longData
                };

                OriginalProducts.Add(productId, product);
            }
            /**/

            // extremely long data
            var extraLargeCategory = ProductCategories[1];

            for (var i = 0; i < extraLargeDataCount; ++i)
            {
                ++count;

                var productId = Guid.NewGuid();

                var longData0 = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData0);

                var longData1 = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData1);

                var longData2 = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData2);

                var longData3 = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData3);

                var longData4 = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData4);

                var longData5 = new byte[random.Next(63000, 65000)];
                random.NextBytes(longData5);

                var longData6 = new byte[random.Next(1000, 2000)];
                random.NextBytes(longData6);

                var product = new Product(extraLargeCategory.ToString(), productId.ToString())
                {
                    CustomerId = Customers[0].ToString(),
                    Name = "Product " + productId + " generated at " + DefaultClock.CurrentTime,
                    PricePerUnit = random.Next(30, 99),
                    Quantity = random.Next(1, 20),
                    NullOrNotNull = i % 2 == 0 ? GenerateRandomString(random) : null,
                    LongData = longData0,
                    P01 = longData1,
                    P02 = longData2,
                    P03 = longData3,
                    P04 = longData4,
                    P05 = longData5,
                    P06 = longData6
                };

                OriginalProducts.Add(productId, product);
            }
            /**/

            Console.WriteLine(@"Original data is created, {0} entities", count);
        }

        private static string GenerateRandomString(Random random)
        {
            var builder = new StringBuilder("some string", 1024);

            var size = random.Next(10, 20);
            for (var i = 0; i < size; i++)
                builder.Append(random.Next());

            var position = random.Next(10, StringSample.Length - 10);
            var length = random.Next(5, StringSample.Length - position);

            builder.Append(StringSample.Substring(position, length));

            return builder.ToString();
        }

        private void PopulateTables()
        {
            var count = 0;
            var batchCount = 0;

            var tmpContext = SimpleContextBase.CreateInstance<IContext>();

            string lastPartitionKey = null;

            // small and medium sized entities
            foreach (var product in OriginalProducts.Values)
            {
                if(product.P01 != null)
                    continue;

                if (lastPartitionKey == null)
                    lastPartitionKey = product.PartitionKey;

                if (batchCount == 40 || lastPartitionKey != product.PartitionKey)
                {
                    batchCount = 0;
                    lastPartitionKey = product.PartitionKey;
                    tmpContext.SaveBatchWithRetries();
                    Console.WriteLine(@"Saved {0} entities", count);
                    tmpContext = SimpleContextBase.CreateInstance<IContext>();
                }

                ++count;
                ++batchCount;

                tmpContext.Products.Add(product);
            }
            tmpContext.SaveBatchWithRetries();

            // large entities
            tmpContext = SimpleContextBase.CreateInstance<IContext>();

            foreach (var product in OriginalProducts.Values)
            {
                if (product.P01 == null)
                    continue;
                ++count;
                tmpContext.Products.Add(product);
            }
            tmpContext.SaveChanges();

            Console.WriteLine(@"Saved {0} entities", count);
        }

        private static void PrintBlobInfo()
        {
            var account = DefaultBackupServiceFactory.Default.GetStorageAccount();

            var blobs = new BlobStorageProvider(account).FlatBlobList(BackupTestContainerName).ToArray();

            Array.Sort(blobs, StringComparer.InvariantCulture);

			Console.WriteLine();
            Console.WriteLine(@"Blob number: {0}", blobs.Length);

			var dataSize = 0L;
			
            foreach (var blobName in blobs)
            {
                var blob = account.CreateCloudBlobClient().GetBlockBlobReference(blobName);

                var blocks = blob.DownloadBlockList(BlockListingFilter.Committed).ToArray();

				var blobSize = blocks.Sum(block => block.Size);

                dataSize += blobSize;
				
                Console.WriteLine(@"Blob: {0}", blobName);
                Console.WriteLine(@"Blocks {0}, size {1}", blocks.Length, blobSize);
            }
			
			Console.WriteLine();
			Console.WriteLine(@"Data size: {0}", dataSize);
			Console.WriteLine();
        }

        private void DeleteOnlySomeProducts()
        {
            var deletedCount = 0;
            var tmpContext = SimpleContextBase.CreateInstance<IContext>();
            var delete = false;
            foreach (var product in tmpContext.Products.Where(x => x.PartitionKey == ProductCategories[0].ToString()))
            {
                if (delete)
                {
                    deletedCount++;
                    tmpContext.Products.Remove(product);
                }
                delete = !delete;
            }
            tmpContext.SaveBatchWithRetries();

            foreach (var product in tmpContext.Products.Where(x => x.PartitionKey == ProductCategories[1].ToString()))
            {
                if (delete)
                {
                    deletedCount++;
                    tmpContext.Products.Remove(product);
                }
                delete = !delete;
            }

            Console.WriteLine(@"Deleted {0} out of {1}", deletedCount, OriginalProducts.Count);
        }

        private static void AssertBlockNumberInBlobs()
        {
            var account = DefaultBackupServiceFactory.Default.GetStorageAccount();

            var blobs = new BlobStorageProvider(account).FlatBlobList(BackupTestContainerName).ToArray();

            Array.Sort(blobs, StringComparer.InvariantCulture);

            foreach (var blobName in blobs)
            {
                var blob = account.CreateCloudBlobClient().GetBlockBlobReference(blobName);

                var blocks = blob.DownloadBlockList(BlockListingFilter.Committed).ToArray();

                Assert.LessOrEqual(blocks.Length, 10);
            }
        }

        private void AssertProducts(int expectedCount)
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var count = 0;

            foreach (var product in newContext.Products)
            {
                AssertProduct(product);
                ++count;
            }

            Console.WriteLine(@"Read {0} entities", count);

            Assert.AreEqual(expectedCount, count, "number of restored products is not equal to the original");
        }

        private void AssertProduct(Product actual)
        {
            var expected = OriginalProducts[Guid.Parse(actual.RowKey)];

            Assert.IsNotNull(expected, "couldn't find the original product");

            Assert.AreEqual(expected.PartitionKey, actual.PartitionKey);
            Assert.AreEqual(expected.CustomerId, actual.CustomerId);
            Assert.AreEqual(expected.Name, actual.Name);
            Assert.AreEqual(expected.PricePerUnit, actual.PricePerUnit, double.Epsilon);
            Assert.AreEqual(expected.Quantity, actual.Quantity);
            Assert.AreEqual(expected.NullOrNotNull, actual.NullOrNotNull);
            CollectionAssert.AreEqual(expected.LongData, actual.LongData);
        }
    
        private static void RecreateBlankTables()
        {
            var context = SimpleContextBase.CreateInstance<IContext>();
            context.DeleteTables();
            context.CreateTablesIfNotExist();
        }
    }
}
