﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaLabs.Storage.DataServicesUtils;
using DocaLabs.Utils;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests
{
    [TestFixture]
    public class TableTransformationTests
    {
        [Table(TableName)]
        public class Product : SimpleEntity
        {
            public const string TableName = "IntegrationTransformationTestsProducts";

            public Product()
            {
            }

            public Product(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public Guid CustomerId { get; set; }

            public string Name { get; set; }

            public double PricePerUnit { get; set; }

            public int Quantity { get; set; }

            public byte[] Data { get; set; }
        }

        [Table(Product.TableName)]
        public class ProductV2 : SimpleEntity
        {
            public ProductV2()
            {
            }

            public ProductV2(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public Guid CustomerId { get; set; }

            public string Name { get; set; }

            public double PricePerUnit { get; set; }

            public int Quantity { get; set; }

            public string NameAndCustomerId { get; set; }

            public double Sum { get; set; }

            public byte[] Data { get; set; }
        }

        public interface IContext : ISimpleContextBase
        {
            IDbSet<Product> Products { get; set; }
        }

        public interface IContextV2 : ISimpleContextBase
        {
            IDbSet<ProductV2> Products { get; set; }
        }


        private const int EntityCountPerCategory = 537;

        private Dictionary<Guid, Product> OriginalProducts { get; set; }

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void TearDown()
        {
            SimpleContextBase.CreateInstance<IContext>().DeleteTables();
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void Test()
        {
            // have three attempts
            for (var i = 0; i < 3;++i)
            {
                try
                {
                    PopulateInitialSample();

                    var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "1", "2", "3", "4", "5", "6", "7", "8", "9", "a", "b", "c", "d", "e", "f"
                    });

                    var target = new TableTransformation(Product.TableName, UpdateAction);

                    target.Transform(ranges, 4);

                    AssertProducts();

                    return;
                }
                catch (Exception e)
                {
                    // the test is highly intermittent as it adds new propertis and the dev fabric behaves irratically as the result
                    // it will bi fine if run against azure storage account
                    if (e.ExtractAtsError() != AtsError.InvalidInput)
                        throw;

                    Console.WriteLine();
                    Console.WriteLine();
                    Console.WriteLine(@" ----------------- Restarting the test -----------------");
                    Console.WriteLine(@"as it failed most likely due the known problem in dev fabric");
                    Console.WriteLine();

                    continue;
                }
            }
        }

        private static void UpdateAction(BackupEntity entity)
        {
            var accessor = entity.GetAtomAccessor();

            var customerId = accessor.GetValue((Product x) => x.CustomerId);
            var name = accessor.GetValue((Product x) => x.Name);
            var quantity = accessor.GetValue((Product x) => x.Quantity);
            var pricePerUnit = accessor.GetValue((Product x) => x.PricePerUnit);

            accessor.Add((ProductV2 x) => x.NameAndCustomerId, name + " -> " + customerId);
            accessor.Add((ProductV2 x) => x.Sum, quantity * pricePerUnit);

            accessor.Update((Product x) => x.Name, "see in NameAndCustomerId");
        }

        private void PopulateInitialSample(bool initializeData = false)
        {
            PopulateTables(CreateOriginalData(initializeData), initializeData);
        }

        private IEnumerable<Product> CreateOriginalData(bool initializeData)
        {
            OriginalProducts = new Dictionary<Guid, Product>();

            var listProducts = new List<Product>();

            var random = new Random(42);

            var categories = new[]
            {
                "56fc6a50-d937-433f-a3d9-cd566aefdca1",
                "1b5639d0-d250-4d5f-b955-8b1273dddd20", 
                "fc87bb62-32c9-4279-b7c0-2d09d1a5c0e3", 
                "3f8b3c6e-0a3b-4219-8830-410b0d2a0844", 
            };

            foreach (var category in categories)
            {
                for (var i = 0; i < EntityCountPerCategory; ++i)
                {
                    var productId = Guid.NewGuid();

                    var product = new Product(category, productId.ToString())
                    {
                        CustomerId = Guid.NewGuid(),
                        Name = "Product " + productId + " generated at " + DefaultClock.CurrentTime,
                        PricePerUnit = random.Next(30, 99),
                        Quantity = random.Next(1, 20),
                    };

                    if (initializeData)
                    {
                        product.Data = new byte[random.Next(327, 959)];
                        random.NextBytes(product.Data);
                    }

                    OriginalProducts.Add(productId, product);
                    listProducts.Add(product);
                }
            }

            Console.WriteLine(@"Original data is created, {0} entities", listProducts.Count);

            return listProducts;
        }

        private static void PopulateTables(IEnumerable<Product> products, bool initializeData)
        {
            var context = SimpleContextBase.CreateInstance<IContext>();

            // recreate tables
            context.DeleteTables();
            context.CreateTablesIfNotExist();

            string lastPartitionKey = null;
            var count = 0;
            var batchCount = 0;
            var maxBatchSize = 100;

            if (initializeData)
                maxBatchSize = 40;

            foreach (var product in products)
            {
                if (lastPartitionKey == null)
                    lastPartitionKey = product.PartitionKey;

                if (batchCount == maxBatchSize || lastPartitionKey != product.PartitionKey)
                {
                    batchCount = 0;
                    lastPartitionKey = product.PartitionKey;
                    context.SaveBatchWithRetries();
                    Console.WriteLine(@"Saved {0} entities", count);
                    context = SimpleContextBase.CreateInstance<IContext>();
                }

                ++count;
                ++batchCount;

                context.Products.Add(product);
            }

            context.SaveBatchWithRetries();

            Console.WriteLine(@"Saved {0} entities", count);
        }

        private void AssertProducts()
        {
            var newContext = SimpleContextBase.CreateInstance<IContextV2>();

            var count = 0;

            foreach (var product in newContext.Products)
            {
                AssertProduct(product);
                ++count;
            }

            Console.WriteLine();
            Console.WriteLine(@"---- Asserted {0} entities", count);

            Assert.AreEqual(OriginalProducts.Count, count, "number of restored products is not equal to the original");
        }

        private void AssertProduct(ProductV2 actual)
        {
            var expected = OriginalProducts[Guid.Parse(actual.RowKey)];

            Assert.IsNotNull(expected, "couldn't find the original product");

            Assert.AreEqual(expected.PartitionKey, actual.PartitionKey);
            Assert.AreEqual(expected.CustomerId, actual.CustomerId);
            Assert.AreEqual(expected.PricePerUnit, actual.PricePerUnit, double.Epsilon);
            Assert.AreEqual(expected.Quantity, actual.Quantity);
            CollectionAssert.AreEqual(expected.Data, actual.Data);

            // annd and changed properties
            Assert.AreEqual("see in NameAndCustomerId", actual.Name);
            StringAssert.EndsWith(" -> " + expected.CustomerId, actual.NameAndCustomerId);
            Assert.AreEqual(expected.Quantity * expected.PricePerUnit, actual.Sum, double.Epsilon);
        }
    }
}
