﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaLabs.Storage.DataServicesUtils;
using NUnit.Framework;

namespace DocaAzure.Backup.IntegrationTests
{
    [TestFixture]
    public class TransformationTableGraphBuilderTests
    {
        #region tests data

        private Guid _customer1Id = Guid.Parse("011dc137-9406-4242-8fad-3f9b667fa4c2");
        private Guid _customer2Id = Guid.Parse("a719ce58-20e7-49f2-a938-b86a0e8af673");

        private Guid _categoryElectronicsId = Guid.Parse("137619d7-1cff-4130-a62a-d69c52e1fd26");
        private Guid _categoryFoodId = Guid.Parse("159ea8bb-6700-4d6e-817f-2681cdbf9ee3");
        private Guid _categorySportId = Guid.Parse("945aed92-9e87-4937-a1cb-41d2e0690e6b");


        [Table(TableName)]
        public class Customer : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsCustomers";

            public Customer()
            {
            }

            public Customer(string customerId)
                : base(customerId, string.Empty)
            {
            }

            public string Name { get; set; }

            public string Address { get; set; }
        }

        public class CustomerV2 : Customer
        {
            public string NewName { get; set; }
        }

        [Table(TableName)]
        public class CustomerUser : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsCustomerUsers";

            public CustomerUser()
            {
            }

            public CustomerUser(string customerId, string userId)
                : base(customerId, userId)
            {
            }

            public string Name { get; set; }
        }

        public class CustomerUserV2 : CustomerUser
        {
            public string NewName { get; set; }    
        }

        [Table(TableName)]
        public class ProductCategory : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProductCategories";

            public ProductCategory()
            {
            }

            public ProductCategory(string customerId, string categoryId)
                : base(customerId, categoryId)
            {
            }

            public string Name { get; set; }
        }

        public class ProductCategoryV2 : ProductCategory
        {
            public string NewName { get; set; }
        }

        [Table(TableName)]
        public class Product : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProducts";

            public Product()
            {
            }

            public Product(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public string CustomerId { get; set; }

            public string Name { get; set; }

            public double PricePerUnit { get; set; }

            public int Quantity { get; set; }
        }

        public class ProductV2 : Product
        {
            public string NewName { get; set; }
        }

        [Table(TableName)]
        public class ProductCategoryColour : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProductCategoruColours";

            public ProductCategoryColour()
            {
            }

            public ProductCategoryColour(string categoryId, string rowId)
                : base(categoryId, rowId)
            {
            }

            public string CustomerId { get; set; }

            public string Colour { get; set; }
        }

        public class ProductCategoryColourV2 : ProductCategoryColour
        {
            public string NewName { get; set; }
        }

        public interface IContext : ISimpleContextBase
        {
            IDbSet<Customer> Customers { get; set; }
            IDbSet<CustomerUser> CustomerUsers { get; set; }
            IDbSet<ProductCategory> ProductCategories { get; set; }
            IDbSet<ProductCategoryColour> ProductCategoryColours { get; set; }
            IDbSet<Product> Products { get; set; }
        }

        public interface IContextV2 : ISimpleContextBase
        {
            IDbSet<CustomerV2> Customers { get; set; }
            IDbSet<CustomerUserV2> CustomerUsers { get; set; }
            IDbSet<ProductCategoryV2> ProductCategories { get; set; }
            IDbSet<ProductCategoryColourV2> ProductCategoryColours { get; set; }
            IDbSet<ProductV2> Products { get; set; }
        }

        public Customer[] OriginalCustomers { get; set; }
        public CustomerUser[] OriginalCustomerUsers { get; set; }
        public ProductCategory[] OriginalProductCategories { get; set; }
        public Product[] OriginalProducts { get; set; }
        public ProductCategoryColour[] OriginalProductCategoryColours { get; set; }
        public IContext Context { get; set; }

        #endregion test data

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void TearDown()
        {
            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void BuildAndTransformTest()
        {
            Arrange();

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            var graph = TransformationTableGraphBuilder.CreateRoot(Customer.TableName, AddNewNameActionFromName)
                            .ChildByPartitionKey(CustomerUser.TableName, AddNewNameActionFromName)
                            .SiblingByPartitionKey(ProductCategory.TableName, AddNewNameActionFromName)
                                .ChildByRowKey(Product.TableName, AddNewNameActionFromName)
                            .BackupRoot;

            graph.Transform(ranges);

            var context = SimpleContextBase.CreateInstance<IContextV2>();

            foreach (var customer in context.Customers)
            {
                Assert.AreEqual(customer.Name + ">$$<" + customer.Name, customer.NewName);
            }

            foreach (var user in context.CustomerUsers)
            {
                Assert.AreEqual(user.Name + ">$$<" + user.Name, user.NewName);
            }

            foreach (var category in context.ProductCategories)
            {
                Assert.AreEqual(category.Name + ">$$<" + category.Name, category.NewName);
            }

            foreach (var product in context.Products)
            {
                Assert.AreEqual(product.Name + ">$$<" + product.Name, product.NewName);
            }
        }

        [Test]
        public void BuildAndTransformTestForSiblingsByRowKey()
        {
            Arrange();

            var ranges = PartitionKeyRange.FromKeyRange(new[]
                    {
                        "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    });

            var graph = TransformationTableGraphBuilder.CreateRoot(ProductCategory.TableName, AddNewNameActionFromName)
                            .ChildByRowKey(ProductCategoryColour.TableName, AddNewNameActionForProductCategoryColour)
                            .SiblingByRowKey(Product.TableName, AddNewNameActionFromName)
                            .BackupRoot;

            graph.Transform(ranges);

            var context = SimpleContextBase.CreateInstance<IContextV2>();

            foreach (var category in context.ProductCategories)
            {
                Assert.AreEqual(category.Name + ">$$<" + category.Name, category.NewName);
            }

            foreach (var colour in context.ProductCategoryColours)
            {
                Assert.AreEqual(colour.Colour + ">$$<" + colour.Colour, colour.NewName);
            }

            foreach (var product in context.Products)
            {
                Assert.AreEqual(product.Name + ">$$<" + product.Name, product.NewName);
            }
        }

        private static void AddNewNameActionFromName(BackupEntity entity)
        {
            var accessor = entity.GetAtomAccessor();

            var exisitingName = accessor.AsString("Name");

            if (accessor.Add("NewName", exisitingName + ">$$<" + exisitingName))
                return;

            // ....f... fabric can add non exixting property event if it's not defined on the entity
            if(TableStorageUtils.TryGuessIsItDevFabric())
            {
                accessor.Update("NewName", exisitingName + ">$$<" + exisitingName);
            }
        }

        private static void AddNewNameActionForProductCategoryColour(BackupEntity entity)
        {
            var accessor = entity.GetAtomAccessor();

            var colour = accessor.AsString("Colour");

            accessor.Add("NewName", colour + ">$$<" + colour);
        }

        private void Arrange()
        {
            CreateOriginalData();

            Context = SimpleContextBase.CreateInstance<IContext>();

            Context.DeleteTables();

            Context.CreateTablesIfNotExist();

            PopulateTables();
        }

        private void CreateOriginalData()
        {
            OriginalCustomers = new[]
                                    {
                                        new Customer(_customer1Id.ToString()) {Name = "Golden Standard", Address = "Somewhere in Ireland"},
                                        new Customer(_customer2Id.ToString()) {Name = "Silver Standard", Address = "Somewhere in UK"}
                                    };

            OriginalCustomerUsers = new[]
                                        {
                                            new CustomerUser(OriginalCustomers[0].PartitionKey, Guid.NewGuid().ToString()) {Name = "John Smith"}, 
                                            new CustomerUser(OriginalCustomers[1].PartitionKey, Guid.NewGuid().ToString()) {Name = "Peter Black"}, 
                                        };

            OriginalProductCategories = new[]
                                            {
                                                new ProductCategory(OriginalCustomers[0].PartitionKey, _categoryElectronicsId.ToString())
                                                    {
                                                        Name = "Electronics"
                                                    }, 
                                                new ProductCategory(OriginalCustomers[0].PartitionKey, _categoryFoodId.ToString())
                                                    {
                                                        Name = "Food"
                                                    },
 
                                                new ProductCategory(OriginalCustomers[1].PartitionKey, _categorySportId.ToString())
                                                    {
                                                        Name = "Sport"
                                                    } 
                                            };

            OriginalProductCategoryColours = new[]
                                         {
                                             new ProductCategoryColour(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                                                {
                                                    CustomerId = OriginalCustomers[0].PartitionKey,
                                                    Colour = "Red"
                                                }, 
                                             new ProductCategoryColour(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                                                {
                                                    CustomerId = OriginalCustomers[0].PartitionKey,
                                                    Colour = "Green"
                                                }, 
                                         };

            OriginalProducts = new[]
                                   {
                                       new Product(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                                           {
                                               CustomerId = OriginalCustomers[0].PartitionKey,
                                               Name = "Fake iPod",
                                               PricePerUnit = 89.99,
                                               Quantity = 10
                                           }, 
                                       new Product(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                                           {
                                               CustomerId = OriginalCustomers[0].PartitionKey,
                                               Name = "Fake EOS 1Ds Mark I",
                                               PricePerUnit = 5589.99,
                                               Quantity = 1
                                           }, 
                                       new Product(OriginalProductCategories[0].RowKey, Guid.NewGuid().ToString())
                                           {
                                               CustomerId = OriginalCustomers[0].PartitionKey,
                                               Name = "Fake HTC Desire",
                                               PricePerUnit = 289.99,
                                               Quantity = 5
                                           }, 

                                       new Product(OriginalProductCategories[1].RowKey, Guid.NewGuid().ToString())
                                           {
                                               CustomerId = OriginalCustomers[0].PartitionKey,
                                               Name = "Ice cream",
                                               PricePerUnit = 1.22,
                                               Quantity = 50
                                           }, 

                                       new Product(OriginalProductCategories[2].RowKey, Guid.NewGuid().ToString())
                                           {
                                               CustomerId = OriginalCustomers[1].PartitionKey,
                                               Name = "Tennis Racket",
                                               PricePerUnit = 289.99,
                                               Quantity = 2
                                           }, 
                                       new Product(OriginalProductCategories[2].RowKey, Guid.NewGuid().ToString())
                                           {
                                               CustomerId = OriginalCustomers[1].PartitionKey,
                                               Name = "Tennis Shoes",
                                               PricePerUnit = 59.99,
                                               Quantity = 7
                                           }, 
                                   };
        }

        private void PopulateTables()
        {
            foreach (var customer in OriginalCustomers)
            {
                Context.Customers.Add(customer);
            }

            foreach (var user in OriginalCustomerUsers)
            {
                Context.CustomerUsers.Add(user);
            }

            foreach (var category in OriginalProductCategories)
            {
                Context.ProductCategories.Add(category);
            }

            foreach (var colour in OriginalProductCategoryColours)
            {
                Context.ProductCategoryColours.Add(colour);
            }

            foreach (var product in OriginalProducts)
            {
                Context.Products.Add(product);
            }

            Context.SaveChanges();
        }
    }
}
