﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using DocaAzure.Messaging.Core;
using DocaAzure.Messaging.Providers;
using DocaAzure.Scheduler;
using DocaAzure.Storage;
using DocaAzure.TestUtils;
using DocaLabs.Utils;
using NUnit.Framework;

namespace DocaAzure.Messaging.Config.IntegrationTests
{
    [TestFixture]
    public class BusServiceConfigurationTests
    {
        [SetUpFixture]
        public class FixtureSetup
        {
            [SetUp]
            public void Setup()
            {
                LoggingUtils.SetupConsoleLogging();

                AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
                DefaultConfigurationManager.ReplaceDefaultBy("DocaAzure.Messaging.Config.IntegrationTests.config");
            }

            [TearDown]
            public void TearDown()
            {
                DefaultConfigurationManager.Default = null;
                AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
            }
        }

        [Test]
        public void CreatesQueuesFromConfiguration()
        {
            var service = new BusService();

            service.ConfigureQueuesFromFile();

            Assert.AreEqual(3, service.QueueServices.Count);

            foreach (var queueService in service.QueueServices)
            {
                Assert.IsTrue(queueService.PurgeOnStartup);

                switch (queueService.QueueName)
                {
                    case "queue10":
                        Assert.AreEqual(1011, queueService.MaxDequeueCount);
                        Assert.AreEqual(1022, queueService.ConcurrentProcessingThrottle);
                        Assert.AreEqual("error10", queueService.ErrorQueueName);
                        Assert.AreEqual(1066L, (long)((AzureQueueTransport)queueService.Transport).MessageInvisibleTime.TotalMilliseconds);
                        Assert.IsInstanceOf<RegularQueueSchedule>((queueService).QueueSchedule);
                        Assert.AreEqual(103L, (long)((RegularQueueSchedule)queueService.QueueSchedule).InitialPollInterval.TotalMilliseconds);
                        Assert.AreEqual(1044L, (long)((RegularQueueSchedule)queueService.QueueSchedule).PollIntervalIncrement.TotalMilliseconds);
                        Assert.AreEqual(1055L, (long)((RegularQueueSchedule)queueService.QueueSchedule).MaximumWaitTimeWhenIdle.TotalMilliseconds);
                        break;
                    case "queue20":
                        Assert.AreEqual(2011, queueService.MaxDequeueCount);
                        Assert.AreEqual(2022, queueService.ConcurrentProcessingThrottle);
                        Assert.AreEqual("error20", queueService.ErrorQueueName);
                        Assert.AreEqual(206L, (long)((AzureQueueTransport)queueService.Transport).MessageInvisibleTime.TotalMilliseconds);
                        Assert.IsInstanceOf<DailySchedule>((queueService).QueueSchedule);
                        Assert.AreEqual(2, ((DailySchedule)(queueService).QueueSchedule).Interval);
                        Assert.AreEqual(2, ((DailySchedule)(queueService).QueueSchedule).Hour);
                        Assert.AreEqual(2, ((DailySchedule)(queueService).QueueSchedule).Guard);
                        Assert.AreEqual(WeekDaysToSkip.SkipSaturday | WeekDaysToSkip.SkipSunday, ((DailySchedule)(queueService).QueueSchedule).WeekDaysToSkip);
                        break;
                    case "queue30":
                        Assert.AreEqual(3011, queueService.MaxDequeueCount);
                        Assert.AreEqual(3022, queueService.ConcurrentProcessingThrottle);
                        Assert.AreEqual("error30", queueService.ErrorQueueName);
                        Assert.AreEqual(3066L, (long)((AzureQueueTransport)queueService.Transport).MessageInvisibleTime.TotalMilliseconds);
                        Assert.IsInstanceOf<HourlySchedule>((queueService).QueueSchedule);
                        Assert.AreEqual(3, ((HourlySchedule)(queueService).QueueSchedule).Interval);
                        Assert.AreEqual(3, ((HourlySchedule)(queueService).QueueSchedule).Minute);
                        break;
                    default:
                        Assert.Fail("Unknown queue service for the queue: {0}", queueService.QueueName);
                        // ReSharper disable HeuristicUnreachableCode
                        break;
                    // ReSharper restore HeuristicUnreachableCode
                }
            }
        }
    
        [Test]
        [ExpectedException(typeof(QueueException), MatchType = MessageMatch.Contains, ExpectedMessage = "queue10")]
        public void ThrowsQueueExceptionForTheQueueWithMoreThanOneSchedule()
        {
            new BusService().ConfigureQueuesFromFile("queueWithTwoSchedules");
        }

        [Test]
        [ExpectedException(typeof(QueueException), MatchType = MessageMatch.Contains, ExpectedMessage = "queue10")]
        public void ThrowsQueueExceptionForTheQueueWithoutSchedule()
        {
            new BusService().ConfigureQueuesFromFile("queueWithNoSchedule");
        }

        [Test]
        [ExpectedException(typeof(QueueException), MatchType = MessageMatch.Contains, ExpectedMessage = "queue10")]
        public void ThrowsQueueExceptionForTheQueueWithoutTransport()
        {
            new BusService().ConfigureQueuesFromFile("queueWithoutTransport");
        }

        [Test]
        [ExpectedException(typeof(QueueException), MatchType = MessageMatch.Contains, ExpectedMessage = "unknownSection10")]
        public void ThrowsQueueExceptionForUnknownSection()
        {
            new BusService().ConfigureQueuesFromFile("unknownSection10");
        }
    }
}
