﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Threading;
using Castle.MicroKernel.Registration;
using DocaAzure.CastleWidsorTypeProvider;
using DocaAzure.Messaging.Core;
using DocaAzure.Messaging.Providers;
using DocaAzure.Scheduler;
using DocaAzure.Scheduler.Providers;
using DocaAzure.Storage;
using DocaAzure.TestUtils;
using DocaLabs.Utils;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaAzure.Messaging.IntegrationTests
{
    [TestFixture]
    public class ComprehensiveQueueTests
    {
        #region test classes

        public class InnerModel : IDisposable
        {
            public void Dispose()
            {
                InnerDisposeCalled = true;
            }
        }

        public class OutterModel : IDisposable
        {
            public InnerModel Inner { get; set; }
            public void Dispose()
            {
                OutterDisposeCalled = true;
            }
        }

        [Endpoint(TestQueueName), Serializable]
        public class PositiveTestMessage : IMessage
        {
            public Guid Value { get; set; }
        }

        public class PositiveTestHandler : IMessageHandler<PositiveTestMessage>
        {
            public OutterModel Model { get; set; }

            public void Handle(PositiveTestMessage message)
            {
                Mark = message.Value;
                ++Counter;
                Signal.Set();
            }
        }

        [Endpoint(TestQueueName), Serializable]
        public class PoisonTestMessage : MessageBase
        {
            public Guid Value { get; set; }
        }

        public class PoisonTestHandler : IDequeueCounter, IMessageHandler<PoisonTestMessage>
        {
            public int DequeueCount { get; set; }

            public int MaxDequeueCount { get; set; }

            public void Handle(PoisonTestMessage message)
            {
                HandlerDequeueCount = DequeueCount;
                Mark = message.Value;
                ++Counter;
                throw new Exception();
            }
        }

        [Endpoint(TestQueueName), Serializable]
        public class WithoutHandlerTestMessage : IMessage
        {
            public Guid Value { get; set; }
        }

        [Endpoint("someunknownqueue"), Serializable]
        public class NonExistingQueueNameTestMessage : IMessage
        {
        }

        #endregion test classes

        public const string ScheduledContextTableName = "intergrationtestschedule";
        public const string TestQueueName = "integrationtestqueue";
        private const string HourlyTestQueueName = "hourlyintegrationtestqueue";
        private const string DailyTestQueueName = "dailyintegrationtestqueue";
        public const string TestErrorQueueName = "integrationerrortestqueue";
        public static Guid Mark { get; set; }
        public static int Counter { get; set; }
        public static ManualResetEvent Signal { get; set; }
        public static bool InnerDisposeCalled { get; set; }
        public static bool OutterDisposeCalled { get; set; }
        public static int HandlerDequeueCount;

        private static HourlySchedule HourlyQueueSchedule { get; set; }
        private static DailySchedule DailyQueueSchedule { get; set; }
        private static IDocaScheduleContext HourlyScheduleContext { get; set; }
        private static IDocaScheduleContext DailyScheduleContext { get; set; }
        private static DateTime ServerInitializedAt { get; set; }
        private static DateTime? FirstHourlyScheduledTime { get; set; }
        private static DateTime? FirstDailyScheduledTime { get; set; }
        
        [SetUpFixture]
        public class FixtureSetup
        {
            [SetUp]
            public void SetupServer()
            {
                PerformanceCountersInstaller.CreateCustomCounters(true);

                LoggingUtils.SetupConsoleLogging();

                AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

                DefaultTypeProvider.Default.Container.Register(

                    Component.For<InnerModel>()
                        .LifestyleScoped(),

                    Component.For<OutterModel>()
                        .LifestyleScoped()
                );


                var errorQueue = AzureStorageAccountFactory.GetQueueReference(TestErrorQueueName);
                errorQueue.CreateIfNotExist();
                errorQueue.Clear();

                var q1 = AzureStorageAccountFactory.GetQueueReference(HourlyTestQueueName);
                q1.CreateIfNotExist();
                q1.Clear();

                var q2 = AzureStorageAccountFactory.GetQueueReference(DailyTestQueueName);
                q2.CreateIfNotExist();
                q2.Clear();

                PrepareQueueSchedules();

                // by default BusService uses "DefaultMessagingTypeFactory" 
                // (which uses "DefaultTypeProvider.Default")
                // to resolve message handler dependencies
                BusService.Service
                    //.Configure(DefaultTypeProvider.Default)
                    .AddAzureQueue(
                                    TestQueueName,
                                    (q, t) =>
                                    {
                                        q.ErrorQueueName = TestErrorQueueName;
                                        q.QueueSchedule = new RegularQueueSchedule
                                        {
                                            InitialPollInterval = TimeSpan.FromMilliseconds(10),
                                            MaximumWaitTimeWhenIdle = TimeSpan.FromMilliseconds(50),
                                            PollIntervalIncrement = TimeSpan.FromMilliseconds(10)
                                        };
                                        t.MessageInvisibleTime = TimeSpan.FromMilliseconds(50);
                                        q.MaxDequeueCount = 5;
                                        q.PurgeOnStartup = true;
                                    })
                    .AddAzureQueue(
                                    HourlyTestQueueName,
                                    (q, t) =>
                                    {
                                        q.QueueSchedule = HourlyQueueSchedule;
                                        q.PurgeOnStartup = false;
                                    })
                    .AddAzureQueue(
                                    DailyTestQueueName,
                                    (q, t) =>
                                    {
                                        q.QueueSchedule = DailyQueueSchedule;
                                        q.PurgeOnStartup = false;
                                    })
                    .RegisterHandlers(typeof(PositiveTestHandler).Assembly);

                BusService.Service.Start();

                ServerInitializedAt = DefaultClock.CurrentTime;

                while ((FirstHourlyScheduledTime = HourlyScheduleContext.ReadNextScheduledTime(HourlyTestQueueName)) == null)
                {
                }

                while ((FirstDailyScheduledTime = DailyScheduleContext.ReadNextScheduledTime(DailyTestQueueName)) == null)
                {
                }
            }

            private static void PrepareQueueSchedules()
            {
                HourlyScheduleContext = new AzureScheduleContext(ScheduledContextTableName);
                DailyScheduleContext = new AzureScheduleContext(ScheduledContextTableName);

                HourlyQueueSchedule = new HourlySchedule
                {
                    PollingTimeout = TimeSpan.FromMilliseconds(100),
                    Interval = 4,
                    Minute = 0,
                    ScheduleContext = HourlyScheduleContext
                };

                DailyQueueSchedule = new DailySchedule
                {
                    PollingTimeout = TimeSpan.FromMilliseconds(100),
                    Interval = 1,
                    Hour = 18,
                    ScheduleContext = DailyScheduleContext
                };
            }

            [TearDown]
            public void RemoveTestQueue()
            {
                BusService.Service.TryStop();
                Thread.Sleep(TimeSpan.FromSeconds(5));

                AzureStorageAccountFactory.GetQueueReference(TestQueueName).Delete();
                AzureStorageAccountFactory.GetQueueReference(HourlyTestQueueName).Delete();
                AzureStorageAccountFactory.GetQueueReference(DailyTestQueueName).Delete();
                AzureStorageAccountFactory.GetQueueReference(TestErrorQueueName).Delete();

                AzureStorageAccountFactory.Create().CreateCloudTableClient().DeleteTableIfExist(ScheduledContextTableName);

                AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
            }
        }

        [Test]
        public void MessageWithoutHandlerIsMovedToErrorQueue()
        {
            Counter = 0;
            Mark = Guid.Empty;
            Signal = new ManualResetEvent(false);

            var guid = Guid.NewGuid();

            new BusClient().Send(new WithoutHandlerTestMessage { Value = guid });

            Thread.Sleep(5000); // should be enough

            Assert.AreEqual(Guid.Empty, Mark);
            Assert.AreEqual(0, Counter);

            Assert.IsNull(AzureStorageAccountFactory.GetQueueReference(TestQueueName).GetMessage());

            var movedMessage = AzureStorageAccountFactory.GetQueueReference(TestErrorQueueName).GetMessage();
            Assert.IsNotNull(movedMessage);
            var transportMessage = new SmartAzureMessageSerializer().FromCloudQueueMessage(movedMessage);
            Assert.IsNotNull(transportMessage);
            Assert.IsNotNull(transportMessage.Body);
            Assert.IsInstanceOf<WithoutHandlerTestMessage>(transportMessage.Body);
            Assert.AreEqual(guid, ((WithoutHandlerTestMessage)transportMessage.Body).Value);
        }

        [Test]
        public void SendThrowsQueueExceptionWhenMessageHasWrongQueueNameAttribute()
        {
            try
            {
                new BusClient().Send(new NonExistingQueueNameTestMessage());
            }
            catch (QueueException e)
            {
                if (e.Message.IndexOf("doesn't exist") >= 0 && e.Message.IndexOf("destination queue") >= 0)
                    return;
            }
// ReSharper disable EmptyGeneralCatchClause
            catch (Exception)
// ReSharper restore EmptyGeneralCatchClause
            {
            }

            Assert.Fail("must throw QueueException for NonExistingQueueNameTestMessage");
        }

        [Test]
        public void PoisonMessageMovedToErrorQueueButCorrectMessageIsProcessed()
        {
            /*1 Poison message*/
            ProcessingPoisonMessage();

            /*2 Correct message*/
            ProcessingCorrectMessage();
        }

        private static void ProcessingPoisonMessage()
        {
            Counter = 0;
            Mark = Guid.Empty;
            Signal = new ManualResetEvent(false);

            var guid = Guid.NewGuid();

            var message = new PoisonTestMessage
                              {
                                  Value = guid
                              };

            new BusClient().Send(message);

            Thread.Sleep(10000); // should be enough to wait for cycle complete    
           

            Assert.AreEqual(guid, Mark);
            Assert.AreEqual(5, Counter);
            Assert.AreEqual(5, HandlerDequeueCount);


            var errorQueue = AzureStorageAccountFactory.GetQueueReference(TestErrorQueueName);

            Assert.IsNull(AzureStorageAccountFactory.GetQueueReference(TestQueueName).GetMessage());


            var movedMessage = errorQueue.GetMessage();
            Assert.IsNotNull(movedMessage);
            var transportMessage = new SmartAzureMessageSerializer().FromCloudQueueMessage(movedMessage);
            Assert.IsNotNull(transportMessage);
            Assert.IsNotNull(transportMessage.Body);
            Assert.IsInstanceOf<PoisonTestMessage>(transportMessage.Body);
            Assert.AreEqual(guid, ((PoisonTestMessage)transportMessage.Body).Value);

            errorQueue.Clear();
        }

        private static void ProcessingCorrectMessage()
        {
            Counter = 0;
            Mark = Guid.Empty;
            Signal = new ManualResetEvent(false);
            InnerDisposeCalled = false;
            OutterDisposeCalled = false;

            var guid = Guid.NewGuid();

            new BusClient().Send(new PositiveTestMessage { Value = guid });

            Signal.WaitOne();
            Thread.Sleep(100);

            Assert.AreEqual(guid, Mark);
            Assert.AreEqual(1, Counter);

            Thread.Sleep(1000); // wait for cleaning up in the message processing cycle

            Assert.IsTrue(OutterDisposeCalled, "the dispose on outter resolved dependency wasn't called");
            Assert.IsTrue(InnerDisposeCalled, "the dispose on inner resolved dependency wasn't called");

            Assert.IsNull(AzureStorageAccountFactory.GetQueueReference(TestQueueName).GetMessage());
            Assert.IsNull(AzureStorageAccountFactory.GetQueueReference(TestErrorQueueName).GetMessage());
        }

        [Test]
        public void CheckThatTheScheduleContextTablesWereInitialized()
        {
            Assert.AreEqual(new DateTime(ServerInitializedAt.Year, ServerInitializedAt.Month, ServerInitializedAt.Day, ServerInitializedAt.Hour, 0, 0).AddHours(4), FirstHourlyScheduledTime.GetValueOrDefault());

// ReSharper disable ConvertIfStatementToConditionalTernaryExpression
            if (ServerInitializedAt < new DateTime(ServerInitializedAt.Year, ServerInitializedAt.Month, ServerInitializedAt.Day, 17, 30, 0))
// ReSharper restore ConvertIfStatementToConditionalTernaryExpression
                Assert.AreEqual(new DateTime(ServerInitializedAt.Year, ServerInitializedAt.Month, ServerInitializedAt.Day, 18, 0, 0), FirstDailyScheduledTime.GetValueOrDefault());
            else
                Assert.AreEqual(new DateTime(ServerInitializedAt.Year, ServerInitializedAt.Month, ServerInitializedAt.Day, 18, 0, 0).AddDays(1), FirstDailyScheduledTime.GetValueOrDefault());
        }

        [Test]
        public void MiniStressTest()
        {
            Counter = 0;

            Signal = new ManualResetEvent(false);

            for (var i = 0; i < 100; ++i)
            {
                var message = new PositiveTestMessage
                {
                    Value = Guid.NewGuid()
                };

                new BusClient().Send(message);
            }

            Thread.Sleep(2000);

            Assert.AreEqual(100, Counter);

            Assert.IsNull(AzureStorageAccountFactory.GetQueueReference(TestQueueName).GetMessage());
            Assert.IsNull(AzureStorageAccountFactory.GetQueueReference(TestErrorQueueName).GetMessage());
        }
    }
}
