﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Threading;
using DocaAzure.Scheduler.Providers;
using DocaAzure.Storage;
using DocaAzure.TestUtils;
using DocaLabs.Utils;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;
using System.Collections.Generic;

namespace DocaAzure.Scheduler.IntegrationTests
{
    [TestFixture]
    public class ComprehensiveSchedulerTests
    {
        private const string ScheduledContextTableName = "intergrationtestscheduletable";
        private const string LeaseTableName = "intergartiontestleasetable";
        private const string HourlyTriggerName = "hourlytrigger";
        private const string DailyTriggerName = "dailytrigger";
        private const int Speed = 250;

        public class TestTrigger : ScheduledTrigger
        {
            public List<DateTime> Calls { get; set; }

            public TestTrigger(string name)
                : base(name)
            {
                Calls = new List<DateTime>();
            }

            protected override void Do()
            {
                Calls.Add(Schedule.GetCurrentLocalTime());
            }

            protected override string GetTenantName()
            {
                return "test tenant for " + Name;
            }

            public override DateTime ValidateNextTime(DateTime now, DateTime next)
            {
                return now.AddMilliseconds(10);
            }
        }

        private HourlySchedule HourlyQueueSchedule { get; set; }
        private DailySchedule DailyQueueSchedule { get; set; }

        private TestTrigger HourlyTrigger { get; set; }
        private TestTrigger DailyTrigger { get; set; }

        private CloudStorageAccount StorageAccount { get; set; }
        private CloudTableClient TableClient { get; set; }

        private IClock _originalClock;
        private SpeedyClock _speedyClock;

        [SetUp]
        public void Setup()
        {
            _originalClock = DefaultClock.Current;

            LoggingUtils.SetupConsoleLogging();

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            StorageAccount = AzureStorageAccountFactory.Create();
            TableClient = new CloudTableClient(StorageAccount.TableEndpoint.ToString(), StorageAccount.Credentials);

            TableClient.DeleteTableIfExist(ScheduledContextTableName);
            TableClient.DeleteTableIfExist(LeaseTableName);

            DefaultClock.Current = _speedyClock = new SpeedyClock(Speed);
        }

        [TearDown]
        public void TearDown()
        {
            DefaultClock.Current = _originalClock;

            TableClient.DeleteTableIfExist(ScheduledContextTableName);
            TableClient.DeleteTableIfExist(LeaseTableName);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        [Category("LongRunninTest")]
        public void SchedulerTests()
        {
            var target = Arrange();

            target.Start();

            var stopMark = DateTime.Parse("2011-02-26T20:10:00.000");

            while (DailyQueueSchedule.GetCurrentLocalTime() < stopMark)
            {
                Thread.Sleep(100);
            }

            target.Stop();

            // Assert when triggers were called
            Assert.AreEqual(8, HourlyTrigger.Calls.Count);
            Assert.LessOrEqual(DateTime.Parse("2011-02-25T16:00:00.000"), HourlyTrigger.Calls[0]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-25T16:30:00.000"), HourlyTrigger.Calls[0]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-25T20:00:00.000"), HourlyTrigger.Calls[1]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-25T20:30:00.000"), HourlyTrigger.Calls[1]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T00:00:00.000"), HourlyTrigger.Calls[2]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T00:30:00.000"), HourlyTrigger.Calls[2]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T04:00:00.000"), HourlyTrigger.Calls[3]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T04:30:00.000"), HourlyTrigger.Calls[3]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T08:00:00.000"), HourlyTrigger.Calls[4]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T08:30:00.000"), HourlyTrigger.Calls[4]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T12:00:00.000"), HourlyTrigger.Calls[5]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T12:30:00.000"), HourlyTrigger.Calls[5]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T16:00:00.000"), HourlyTrigger.Calls[6]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T16:30:00.000"), HourlyTrigger.Calls[6]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T20:00:00.000"), HourlyTrigger.Calls[7]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T20:30:00.000"), HourlyTrigger.Calls[7]);

            Assert.AreEqual(2, DailyTrigger.Calls.Count);
            Assert.LessOrEqual(DateTime.Parse("2011-02-25T18:00:00.000"), DailyTrigger.Calls[0]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-25T18:30:00.000"), DailyTrigger.Calls[0]);
            Assert.LessOrEqual(DateTime.Parse("2011-02-26T18:00:00.000"), DailyTrigger.Calls[1]);
            Assert.GreaterOrEqual(DateTime.Parse("2011-02-26T18:30:00.000"), DailyTrigger.Calls[1]);
        }

        private SchedulerService Arrange()
        {
            PrepareSchedules();

            var service = new SchedulerService();

            service
                .AddTrigger(HourlyTrigger)
                .AddTrigger(DailyTrigger);

            return service;
        }

        private void PrepareSchedules()
        {
            _speedyClock.InitialTime = DateTime.Parse("2011-02-25T12:31:45.000");

            var scheduleContext = new AzureScheduleContext(ScheduledContextTableName);

            HourlyQueueSchedule = new HourlySchedule
            {
                PollingTimeout = TimeSpan.FromMilliseconds(100),
                ScheduleContext = scheduleContext,
                Interval = 4,
                Minute = 0
            };

            DailyQueueSchedule = new DailySchedule
            {
                PollingTimeout = TimeSpan.FromMilliseconds(100),
                ScheduleContext = scheduleContext,
                Interval = 1,
                Hour = 18
            };

            var leaseProvider = new AzureScheduleLeaseProvider(LeaseTableName);

            HourlyTrigger = new TestTrigger(HourlyTriggerName)
                                {
                                    Schedule = HourlyQueueSchedule,
                                    LeaseProvider = leaseProvider
                                };

            DailyTrigger = new TestTrigger(DailyTriggerName)
                               {
                                   Schedule = DailyQueueSchedule,
                                   LeaseProvider = leaseProvider
                               };
        }
    }
}
