﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Threading;
using DocaAzure.Scheduler.Providers;
using DocaAzure.Storage;
using DocaAzure.TestUtils;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaAzure.Scheduler.IntegrationTests.Providers
{
    [TestFixture]
    public class AzureScheduleLeaseProviderTests
    {
        private const string TestTableName = "integrationleaseprovidertests";

        public class RaceConditionTestTarget : AzureScheduleLeaseProvider
        {
            public Type ExceptionToBeThrownFromSaveChanges { get; set; }
            public ManualResetEventSlim ContinueWithSave { get; private set; }
            public int NumberOfSaveCalls { get; set; }
            public int Instance { get; set; }

            public RaceConditionTestTarget(string tableName, int instance)
                : base(tableName)
            {
                Instance = instance;
                ContinueWithSave = new ManualResetEventSlim(false);
            }

            protected override void SaveChanges(TableServiceContext context)
            {
                ++NumberOfSaveCalls;

                Console.WriteLine(@"Instance {0} is waiting for event to save", Instance);
                ContinueWithSave.Wait();

                if (ExceptionToBeThrownFromSaveChanges != null)
                {
                    Console.WriteLine(@"Instance {0} is throwing exception {1}", Instance, ExceptionToBeThrownFromSaveChanges);
                    throw (Exception) Activator.CreateInstance(ExceptionToBeThrownFromSaveChanges);
                }

                Console.WriteLine(@"Instance {0} is trying to save", Instance);
                base.SaveChanges(context);
            }
        }

        public class ExtendedRaceConditionTarget
        {
            public bool LastLeaseResult { get; set; }
            public int LeaseCount { get; set; }
            public ManualResetEventSlim IterationCompleted1 { get; set; }
            public ManualResetEventSlim IterationCompleted2 { get; set; }
            public string TenantName { get; set; }
            public AzureScheduleLeaseProvider Target { get; set; }

            public ManualResetEventSlim SyncEvent1 { get; set; }
            public ManualResetEventSlim SyncEvent2 { get; set; }

            internal volatile bool Stopping;

            public ExtendedRaceConditionTarget(string tenantName)
            {
                Target = new AzureScheduleLeaseProvider(TestTableName);
                IterationCompleted1 = new ManualResetEventSlim(false);
                IterationCompleted2 = new ManualResetEventSlim(false);
                TenantName = tenantName;
            }

            public void Run()
            {
                while (!Stopping)
                {
                    SyncEvent1.Wait();

                    var now = DateTime.Now;

                    LastLeaseResult = Target.TryLeasing("schedule1", TenantName, now, now.AddMilliseconds(300));
                    if (LastLeaseResult)
                        ++LeaseCount;

                    IterationCompleted1.Set();

                    SyncEvent2.Wait();

                    IterationCompleted2.Set();
                }
            }
        }

        private CloudStorageAccount StorageAccount { get; set; }
        private CloudTableClient TableClient { get; set; }

        [SetUp]
        public void Setup()
        {
            LoggingUtils.SetupConsoleLogging();

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;

            StorageAccount = AzureStorageAccountFactory.Create();
            TableClient = new CloudTableClient(StorageAccount.TableEndpoint.ToString(), StorageAccount.Credentials);

            TableClient.DeleteTableIfExist(TestTableName);
        }

        [TearDown]
        public void TearDown()
        {
            TableClient.DeleteTableIfExist(TestTableName);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void TestsSequentialLeasingOfExpiredLeases()
        {
            // emulate real runtime environment when workers will have their oen instances of the AzureScheduleLeaseProvider
            var target1 = new AzureScheduleLeaseProvider(TestTableName);
            var target2 = new AzureScheduleLeaseProvider(TestTableName);
            var target3 = new AzureScheduleLeaseProvider(TestTableName);

            // aquaire lase to tenant1
            var now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target1.TryLeasing("schedule1", "tenant1", now, now.AddMilliseconds(1000)));

            // shouldn't be problem to aquare lease to another schedule
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target3.TryLeasing("another schedule", "another tenant", now, now.AddMilliseconds(1000)));

            // lease should be refused to tenant2 as it's still active for tenant1
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsFalse(target2.TryLeasing("schedule1", "tenant2", now, now.AddMilliseconds(1000)));

            // wait until lease expires
            Thread.Sleep(1100);

            // now the lease should be aquired for tenant2
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target2.TryLeasing("schedule1", "tenant2", now, now.AddMilliseconds(1000)));

            // and refused for tenant1
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsFalse(target1.TryLeasing("schedule1", "tenant1", now, now.AddMilliseconds(1000)));
        }

        [Test]
        public void TestsSequentialLeasing()
        {
            // emulate real runtime environment when workers will have their oen instances of the AzureScheduleLeaseProvider
            var target1 = new AzureScheduleLeaseProvider(TestTableName);
            var target2 = new AzureScheduleLeaseProvider(TestTableName);
            var target3 = new AzureScheduleLeaseProvider(TestTableName);

            // aquaire lase to tenant1
            var now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target1.TryLeasing("schedule1", "tenant1", now, now.AddMilliseconds(1000)));

            // shouldn't be problem to aquare lease to another schedule
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target3.TryLeasing("another schedule", "another tenant", now, now.AddMilliseconds(1000)));

            // lease should be refused to tenant2 as it's still active for tenant1
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsFalse(target2.TryLeasing("schedule1", "tenant2", now, now.AddMilliseconds(1000)));

            // surrender lease for tenant1
            Thread.Sleep(1100);

            // now the lease should be aquired for tenant2
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target2.TryLeasing("schedule1", "tenant2", now, now.AddMilliseconds(1000)));

            // and refused for tenant1
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsFalse(target1.TryLeasing("schedule1", "tenant1", now, now.AddMilliseconds(1000)));

            // surrender lease for tenant1
            Thread.Sleep(1100);

            // aquire lease for tenant1
            now = DateTime.Now;
            Console.WriteLine(now.ToString("o"));
            Assert.IsTrue(target1.TryLeasing("schedule1", "tenant1", now, now.AddMilliseconds(1000)));
        }
    
        [Test]
        public void TryLeasingRaceConditionTestWhenRowIsJustCreated()
        {
            // emulate real runtime environment when workers will have their oen instances of the AzureScheduleLeaseProvider
            var target1 = new RaceConditionTestTarget(TestTableName, 1);
            var target2 = new RaceConditionTestTarget(TestTableName, 2);

            bool? b1 = null, b2 = null;

            var thread1 = new Thread(() =>
            {
                var now1 = DateTime.Now;
                b1 = target1.TryLeasing("schedule1", "tenant1", now1, now1.AddMilliseconds(1000));
            }) { IsBackground = true };
            thread1.Start();
           
            // give chance for the first call to create the item (it's precondition of all tests)
            Thread.Sleep(1100);

            var thread2 = new Thread(() =>
            {
                var now2 = DateTime.Now;
                b2 = target2.TryLeasing("schedule1", "tenant2", now2, now2.AddMilliseconds(1000));
            }) { IsBackground = true };
            thread2.Start();

            // release the second
            target2.ContinueWithSave.Set();
            Thread.Sleep(1100);

            // release the first which would force 'already exists' error condition in the first
            target1.ContinueWithSave.Set();
            Thread.Sleep(1100);

            // assert that lease was aquired by the second
            Assert.IsNotNull(b1);
            Assert.IsNotNull(b2);
            Assert.IsFalse(b1.GetValueOrDefault());
            Assert.IsTrue(b2.GetValueOrDefault());

            Assert.AreEqual(1, target1.NumberOfSaveCalls);
            Assert.AreEqual(1, target2.NumberOfSaveCalls);
        }

        [Test]
        public void TryLeasingOptimisticConcurrenyTest()
        {
            // insure that the lease record exists
            var now = DateTime.Now;
            var target0 = new AzureScheduleLeaseProvider(TestTableName);
            Assert.IsTrue(target0.TryLeasing("schedule1", "tenant1", now, now.AddMilliseconds(10)));
            Thread.Sleep(100);

            // emulate real runtime environment when workers will have their oen instances of the AzureScheduleLeaseProvider
            var target1 = new RaceConditionTestTarget(TestTableName, 1);
            var target2 = new RaceConditionTestTarget(TestTableName, 2);

            bool? b1 = null, b2 = null;

            var thread1 = new Thread(() =>
            {
                var now1 = DateTime.Now;
                b1 = target1.TryLeasing("schedule1", "tenant1", now1, now1.AddMilliseconds(1000));
            }) { IsBackground = true };
            thread1.Start();

            // give chance for the first call to create the item (it's precondition of all tests)
            Thread.Sleep(1100);

            var thread2 = new Thread(() =>
            {
                var now2 = DateTime.Now;
                b2 = target2.TryLeasing("schedule1", "tenant2", now2, now2.AddMilliseconds(1000));
            }) { IsBackground = true };
            thread2.Start();

            // release the second
            target2.ContinueWithSave.Set();
            Thread.Sleep(1100);

            // release the first which would force 'already exists' error condition in the first
            target1.ContinueWithSave.Set();
            Thread.Sleep(1100);

            // assert that lease was aquired by the second
            Assert.IsNotNull(b1);
            Assert.IsNotNull(b2);
            Assert.IsFalse(b1.GetValueOrDefault());
            Assert.IsTrue(b2.GetValueOrDefault());

            Assert.AreEqual(1, target1.NumberOfSaveCalls);
            Assert.AreEqual(1, target2.NumberOfSaveCalls);
        }

        [Test]
        [Category("LongRunninTest")]
        public void TryLeasingOptimisticConcurrenyExtendedTest()
        {
            var synchEvent1 = new ManualResetEventSlim(false);
            var synchEvent2 = new ManualResetEventSlim(false);

            // emulate real runtime environment when workers will have their oen instances of the AzureScheduleLeaseProvider
            var target1 = new ExtendedRaceConditionTarget("tenant1") { SyncEvent1 = synchEvent1, SyncEvent2 = synchEvent2 };
            var target2 = new ExtendedRaceConditionTarget("tenant2") { SyncEvent1 = synchEvent1, SyncEvent2 = synchEvent2 };
            var target3 = new ExtendedRaceConditionTarget("tenant3") { SyncEvent1 = synchEvent1, SyncEvent2 = synchEvent2 };
            var target4 = new ExtendedRaceConditionTarget("tenant4") { SyncEvent1 = synchEvent1, SyncEvent2 = synchEvent2 };

            var thread1 = new Thread(target1.Run) { IsBackground = true };
            thread1.Start();

            var thread2 = new Thread(target2.Run) { IsBackground = true };
            thread2.Start();

            var thread3 = new Thread(target3.Run) { IsBackground = true };
            thread3.Start();

            var thread4 = new Thread(target4.Run) { IsBackground = true };
            thread4.Start();

            for (var i = 0; i < 100; i++)
            {
                synchEvent1.Set();

                target1.IterationCompleted1.Wait();
                target2.IterationCompleted1.Wait();
                target3.IterationCompleted1.Wait();
                target4.IterationCompleted1.Wait();

                synchEvent1.Reset();

                target1.IterationCompleted1.Reset();
                target2.IterationCompleted1.Reset();
                target3.IterationCompleted1.Reset();
                target4.IterationCompleted1.Reset();

                Assert.IsFalse(target1.LastLeaseResult == target2.LastLeaseResult == target3.LastLeaseResult == target4.LastLeaseResult, "At least one and only one thread must lease");

                synchEvent2.Set();

                target1.IterationCompleted2.Wait();
                target2.IterationCompleted2.Wait();
                target3.IterationCompleted2.Wait();
                target4.IterationCompleted2.Wait();

                synchEvent2.Reset();

                target1.IterationCompleted2.Reset();
                target2.IterationCompleted2.Reset();
                target3.IterationCompleted2.Reset();
                target4.IterationCompleted2.Reset();

                Thread.Sleep(310);
            }

            target1.Stopping = true;
            target2.Stopping = true;
            target3.Stopping = true;
            target4.Stopping = true;

            Assert.AreEqual(100, target1.LeaseCount + target2.LeaseCount + target3.LeaseCount + target4.LeaseCount);
        }

        [Test]
        public void TryLeasingExceptionTest()
        {
            // emulate real runtime environment when workers will have their oen instances of the AzureScheduleLeaseProvider
            var target1 = new RaceConditionTestTarget(TestTableName, 1);
            var target2 = new RaceConditionTestTarget(TestTableName, 2)
            {
                ExceptionToBeThrownFromSaveChanges = typeof(InvalidOperationException)
            };

            bool? b1 = null, b2 = null;

            var thread1 = new Thread(() =>
            {
                var now1 = DateTime.Now;
                b1 = target1.TryLeasing("schedule1", "tenant1", now1, now1.AddMilliseconds(1000));
            }) { IsBackground = true };
            thread1.Start();

            // give chance for the first call to create the item (it's precondition of all tests)
            Thread.Sleep(1000);

            var thread2 = new Thread(() =>
            {
                var now2 = DateTime.Now;
                b2 = target2.TryLeasing("schedule1", "tenant2", now2, now2.AddMilliseconds(1000));
            }) { IsBackground = true };
            thread2.Start();

            // release the second
            target2.ContinueWithSave.Set();
            Thread.Sleep(1000);

            // release the first which will aquire the lease as the second will fail to save any change
            target1.ContinueWithSave.Set();
            Thread.Sleep(1000);

            // assert that lease was aquired by the second
            Assert.IsNotNull(b1);
            Assert.IsNotNull(b2);
            Assert.IsTrue(b1.GetValueOrDefault());
            Assert.IsFalse(b2.GetValueOrDefault());

            Assert.AreEqual(1, target1.NumberOfSaveCalls);
            Assert.AreEqual(1, target2.NumberOfSaveCalls);
        }
    }
}
