﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using DocaAzure.Utils;
using Microsoft.WindowsAzure.StorageClient;
using Moq;
using NUnit.Framework;

namespace DocaAzure.Storage.IntegrationTests
{
    [TestFixture]
    public class BlobStorageProviderExtensionsTests
    {
        public class TestException : Exception
        {
        }

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void Cleanup()
        {
            var account = AzureStorageAccountFactory.Create();
            var blobClient = new CloudBlobClient(account.BlobEndpoint, account.Credentials);

            CommonTestUtils.SafeDeleteContainer(blobClient, TestBlobName.TestContainerName1);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void RoundtripForDeflateAndPutBlobAndGetBlobAndInflateAndThenDeleteBlob()
        {
            var provider = new BlobStorageProvider();

            // create
            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mycompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.DeflateAndPutBlob(blobName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            using (var stream = provider.GetBlobAndInflate(blobName))
            {
                result = stream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(blobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));
        }

        [Test]
        public void RoundtripForDeflateAndPutBlobAndGetBlobAndInfalteAndThenDeleteBlobAndCreatesContainerWhenItDoesntExist()
        {
            var provider = new BlobStorageProvider();

            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));

            var blobName = new TestBlobName { TestName = "mycompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.DeflateAndPutBlob(blobName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            using (var stream = provider.GetBlobAndInflate(blobName))
            {
                result = stream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(blobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));
        }

        [Test]
        public void RoundtripForPutObjectAndGetObjectAndDeleteBlob()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "myobjectblob" };

            var guid = Guid.NewGuid();
            var o = new TestSerializationClass
            {
                Property3 = new InnerTestSerializationClass
                {
                    Property1 = CommonTestUtils.SampleData,
                    Property2 = 42
                },
                Property4 = guid
            };

            provider.PutObject(blobName, o);

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            var result = provider.GetObject<TestSerializationClass>(blobName);

            Assert.AreEqual(CommonTestUtils.SampleData, result.Property3.Property1);
            Assert.AreEqual(42, result.Property3.Property2);
            Assert.AreEqual(guid, result.Property4);

            provider.DeleteBlob(blobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));
        }

        [Test]
        public void RoundtripForPutObjectAndGetObjectAndDeleteBlobAndCreatesContainerWhenItDoesntExist()
        {
            var provider = new BlobStorageProvider();

            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));

            var blobName = new TestBlobName { TestName = "myobjectblob" };

            var guid = Guid.NewGuid();
            var o = new TestSerializationClass
            {
                Property3 = new InnerTestSerializationClass
                {
                    Property1 = CommonTestUtils.SampleData,
                    Property2 = 42
                },
                Property4 = guid
            };

            provider.PutObject(blobName, o);

            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            var result = provider.GetObject<TestSerializationClass>(blobName);

            Assert.AreEqual(CommonTestUtils.SampleData, result.Property3.Property1);
            Assert.AreEqual(42, result.Property3.Property2);
            Assert.AreEqual(guid, result.Property4);

            provider.DeleteBlob(blobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));
        }

        [Test]
        public void GetObjectReturnsNullForNonExistentBlob()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            var result = provider.GetObject<TestSerializationClass>(blobName);

            Assert.IsNull(result);
        }

        [Test]
        public void GetBlobAndInflateReturnsNullForNonExistentBlob()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            var result = provider.GetBlobAndInflate(blobName);

            Assert.IsNull(result);
        }

        [Test]
        public void ReadAllReturnsAllBlobContent()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(blobName, sourceStream);
            }

            string result;
            using (var resultStream = provider.ReadAll(blobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);
        }

        [Test]
        public void ReadAllTextReturnsAllBlobContent()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(blobName, sourceStream);
            }

            string result = provider.ReadAllText(blobName);

            Assert.AreEqual(CommonTestUtils.SampleData, result);
        }

        [Test]
        public void ReadAllReturnsNullForBlobInNonExistingContainer()
        {
            var provider = new BlobStorageProvider();

            var blobName = new TestBlobName("very-strange-container") { TestName = "mynoncompressedtestblob" };

            var result = provider.ReadAll(blobName);

            Assert.IsNull(result);
        }

        [Test]
        public void ReadAllReturnsNullForNonExistentBlobInExistingContainer()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            var result = provider.ReadAll(blobName);

            Assert.IsNull(result);
        }

        [Test]
        public void ReadAllTextReturnsNullForBlobInNonExistingContainer()
        {
            var provider = new BlobStorageProvider();

            var blobName = new TestBlobName("very-strange-container") { TestName = "mynoncompressedtestblob" };

            var result = provider.ReadAllText(blobName);

            Assert.IsNull(result);
        }

        [Test]
        public void ReadAllTextReturnsNullForNonExistentBlobInExistingContainer()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            var result = provider.ReadAllText(blobName);

            Assert.IsNull(result);
        }

        [Test]
        [ExpectedException(typeof(TestException))]
        public void ReadAllThrowsExceptionWhenThereIsExceptionOtherThanStorageClientException()
        {
            var provider = new Mock<IBlobStorageProvider>();
            provider.Setup(x => x.GetBlob(It.IsAny<BlobName>())).Throws(new TestException());

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            provider.Object.ReadAllText(blobName);
        }

        [Test]
        [ExpectedException(typeof(StorageClientException))]
        public void ReadAllThrowsExceptionWhenThereIsStorageClientExceptionOtherThanResourceNotFound()
        {
            var provider = new Mock<IBlobStorageProvider>();
            provider.Setup(x => x.GetBlob(It.IsAny<BlobName>())).Throws(new StorageClientException());

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            provider.Object.ReadAllText(blobName);
        }

        [Test]
        [ExpectedException(typeof(TestException))]
        public void GetObjectThrowsExceptionWhenThereIsExceptionOtherThanStorageClientException()
        {
            var provider = new Mock<IBlobStorageProvider>();
            provider.Setup(x => x.GetBlob(It.IsAny<BlobName>())).Throws(new TestException());

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            provider.Object.GetObject<TestSerializationClass>(blobName);
        }

        [Test]
        [ExpectedException(typeof(StorageClientException))]
        public void GetObjectThrowsExceptionWhenThereIsStorageClientExceptionOtherThanResourceNotFound()
        {
            var provider = new Mock<IBlobStorageProvider>();
            provider.Setup(x => x.GetBlob(It.IsAny<BlobName>())).Throws(new StorageClientException());

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            provider.Object.GetObject<TestSerializationClass>(blobName);
        }

        [Test]
        [ExpectedException(typeof(TestException))]
        public void GetBlobAndInflateThrowsExceptionWhenThereIsExceptionOtherThanStorageClientException()
        {
            var provider = new Mock<IBlobStorageProvider>();
            provider.Setup(x => x.GetBlob(It.IsAny<BlobName>())).Throws(new TestException());

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            provider.Object.GetBlobAndInflate(blobName);
        }

        [Test]
        [ExpectedException(typeof(StorageClientException))]
        public void GetBlobAndInflateThrowsExceptionWhenThereIsStorageClientExceptionOtherThanResourceNotFound()
        {
            var provider = new Mock<IBlobStorageProvider>();
            provider.Setup(x => x.GetBlob(It.IsAny<BlobName>())).Throws(new StorageClientException());

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            provider.Object.GetBlobAndInflate(blobName);
        }
    }
}
