﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Threading;
using DocaAzure.Utils;
using DocaLabs.Utils;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaAzure.Storage.IntegrationTests
{
    [TestFixture]
    public class BlobStorageProviderTests
    {
        public const string TestContainerName2 = "blob-test-container-2";
        public const string TestContainerName3 = "blob-test-container-3";

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void Cleanup()
        {
            var account = AzureStorageAccountFactory.Create();
            var blobClient = new CloudBlobClient(account.BlobEndpoint, account.Credentials);

            CommonTestUtils.SafeDeleteContainer(blobClient, TestBlobName.TestContainerName1);
            CommonTestUtils.SafeDeleteContainer(blobClient, TestContainerName2);
            CommonTestUtils.SafeDeleteContainer(blobClient, TestContainerName3);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void CreateAndDeleteContainerWithPrivateBlobs()
        {
            var provider = new BlobStorageProvider();

            // create
            var result = provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsTrue(result);
            Assert.IsTrue(CommonTestUtils.IsContainerWithPrivateBlobs(TestBlobName.TestContainerName1));


            // delete
            result = provider.DeleteContainer(TestBlobName.TestContainerName1);

            Assert.IsTrue(result);
            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
        }

        [Test]
        public void CreateContainerReturnFalseIfItAlreadyExists()
        {
            var provider = new BlobStorageProvider();

            // create first time
            var result = provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsTrue(result);
            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));

            // tries to create second time
            result = provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsFalse(result);
            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
        }

        [Test]
        [Category("RequiresRealAccountCredentials")]
        [Ignore("The test requires the real azure storage account as the emulator doesn't reproduce completely the behaviour when a container is deleted")]
        public void CreatesContainerImmediatellyAfterDeletion()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;

            CloudStorageAccount.SetConfigurationSettingPublisher(AzureStorageAccountFactory.GetConfigurationSettingPublisher());

            AzureStorageAccountFactory.DefaultConnectionString = "<your account connection string>";
            
            var provider = new BlobStorageProvider();

            // create first time
            var result = provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsTrue(result);
            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                for (var i = 0; i < 100; ++i)
                {
                    sourceStream.ResetToBegining();
                    provider.PutBlob(new TestBlobName { TestName = "inblob" + i }, sourceStream);
                }
            }
            
            provider.DeleteContainer(TestBlobName.TestContainerName1);

            // tries to create second time
            result = provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsTrue(result);
            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
        }

        [Test]
        public void CreateAndDeleteContainerWithPublicBlobs()
        {
            var provider = new BlobStorageProvider();

            // create
            var result = provider.CreateContainer(TestBlobName.TestContainerName1, true);

            Assert.IsTrue(result);
            Assert.IsFalse(CommonTestUtils.IsContainerWithPrivateBlobs(TestBlobName.TestContainerName1));


            // delete
            result = provider.DeleteContainer(TestBlobName.TestContainerName1);

            Assert.IsTrue(result);
            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
        }

        [Test]
        public void DeleteContainerReturnFalseForNonExistentContainer()
        {
            var result = new BlobStorageProvider().DeleteContainer("correctcontainernamebutdoesntexists");
            Assert.IsFalse(result);
        }

        [Test]
        [ExpectedException]
        public void CreateContainerThrowsExceptionIfThereIsErrorOtherThanAlreadyExists()
        {
            new BlobStorageProvider().CreateContainer("//Bad container name&");
        }

        [Test]
        [ExpectedException]
        public void DeleteContainerThrowsExceptionIfThereIsErrorOtherThanDoesntExists()
        {
            new BlobStorageProvider().DeleteContainer("//bad container name&");
        }

        [Test]
        public void ListAllContainersWithGivenPrefix()
        {
            var provider = new BlobStorageProvider();

            // create
            provider.CreateContainer(TestBlobName.TestContainerName1);
            provider.CreateContainer(TestContainerName2);
            provider.CreateContainer(TestContainerName3);

            var result = provider.ListContainers("blob-test-").ToList();

            Assert.AreEqual(3, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1);
            CollectionAssert.Contains(result, TestContainerName2);
            CollectionAssert.Contains(result, TestContainerName3);

            // delete
            provider.DeleteContainer(TestBlobName.TestContainerName1);
            provider.DeleteContainer(TestContainerName2);
            provider.DeleteContainer(TestContainerName3);
        }

        [Test]
        public void ListAllBlobsInGivenContainerWithOrWithoutGivenPrefix()
        {
            var provider = new BlobStorageProvider();

            // create
            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName1 = new TestBlobName {TestName = "inblob1"};
            var blobName2 = new TestBlobName { TestName = "inblob2" };
            var blobName3 = new TestBlobName { TestName = "in/blob3" };
            var blobName4 = new TestBlobName { TestName = "outblob4" };
            var blobName5 = new TestBlobName { TestName = "outblob5" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                sourceStream.ResetToBegining();
                provider.PutBlob(blobName1, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName2, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName3, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName4, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName5, sourceStream);
            }

            var result = provider.ListBlobs(TestBlobName.TestContainerName1, "in").ToList();
            Assert.AreEqual(3, result.Count);
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "in/");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "inblob1");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "inblob2");

            result = provider.ListBlobs(TestBlobName.TestContainerName1, "in/").ToList();
            Assert.AreEqual(1, result.Count);
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");

            result = provider.ListBlobs(TestBlobName.TestContainerName1 + "/", "in/").ToList();
            Assert.AreEqual(1, result.Count);
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");

            result = provider.ListBlobs(TestBlobName.TestContainerName1).ToList();
            Assert.AreEqual(5, result.Count);
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "inblob1");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "inblob2");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "in/");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "outblob4");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "outblob5");

            result = provider.ListBlobs(TestBlobName.TestContainerName1 + "/").ToList();
            Assert.AreEqual(5, result.Count);
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "inblob1");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "inblob2");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "in/");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "outblob4");
            AssertBlobName(result, TestBlobName.TestContainerName1 + "/" + "outblob5");

            // delete
            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void ListBlobsReturnsEmptyCollectionForNonExistentContainer()
        {
            var result = new BlobStorageProvider().ListBlobs("somethingwedontknowyet").ToList();
            Assert.AreEqual(0, result.Count);
        }

        [Test]
        public void FlatBlobListReturnsEmptyCollectionForNonExistentContainer()
        {
            var result = new BlobStorageProvider().FlatBlobList("somethingwedontknowyet").ToList();
            Assert.AreEqual(0, result.Count);
        }

        [Test]
        public void FlatBlobListListsAllBlobsInGivenContainer()
        {
            var provider = new BlobStorageProvider();

            // create
            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName1 = new TestBlobName { TestName = "inblob1" };
            var blobName2 = new TestBlobName { TestName = "inblob2" };
            var blobName3 = new TestBlobName { TestName = "in/blob3" };
            var blobName4 = new TestBlobName { TestName = "outblob4" };
            var blobName5 = new TestBlobName { TestName = "outblob5" };

            DateTime markerTime;

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                sourceStream.ResetToBegining();
                provider.PutBlob(blobName1, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName2, sourceStream);

                markerTime = DefaultClock.CurrentTime;
                Thread.Sleep(1000);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName3, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName4, sourceStream);

                sourceStream.ResetToBegining();
                provider.PutBlob(blobName5, sourceStream);
            }

            var result = provider.FlatBlobList(TestBlobName.TestContainerName1, null, BlobStorageProvider.IfNotModifiedSince(markerTime)).ToList();
            Assert.AreEqual(2, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "inblob1");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "inblob2");


            result = provider.FlatBlobList(TestBlobName.TestContainerName1, null, BlobStorageProvider.IfModifiedSince(markerTime)).ToList();
            Assert.AreEqual(3, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "outblob4");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "outblob5");


            result = provider.FlatBlobList(TestBlobName.TestContainerName1).ToList();
            Assert.AreEqual(5, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "inblob1");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "inblob2");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "outblob4");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "outblob5");


            result = provider.FlatBlobList(TestBlobName.TestContainerName1 + "/").ToList();
            Assert.AreEqual(5, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "inblob1");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "inblob2");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "outblob4");
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "outblob5");

            result = provider.FlatBlobList(TestBlobName.TestContainerName1, "in/").ToList();
            Assert.AreEqual(1, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");

            result = provider.FlatBlobList(TestBlobName.TestContainerName1 + "/", "in/").ToList();
            Assert.AreEqual(1, result.Count);
            CollectionAssert.Contains(result, TestBlobName.TestContainerName1 + "/" + "in/blob3");

            // delete
            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void RoundtripPutBlobAndGetBlobAndDeleteBlob()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName {TestName = "mynoncompressedtestblob"};

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(blobName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            using (var resultStream = provider.GetBlob(blobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(blobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void PutBlobWithContentTypeAndMetadataUsingExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                ((IBlobOperations)provider).PutBlob(blobName.FullName, sourceStream, "image/jpeg", new NameValueCollection { { "Key1", "Value1" } });
            }

            using (var resultStream = provider.GetBlob(blobName))
            {
                var result = resultStream.StreamToString();

                Assert.AreEqual(CommonTestUtils.SampleData, result);

                Assert.AreEqual("image/jpeg", resultStream.GetBlobProperties().ContentType);
                Assert.AreEqual(1, resultStream.GetBlobMetadata().Count);
                Assert.AreEqual("Value1", resultStream.GetBlobMetadata()["Key1"]);
            }
        }

        [Test]
        public void RoundtripPutBlobAndGetBlobAndDeleteBlobUsinExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                ((IBlobOperations)provider).PutBlob(blobName.FullName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            using (var resultStream = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            ((IBlobOperations)provider).DeleteBlob(blobName.FullName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void RoundtripPutBlobAndGetBlobAndDeleteBlobAndCreatesContainerWhenItDoesntExist()
        {
            var provider = new BlobStorageProvider();

            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(blobName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestBlobName.TestContainerName1));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            using (var resultStream = provider.GetBlob(blobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(blobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(blobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void CopiesBlob()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var sourceBlobName = new TestBlobName { TestName = "mynoncompressedtestblob1" };
            var destBlobName = new TestBlobName { TestName = "mynoncompressedtestblob2" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(sourceBlobName, sourceStream);
            }


            provider.CopyBlob(sourceBlobName, destBlobName);

            Assert.IsTrue(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            string result;
            using (var resultStream = provider.GetBlob(destBlobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(sourceBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));

            provider.DeleteBlob(destBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void CopiesBlobUsinExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var sourceBlobName = new TestBlobName { TestName = "mynoncompressedtestblob1" };
            var destBlobName = new TestBlobName { TestName = "mynoncompressedtestblob2" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                ((IBlobOperations)provider).PutBlob(sourceBlobName.FullName, sourceStream);
            }


            provider.CopyBlob(sourceBlobName, destBlobName);

            Assert.IsTrue(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            string result;
            using (var resultStream = ((IBlobOperations)provider).GetBlob(destBlobName.FullName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(sourceBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));

            ((IBlobOperations)provider).DeleteBlob(destBlobName.FullName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void CopiesBlobAndCreatesContainerWhenDestinationContainerDoesntExist()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestContainerName2));

            var sourceBlobName = new TestBlobName { TestName = "mynoncompressedtestblob1" };
            var destBlobName = new TestBlobName(TestContainerName2) { TestName = "mynoncompressedtestblob2" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(sourceBlobName, sourceStream);
            }


            provider.CopyBlob(sourceBlobName, destBlobName);

            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestContainerName2));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            string result;
            using (var resultStream = provider.GetBlob(destBlobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(sourceBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));

            provider.DeleteBlob(destBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
            provider.DeleteContainer(TestContainerName2);
        }

        [Test]
        public void MovesBlob()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var sourceBlobName = new TestBlobName { TestName = "mynoncompressedtestblob1" };
            var destBlobName = new TestBlobName { TestName = "mynoncompressedtestblob2" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(sourceBlobName, sourceStream);
            }


            provider.MoveBlob(sourceBlobName, destBlobName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            string result;
            using (var resultStream = provider.GetBlob(destBlobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(destBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void MovesBlobUsingExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var sourceBlobName = new TestBlobName { TestName = "mynoncompressedtestblob1" };
            var destBlobName = new TestBlobName { TestName = "mynoncompressedtestblob2" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                ((IBlobOperations)provider).PutBlob(sourceBlobName.FullName, sourceStream);
            }


            ((IBlobOperations)provider).MoveBlob(sourceBlobName.FullName, destBlobName.FullName);

            Assert.IsFalse(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            string result;
            using (var resultStream = ((IBlobOperations)provider).GetBlob(destBlobName.FullName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            ((IBlobOperations)provider).DeleteBlob(destBlobName.FullName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
        }

        [Test]
        public void MovesBlobAndCreatesContainerWhenDestinationContainerDoesntExist()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            Assert.IsFalse(CommonTestUtils.IsContainerExists(TestContainerName2));

            var sourceBlobName = new TestBlobName { TestName = "mynoncompressedtestblob1" };
            var destBlobName = new TestBlobName(TestContainerName2) { TestName = "mynoncompressedtestblob2" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(sourceBlobName, sourceStream);
            }


            provider.MoveBlob(sourceBlobName, destBlobName);

            Assert.IsTrue(CommonTestUtils.IsContainerExists(TestContainerName2));
            Assert.IsFalse(CommonTestUtils.IsBlobExists(sourceBlobName.FullName));
            Assert.IsTrue(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            string result;
            using (var resultStream = provider.GetBlob(destBlobName))
            {
                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);

            provider.DeleteBlob(destBlobName);
            Assert.IsFalse(CommonTestUtils.IsBlobExists(destBlobName.FullName));

            provider.DeleteContainer(TestBlobName.TestContainerName1);
            provider.DeleteContainer(TestContainerName2);
        }

        [Test]
        public void DeleteBlobReturnsFalseForBlobInNonExistentContainer()
        {
            var provider = new BlobStorageProvider();

            var blobName = new TestBlobName("very-strange-container") { TestName = "mynoncompressedtestblob" };

            var result = provider.DeleteBlob(blobName);

            Assert.IsFalse(result);
        }

        [Test]
        public void DeleteBlobReturnsFalseForNonExistentBlobInExistingContainer()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            var result = provider.DeleteBlob(blobName);

            Assert.IsFalse(result);
        }

        [Test]
        public void DeleteBlobReturnsFalseFroBlobInNonExistentContainerUsingExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            var blobName = new TestBlobName("very-strange-container") { TestName = "mynoncompressedtestblob" };

            var result = ((IBlobOperations)provider).DeleteBlob(blobName.FullName);

            Assert.IsFalse(result);
        }

        [Test]
        public void DeleteBlobReturnsFalseForNonExistentBlobInExistingContainerUsingExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            var result = ((IBlobOperations)provider).DeleteBlob(blobName.FullName);

            Assert.IsFalse(result);
        }

        [Test]
        public void GeneratesBlobName()
        {
            var provider = new BlobStorageProvider(); 
            
            var blobName = new TestBlobName { TestName = "myblob" };

            var result = provider.GenerateBlobUrl(blobName);

            Assert.IsTrue(result.StartsWith(@"http://127.0.0.1"));
            Assert.IsTrue(result.EndsWith(@"/devstoreaccount1/blob-test-container-1/myblob"));
        }

        [Test]
        public void ExtractsBlobName()
        {
            var blob = AzureStorageAccountFactory.GetBlobReference("test-container/sub1/blob");

            var result = BlobStorageProvider.ExtractBlobName(blob);

            Assert.AreEqual("test-container/sub1/blob", result);
        }

        private static void AssertBlobName(IEnumerable<IListBlobItem> list, string expected)
        {
            if (list.Any(item => item.Uri.ToString().EndsWith(expected)))
                return;

            Assert.Fail("Collection doesn't contain blob definition for {0}", expected);
        }
    }
}
