﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Specialized;
using System.IO;
using DocaAzure.Utils;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaAzure.Storage.IntegrationTests
{
    [TestFixture]
    public class BlobStreamExtensionsTests
    {
        public class BlobStreamTest : BlobStream
        {
            public override void Flush()
            {
                throw new NotImplementedException();
            }

            public override long Seek(long offset, SeekOrigin origin)
            {
                throw new NotImplementedException();
            }

            public override void SetLength(long value)
            {
                throw new NotImplementedException();
            }

            public override int Read(byte[] buffer, int offset, int count)
            {
                throw new NotImplementedException();
            }

            public override void Write(byte[] buffer, int offset, int count)
            {
                throw new NotImplementedException();
            }

            public override bool CanRead
            {
                get { throw new NotImplementedException(); }
            }

            public override bool CanSeek
            {
                get { throw new NotImplementedException(); }
            }

            public override bool CanWrite
            {
                get { throw new NotImplementedException(); }
            }

            public override long Length
            {
                get { throw new NotImplementedException(); }
            }

            public override long Position
            {
                get { throw new NotImplementedException(); }
                set { throw new NotImplementedException(); }
            }
        }

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void Cleanup()
        {
            var account = AzureStorageAccountFactory.Create();
            var blobClient = new CloudBlobClient(account.BlobEndpoint, account.Credentials);

            CommonTestUtils.SafeDeleteContainer(blobClient, TestBlobName.TestContainerName1);

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void MetadataAndPropertiesReadingAfterStreamUsinExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                ((IBlobOperations)provider).PutBlob(blobName.FullName, sourceStream, null, new NameValueCollection
                                {
                                    {"testKey1", "test-value1"},
                                    {"testKey2", "test-value2"}
                                });
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            NameValueCollection metadata;
            BlobProperties properties;

            using (var resultStream = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                result = resultStream.StreamToString();

                metadata = resultStream.GetBlobMetadata();
                properties = resultStream.GetBlobProperties();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);


            Assert.AreEqual(2, metadata.Count);
            CollectionAssert.Contains(metadata.AllKeys, "testKey1");
            CollectionAssert.Contains(metadata.AllKeys, "testKey2");
            Assert.AreEqual("test-value1", metadata["testKey1"]);
            Assert.AreEqual("test-value2", metadata["testKey2"]);

            Assert.IsNotNull(properties.ETag);
            Assert.Less(0, properties.Length);
            Assert.Less(DateTime.MinValue, properties.LastModifiedUtc);
        }

        [Test]
        public void MetadataAndPropertiesReadingBeforeStreamUsinExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                ((IBlobOperations)provider).PutBlob(blobName.FullName, sourceStream, null, new NameValueCollection
                                {
                                    {"testKey1", "test-value1"},
                                    {"testKey2", "test-value2"}
                                });
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            string result;
            NameValueCollection metadata;
            BlobProperties properties;

            using (var resultStream = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                metadata = resultStream.GetBlobMetadata();
                properties = resultStream.GetBlobProperties();

                result = resultStream.StreamToString();
            }

            Assert.AreEqual(CommonTestUtils.SampleData, result);


            Assert.AreEqual(2, metadata.Count);
            CollectionAssert.Contains(metadata.AllKeys, "testKey1");
            CollectionAssert.Contains(metadata.AllKeys, "testKey2");
            Assert.AreEqual("test-value1", metadata["testKey1"]);
            Assert.AreEqual("test-value2", metadata["testKey2"]);

            Assert.IsNotNull(properties.ETag);
            Assert.Less(0, properties.Length);
            Assert.Less(DateTime.MinValue, properties.LastModifiedUtc);
        }

        [Test]
        public void ExistsReturnTrueForExistingBlobWhenCalledBeforeAnyOtherMethod()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(blobName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            using (var resultStream = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                Assert.IsTrue(resultStream.Exists());
            }
        }

        [Test]
        public void ExistsReturnTrueForExistingBlobWhenCalledAfterOtherMethods()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var sourceStream = CommonTestUtils.SampleData.StringToStream())
            {
                provider.PutBlob(blobName, sourceStream);
            }

            Assert.IsTrue(CommonTestUtils.IsBlobExists(blobName.FullName));

            using (var resultStream = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                resultStream.StreamToString();

                Assert.IsTrue(resultStream.Exists());
            }
        }

        [Test]
        public void ExistsReturnsFalseForNonExistentBlobInExistingContainerUsingExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            provider.CreateContainer(TestBlobName.TestContainerName1);

            var blobName = new TestBlobName { TestName = "mynoncompressedtestblob" };

            using (var result = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                Assert.IsFalse(result.Exists());
            }
        }

        [Test]
        public void ExistsReturnsFalseForBlobInNonExistingContainerUsingExplicitInterface()
        {
            var provider = new BlobStorageProvider();

            var blobName = new TestBlobName("very-strange-blob") { TestName = "mynoncompressedtestblob" };

            using (var result = ((IBlobOperations)provider).GetBlob(blobName.FullName))
            {
                Assert.IsFalse(result.Exists());
            }
        }
    
        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void ExistsThrowsArgumentNullExceptionForNullStream()
        {
            ((BlobStream) null).Exists();
        }

        [Test]
        [ExpectedException(typeof(InvalidOperationException))]
        public void ExistsThrowsInvalidOperationExceptionForStreamWithoutBlob()
        {
            new BlobStreamTest().Exists();
        }

        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void FetchAttributesThrowsArgumentNullExceptionForNullStream()
        {
            ((BlobStream)null).FetchBlobAttributes();
        }

        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void GetBlobMetadataThrowsArgumentNullExceptionForNullStream()
        {
            ((BlobStream)null).GetBlobMetadata();
        }

        [Test]
        [ExpectedException(typeof(ArgumentNullException))]
        public void GetBlobPropertiesThrowsArgumentNullExceptionForNullStream()
        {
            ((BlobStream)null).GetBlobProperties();
        }
    }
}
