﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.Data.Entity;
using System.Linq;
using DocaLabs.Storage.DataServicesUtils;
using DocaLabs.Storage.Repository;
using DocaLabs.Utils.AutoMapping;
using NUnit.Framework;

namespace DocaAzure.Storage.IntegrationTests
{
    [TestFixture]
    public class SimpleContextBaseTests
    {
        private const int EntitiesPerCategory = 1237;

        [Table(TableName)]
        public class Product : SimpleEntity
        {
            public const string TableName = "IntegrationBackupTestsProducts";

            public Product()
            {
            }

            public Product(string categoryId, string productId)
                : base(categoryId, productId)
            {
            }

            public string CustomerId { get; set; }

            public string Name { get; set; }

            public double PricePerUnit { get; set; }

            public int Quantity { get; set; }
        }

        public interface IContext : ISimpleContextBase
        {
            IDbSet<Product> Products { get; set; }
        }

        public Guid[] ProductCategories { get; set; }
        public Guid[] Customers { get; set; }
        public Guid ProductCategoryIdOf10ThEntity { get; set; }
        public Guid ProductIdOf10ThEntity { get; set; }
        public Dictionary<Guid, Product> OriginalProducts { get; set; }
        public IContext Context { get; set; }

        [SetUp]
        public void Setup()
        {
            AzureStorageAccountFactory.UseDevelopmentStorageAccount = true;
        }

        [TearDown]
        public void TearDown()
        {
            SimpleContextBase.CreateInstance<IContext>().DeleteTables();

            AzureStorageAccountFactory.UseDevelopmentStorageAccount = false;
        }

        [Test]
        public void TestQueries()
        {
            Arrange();

            AssertAllProducts();

            AssertAllProductsUsingToList();

            AssertProductsForCategory(ProductCategoryIdOf10ThEntity.ToString());

            AssertProductByWhere(ProductCategoryIdOf10ThEntity.ToString(), ProductIdOf10ThEntity.ToString());

            AssertProductByWhereWithNonGenericQueryable(ProductCategoryIdOf10ThEntity.ToString(), ProductIdOf10ThEntity.ToString());

            AssertProductByFind(ProductCategoryIdOf10ThEntity.ToString(), ProductIdOf10ThEntity.ToString());

            AssertFirstProduct();

            AssertProductNotFound();
        }

        private void Arrange()
        {
            PrepareEnvironment();
        }

        private void PrepareEnvironment()
        {
            CreateOriginalData();

            Context = SimpleContextBase.CreateInstance<IContext>();

            Context.DeleteTables();

            Context.CreateTablesIfNotExist();

            PopulateTables();
        }

        private void CreateOriginalData()
        {
            ProductCategories = new[]
                                    {
                                        new Guid("56fc6a50-d937-433f-a3d9-cd566aefdca1"),
                                        new Guid("1b5639d0-d250-4d5f-b955-8b1273dddd20"), 
                                        new Guid("fc87bb62-32c9-4279-b7c0-2d09d1a5c0e3"), 
                                    };

            Customers = new[]
                            {
                                new Guid("ae1ae27f-f26b-41cc-a657-c7547a62188f"), 
                                new Guid("83150a70-1669-477a-ac6b-63270ac6e0fe"), 
                            };

            OriginalProducts = new Dictionary<Guid, Product>();

            var random = new Random();

            var count = 0;

            foreach (var productCategory in ProductCategories)
            {
                for (var i = 0; i < EntitiesPerCategory; ++i)
                {
                    ++count;

                    var productId = Guid.NewGuid();

                    var product = new Product(productCategory.ToString(), productId.ToString())
                    {
                        CustomerId = Customers[productCategory.GetHashCode() % 2].ToString(),
                        Name = "Product" + productId,
                        PricePerUnit = random.Next(30, 99),
                        Quantity = random.Next(1, 20),
                    };

                    if (count == 10)
                    {
                        ProductCategoryIdOf10ThEntity = productCategory;
                        ProductIdOf10ThEntity = productId;
                    }

                    OriginalProducts.Add(productId, product);
                }
            }

            Console.WriteLine(@"Original data is created, {0} entities", count);
        }

        private void PopulateTables()
        {
            var count = 0;
            var batchCount = 0;

            var tmpContext = SimpleContextBase.CreateInstance<IContext>();

            string lastPartitionKey = null;

            foreach (var product in OriginalProducts.Values)
            {
                if (lastPartitionKey == null)
                    lastPartitionKey = product.PartitionKey;

                if (batchCount == 100 || lastPartitionKey != product.PartitionKey)
                {
                    batchCount = 0;
                    lastPartitionKey = product.PartitionKey;
                    tmpContext.SaveBatchWithRetries();
                    Console.WriteLine(@"Saved {0} entities", count);
                    tmpContext = SimpleContextBase.CreateInstance<IContext>();
                }

                ++count;
                ++batchCount;

                tmpContext.Products.Add(TypeMapper.Default.Map(product).ToNew<Product>());
            }

            tmpContext.SaveBatchWithRetries();
            Console.WriteLine(@"Saved {0} entities", count);
        }

        private void AssertAllProducts()
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var count = 0;

            foreach (var product in newContext.Products)
            {
                AssertProduct(product);
                ++count;
            }

            Console.WriteLine(@"Read {0} entities", count);

            Assert.AreEqual(OriginalProducts.Count, count, "number of restored products is not equal to the original");
        }

        private void AssertAllProductsUsingToList()
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var count = 0;

            foreach (var product in newContext.Products.ToList())
            {
                AssertProduct(product);
                ++count;
            }

            Console.WriteLine(@"Read {0} entities", count);

            Assert.AreEqual(OriginalProducts.Count, count, "number of restored products is not equal to the original");
        }

        private void AssertProductsForCategory(string categoryId)
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var count = 0;

            foreach (var product in newContext.Products.Where(x => x.PartitionKey == categoryId))
            {
                AssertProduct(product);
                ++count;
            }

            Console.WriteLine(@"Read {0} entities", count);

            Assert.AreEqual(EntitiesPerCategory, count, "number of restored products is not equal to the original");
        }

        private void AssertProductByWhere(string categoryId, string productId)
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var count = 0;

            foreach (var product in newContext.Products.Where(x => x.PartitionKey == categoryId && x.RowKey == productId))
            {
                AssertProduct(product);
                ++count;
            }

            Console.WriteLine(@"Read {0} entities", count);

            Assert.AreEqual(1, count, "number of restored products is not equal to the original");
        }

        private void AssertProductByWhereWithNonGenericQueryable(string categoryId, string productId)
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            AssertProduct(newContext.Products.Where(x => x.PartitionKey == categoryId && x.RowKey == productId));
        }

        private void AssertProduct(IQueryable queryable)
        {
            Assert.AreEqual(typeof (Product), queryable.ElementType);

            var enumerator = queryable.GetEnumerator();

            Assert.IsTrue(enumerator.MoveNext());

            var product = enumerator.Current as Product;

            AssertProduct(product);
        }

        private void AssertProductByFind(string categoryId, string productId)
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var product = newContext.Products.Find(categoryId, productId);

            Assert.IsNotNull(product);

            AssertProduct(product);
        }

        private void AssertFirstProduct()
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var product = newContext.Products.FirstOrDefault();

            Assert.IsNotNull(product);

            AssertProduct(product);
        }

        private static void AssertProductNotFound()
        {
            var newContext = SimpleContextBase.CreateInstance<IContext>();

            var result = newContext.Products.Where(x => x.PartitionKey == Guid.NewGuid().ToString() && x.RowKey == Guid.NewGuid().ToString()).ToArray().Count();

            Assert.AreEqual(0, result);
        }

        private void AssertProduct(Product actual)
        {
            var expected = OriginalProducts[Guid.Parse(actual.RowKey)];

            Assert.IsNotNull(expected, "couldn't find the original product");

            Assert.AreEqual(expected.PartitionKey, actual.PartitionKey);
            Assert.AreEqual(expected.CustomerId, actual.CustomerId);
            Assert.AreEqual(expected.Name, actual.Name);
            Assert.AreEqual(expected.PricePerUnit, actual.PricePerUnit, double.Epsilon);
            Assert.AreEqual(expected.Quantity, actual.Quantity);
        }
    }
}
