﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.Data;
using System.Data.Entity;
using System.Data.SqlClient;
using System.IO;
using System.Threading;
using DocaLabs.Storage.Partitioning;
using DocaLabs.Storage.StorageConfiguration;
using NUnit.Framework;

namespace DocaLabs.Storage.IntegrationTests.Partitioning
{
    public class Category
    {
        public Guid CategoryId { get; set; }
        public string Name { get; set; }
        public virtual ICollection<Product> Products { get; set; }
    }

    public class Product
    {
        public Guid ProductId { get; set; }
        public Guid CategoryId { get; set; }
        public string Name { get; set; }

        public Category Category { get; set; }
    }

    public interface IProductContext : IRelationalContextBase
    {
        IDbSet<Category> Categories { get; set; }
        IDbSet<Product> Products { get; set; }

        RelationalContextBase Context { get; }
    }

    [TestFixture]
    public class ComprehensivePartitioningTests
    {
        private const string PartitionMapConnectionString = @"Data Source=.\SQLEXPRESS;Initial Catalog=PartitionsOfIntergrationTests;Integrated Security=SSPI;";
        private const string Partition0ConnectionString = @"Data Source=.\SQLEXPRESS;Initial Catalog=DataForPartition0OfIntergrationTests;Integrated Security=SSPI;";
        private const string Partition1ConnectionString = @"Data Source=.\SQLEXPRESS;Initial Catalog=DataForPartition1OfIntergrationTests;Integrated Security=SSPI;";
        private const string Partition2ConnectionString = @"Data Source=.\SQLEXPRESS;Initial Catalog=DataForPartition2OfIntergrationTests;Integrated Security=SSPI;";

        public class TestPartitionInfo : IPartitionInfo
        {
            public Guid PartitionId { get; set; }

            public Guid GetPartitionId()
            {
                return PartitionId;
            }
        }

        [TearDown]
        public void TearDown()
        {
            new DbContext(PartitionMapConnectionString).Database.Delete();
            RelationalContextBase.CreateInstance<IProductContext>(Partition0ConnectionString).Context.Database.Delete();
            RelationalContextBase.CreateInstance<IProductContext>(Partition1ConnectionString).Context.Database.Delete();
            RelationalContextBase.CreateInstance<IProductContext>(Partition2ConnectionString).Context.Database.Delete();
        }

        [Test]
        public void PartitioningTests()
        {
            Arrange();

            var allConnectionStrings = DefaultPartitionProxy.Default.GetAllConnectionStrings();
            Assert.AreEqual(3, allConnectionStrings.Count);
            CollectionAssert.Contains(allConnectionStrings, Partition0ConnectionString);
            CollectionAssert.Contains(allConnectionStrings, Partition1ConnectionString);
            CollectionAssert.Contains(allConnectionStrings, Partition2ConnectionString);


            var map1 = new Dictionary<string, List<Guid>>();
            var map2 = new Dictionary<string, List<Guid>>();
            var map3 = new Dictionary<string, List<Guid>>();
            var map4 = new Dictionary<string, List<Guid>>();

            var startEvent = new ManualResetEvent(false);

            var thread1 = new Thread(() => AutoAssignAct(startEvent, map1)) { IsBackground = true };
            thread1.Start();
            var thread2 = new Thread(() => AutoAssignAct(startEvent, map2)) { IsBackground = true };
            thread2.Start();
            var thread3 = new Thread(() => AutoAssignAct(startEvent, map3)) { IsBackground = true };
            thread3.Start();
            var thread4 = new Thread(() => ManualAct(startEvent, map4)) { IsBackground = true };
            thread4.Start();

            Thread.Sleep(100);
            startEvent.Set();

            thread1.Join();
            thread2.Join();
            thread3.Join();
            thread4.Join();

            var part0Count = map1[Partition0ConnectionString].Count + map2[Partition0ConnectionString].Count + map3[Partition0ConnectionString].Count;
            var part1Count = map1[Partition1ConnectionString].Count + map2[Partition1ConnectionString].Count + map3[Partition1ConnectionString].Count;

            Assert.AreEqual(150, part0Count);
            Assert.AreEqual(150, part1Count);
            Assert.AreEqual(100, map4[Partition2ConnectionString].Count);

            // this method should not throw any exception for a key from manual partition
            var connectionString = DefaultPartitionProxy.Default.GetOrAutoAssignConnectionString(map4[Partition2ConnectionString][0]);
            Assert.AreEqual(Partition2ConnectionString, connectionString);
            /**/

            // Test the smart repository generator
            var categoryId = Guid.NewGuid();
            Guid key;
            if (map2[Partition1ConnectionString].Count > 0)
            {
                key = map2[Partition1ConnectionString][0];
                connectionString = Partition1ConnectionString;
            }
            else
            {
                key = map2[Partition0ConnectionString][0];
                connectionString = Partition0ConnectionString;
            }

            RelationalContextBase.PartitionInfoProvider = new TestPartitionInfo { PartitionId = key };
            using (var productContext = RelationalContextBase.SmartCreatePatitionedInstance<IProductContext>())
            {
                productContext.Categories.Add(new Category { CategoryId = categoryId, Name = "Foo" });
                productContext.SaveChanges();
            }
            using (var productContext = RelationalContextBase.CreateInstance<IProductContext>(connectionString))
            {
                var category = productContext.Categories.Find(categoryId);
                Assert.IsNotNull(category);
                Assert.AreEqual("Foo", category.Name);
            }
            /**/

            try
            {
                DefaultPartitionProxy.Default.GetConnectionString(Guid.NewGuid());
            }
            catch (PartitioningException)
            {
                try
                {
                    ((DefaultPartitionProxy)DefaultPartitionProxy.Default).AdminAddNewManualPartition(2, Partition2ConnectionString);
                }
                catch (PartitioningException)
                {
                    try
                    {
                        ((DefaultPartitionProxy)DefaultPartitionProxy.Default).AdminAddNewAutoAssignPartition(0, Partition0ConnectionString);
                    }
                    catch (PartitioningException)
                    {
                        try
                        {
                            DefaultPartitionProxy.Default.AddKeyToManualPartition(map4[Partition2ConnectionString][0], 2);
                        }
                        catch (PartitioningException)
                        {
                            try
                            {
                                DefaultPartitionProxy.Default.AddKeyToManualPartition(Guid.NewGuid(), 0);
                            }
                            catch (PartitioningException)
                            {
                                try
                                {
                                    DefaultPartitionProxy.Default.GetConnectionString(17);
                                }
                                catch (PartitioningException)
                                {
                                    return;
                                }

                                Assert.Fail("Expected the PartitioningException when trying get connection string for non defined partition");
                            }

                            Assert.Fail("Expected the PartitioningException when trying manually to add new key into auto assign partition");
                        }

                        Assert.Fail("Expected the PartitioningException when trying to add already existing key into manual partition");
                    }

                    Assert.Fail("Expected the PartitioningException when trying to add already existing auto assign partition");
                }

                Assert.Fail("Expected the PartitioningException when trying to add already existing manul partition");
            }

            Assert.Fail("Expected the PartitioningException for unmapped partition key");
        }

        private static void AutoAssignAct(WaitHandle startEvent, IDictionary<string, List<Guid>> map)
        {
            map.Add(Partition0ConnectionString, new List<Guid>());
            map.Add(Partition1ConnectionString, new List<Guid>());

            startEvent.WaitOne();
            for (var i = 0; i < 100; i++)
            {
                var key = Guid.NewGuid();
                map[DefaultPartitionProxy.Default.GetOrAutoAssignConnectionString(key)].Add(key);
            }
        }

        private static void ManualAct(WaitHandle startEvent, IDictionary<string, List<Guid>> map)
        {
            map.Add(Partition2ConnectionString, new List<Guid>());

            startEvent.WaitOne();
            for (var i = 0; i < 100; i++)
            {
                var key = Guid.NewGuid();
                map[DefaultPartitionProxy.Default.AddKeyToManualPartition(key, 2)].Add(key);
            }
        }

        private static void Arrange()
        {
            CreatePartitionDatabase();

            CreatePartition(Partition0ConnectionString);
            CreatePartition(Partition1ConnectionString);
            CreatePartition(Partition2ConnectionString);

            ((DefaultPartitionProxy)DefaultPartitionProxy.Default).AdminAddNewAutoAssignPartition(0, Partition0ConnectionString);
            ((DefaultPartitionProxy)DefaultPartitionProxy.Default).AdminAddNewAutoAssignPartition(1, Partition1ConnectionString);
            ((DefaultPartitionProxy)DefaultPartitionProxy.Default).AdminAddNewManualPartition(2, Partition2ConnectionString);
        }

        private static void CreatePartitionDatabase()
        {
            var db = new DbContext(PartitionMapConnectionString).Database;

            // drop the database if exists
            db.Delete();

            // create new database
            db.Create();

            // create tables & sprocs
            foreach (var part in FlatFileToSqlCommands(File.ReadAllText(@"partitioning-create-tables-and-sps.sql")))
            {
                ExecuteNonQuery(DocaStorageConfiguration.GetDefaultSection().PartitionMapConnection, part);                
            }
        }

        private static void CreatePartition(string connectionString)
        {
            var db = RelationalContextBase.CreateInstance<IProductContext>(connectionString).Context.Database;

            // drop the database if exists
            db.Delete();

            // create new database and tables
            db.Create();
        }

        private static IEnumerable<string> FlatFileToSqlCommands(string flatFile)
        {
            return flatFile.Split(new [] {"\r\nGO\r\n"}, StringSplitOptions.RemoveEmptyEntries);
        }

        private static void ExecuteNonQuery(string connectionString, string query)
        {
            using (var connection = new SqlConnection(connectionString))
            using (var command = new SqlCommand(query, connection) { CommandType = CommandType.Text })
            {
                command.Connection.Open();
                command.ExecuteNonQuery();
            }
        }
    }
}
