﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Data.Entity;
using System.Linq;
using DocaLabs.Storage.DataServicesUtils;
using DocaLabs.Storage.IntegrationTests.TestUtils;
using DocaLabs.Storage.Repository;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using NUnit.Framework;

namespace DocaLabs.Storage.IntegrationTests.Repository
{
    [TestFixture]
    public class RepositoryGeneratorTests
    {
        public class AzureContextBaseTestExtension : AzureContextBase
        {
            public string ReturnPassedText(string text, int value)
            {
                return text + value;
            }

            public string ReturnHelloWorld()
            {
                return "Hello World!";
            }
        }

        public class AzureTestEntity42 : SimpleEntity
        {
            public AzureTestEntity42()
            {
            }

            public AzureTestEntity42(string partitionKey, string rowKey)
                : base(partitionKey, rowKey)
            {
            }

            public string Property1 { get; set; }
        }

        public class AzureTestEntity44 : SimpleEntity
        {
            public AzureTestEntity44()
            {
            }

            public AzureTestEntity44(string partitionKey, string rowKey)
                : base(partitionKey, rowKey)
            {
            }

            public string Property2 { get; set; }
        }

        public interface IAzureTestRepository : IDisposable
        {
            IDbSet<AzureTestEntity42> TestEntities42 { get; set; }

            AzureContextBaseTestExtension Context { get; }

            void SaveChanges();
            void MarkAsModified(object entity);

            string ReturnPassedText(string text, int value);
            string ReturnHelloWorld();
        }

        public interface IAzureTestRepository2 : IAzureTestRepository
        {
            IDbSet<AzureTestEntity44> TestEntities44 { get; set; }
        }
        
        [Test]
        public void GeneratesFullyFunctionalAzureTableRepositoryWithCustomMethods()
        {
            var target1 = RepositoryGenerator.CreateInstance<IAzureTestRepository2>(typeof(AzureContextBaseTestExtension));

            Assert.AreEqual("My value in text is 42", target1.ReturnPassedText("My value in text is ", 42));
            Assert.AreEqual("Hello World!", target1.ReturnHelloWorld());

            // in order to start from controllable state in case if previous test were aborted
            target1.Context.DropTables();

            var tableClient = CloudStorageAccount.DevelopmentStorageAccount.CreateCloudTableClient();

            try
            {
                target1.Context.CreateTablesIfNotExist();

                Assert.IsTrue(tableClient.DoesTableExist("AzureTestEntity42"));
                Assert.IsTrue(tableClient.DoesTableExist("AzureTestEntity44"));

                target1.TestEntities42.Add(new AzureTestEntity42("part1", "row1") { Property1 = "42" });
                target1.TestEntities42.Add(new AzureTestEntity42("part1", "row2") { Property1 = "33" });
                target1.TestEntities44.Add(new AzureTestEntity44("part44", "row44") { Property2 = "44" });
                target1.SaveChanges();

                var target2 = RepositoryGenerator.CreateInstance<IAzureTestRepository2>(typeof(AzureContextBaseTestExtension));
                var entity = target2.TestEntities42.Where(x => x.PartitionKey == "part1" && x.RowKey == "row2").FirstOrDefault();
                Assert.IsNotNull(entity);
                Assert.AreEqual("part1", entity.PartitionKey);
                Assert.AreEqual("row2", entity.RowKey);
                Assert.AreEqual("33", entity.Property1);
                entity.Property1 = "77";
                target2.MarkAsModified(entity);
                target2.SaveChanges();

                var target3 = RepositoryGenerator.CreateInstance<IAzureTestRepository2>(typeof(AzureContextBaseTestExtension));
                entity = target3.TestEntities42.Where(x => x.PartitionKey == "part1" && x.RowKey == "row2").FirstOrDefault();
                Assert.IsNotNull(entity);
                Assert.AreEqual("part1", entity.PartitionKey);
                Assert.AreEqual("row2", entity.RowKey);
                Assert.AreEqual("77", entity.Property1);

                var entity44 = target3.TestEntities44.Find("part44",  "row44");
                Assert.IsNotNull(entity44);
                Assert.AreEqual("part44", entity44.PartitionKey);
                Assert.AreEqual("row44", entity44.RowKey);
                Assert.AreEqual("44", entity44.Property2);


                Assert.IsNotNull(target3.Context);
                target3.Dispose();
                Assert.IsNull(target3.Context);
            }
            finally
            {
                target1.Context.DropTables();
            }

            Assert.IsFalse(tableClient.DoesTableExist("AzureTestEntity42"));
        }
    }
}
