﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaAzure.Utils;
using Microsoft.WindowsAzure.StorageClient;

namespace DocaAzure.Backup
{
    public class BlobContainerBackupToLocalFileSystem
    {
        /// <summary>
        /// Set this property to false if you don't expect any page blobs in the target container to minimize storega transaction count.
        /// There is an extra request to determine type of the blob if this property is set to true.
        /// The default value is true.
        /// </summary>
        public bool ExpectPageBlobs { get; set; }

        public BlobContainerBackupToLocalFileSystem()
        {
            ExpectPageBlobs = true;
        }

        /// <summary>
        /// Backs up whole blob container to specified local folder.
        /// </summary>
        /// <remarks>
        /// It's up to a caller to delete whatever files are already in the destination folder before starting backup.
        /// Not doing so may cause
        /// </remarks>
        /// <param name="sourceContainerName">Container to be backed up</param>
        /// <param name="destinationBaseFolder">Local root folder</param>
        /// <param name="onlyIfModifiedSince"></param>
        /// <param name="maxDegreeOfParallelism"></param>
        public void Backup(string sourceContainerName, string destinationBaseFolder, DateTime? onlyIfModifiedSince, int maxDegreeOfParallelism = 1)
        {
            var provider = new BlobStorageProvider();

            var filter = onlyIfModifiedSince == null
                    ? null
                    : BlobStorageProvider.IfModifiedSince(onlyIfModifiedSince.GetValueOrDefault());

            Parallel.ForEach(
                        provider.FlatBlobList(sourceContainerName, null, filter),
                        new ParallelOptions { MaxDegreeOfParallelism = maxDegreeOfParallelism },
                        sourceBlobName => BackupBlob(sourceBlobName, destinationBaseFolder));
        }

        /// <summary>
        /// Downloads the blob to the file
        /// </summary>
        /// <remarks>
        /// The file name must conform to the normal file name cosntrains. The length must be 256
        /// as the method create side car file with metadata, the extra four charaters are required
        /// for the this file extension
        /// </remarks>
        /// <param name="sourceBlobName">Blob's name, must include the container name as well</param>
        /// <param name="destinationFileName">File name where to download</param>
        public void DownloadBlob(string sourceBlobName, string destinationFileName)
        {
            // ReSharper disable AssignNullToNotNullAttribute
            Directory.CreateDirectory(Path.GetDirectoryName(destinationFileName));
            // ReSharper restore AssignNullToNotNullAttribute

            using (var backupStream = new FileStream(destinationFileName, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                var sourceBlob = AzureStorageAccountFactory.GetBlobReference(sourceBlobName);

                if (ExpectPageBlobs)
                {
                    sourceBlob.FetchAttributes();

                    if (sourceBlob.Properties.BlobType == BlobType.BlockBlob)
                        DownloadBlockBlob(destinationFileName, sourceBlob.ToBlockBlob, backupStream);
                    else
                        DownloadPageBlob(destinationFileName, sourceBlob.ToPageBlob, backupStream);
                }
                else
                {
                    DownloadBlockBlob(destinationFileName, sourceBlob.ToBlockBlob, backupStream);
                }
            }
        }

        private void BackupBlob(string sourceBlobName, string destinationBaseFolder)
        {
            string metadataFileName;

            var idx = sourceBlobName[0] != '/' ? sourceBlobName.IndexOf("/") : sourceBlobName.IndexOf("/", 1);
            if (idx <= 0)
                throw new ArgumentException(string.Format(Resources.Text.no_container_in_blob_name, sourceBlobName), "sourceBlobName");

            var sourceContainerName = sourceBlobName.Substring(0, idx);

            try
            {
                metadataFileName = Path.GetFullPath(Path.Combine(destinationBaseFolder, sourceBlobName + FileBackupBlobInfo.MetaFileExtension));
            }
            catch (ArgumentException)
            {
                // could be if the name contains invalid characters for the file name
                metadataFileName = FileBackupBlobInfo.MakeSpecialMetaFileNameFor(destinationBaseFolder, sourceContainerName);
            }
            catch (PathTooLongException)
            {
                // could be if the name is too long for the file name
                metadataFileName = FileBackupBlobInfo.MakeSpecialMetaFileNameFor(destinationBaseFolder, sourceContainerName);
            }

            DownloadBlob(sourceBlobName, FileBackupBlobInfo.MetaFileNameToContent(metadataFileName));
        }

        private static void DownloadBlockBlob(string destinationFileName, CloudBlockBlob blob, Stream destinationStream)
        {
            blob.DownloadToStream(destinationStream);

            FileBackupBlobInfo.Save(destinationFileName, blob, null);
        }

        private static void DownloadPageBlob(string destinationFileName, CloudPageBlob blob, Stream destinationStream)
        {
            var pageRanges = blob.GetPageRanges().ToArray();

            using (var blobStream = blob.OpenRead())
            {
                foreach (var range in pageRanges)
                {
                    // EndOffset is inclusive... so need to add 1
                    var rangeSize = (int)(range.EndOffset + 1 - range.StartOffset);

                    blobStream.CopyRangeTo(range.StartOffset, rangeSize, destinationStream);
                }
            }

            FileBackupBlobInfo.Save(destinationFileName, blob, pageRanges);
        }
    }
}
