﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.IO;
using System.Threading.Tasks;
using DocaAzure.Backup.Impl;
using DocaAzure.Storage;
using DocaAzure.Utils;
using DocaAzure.Utils.Serialization;
using Microsoft.WindowsAzure.StorageClient;
using NLog;

namespace DocaAzure.Backup
{
    public static class BlobContainerRestoreFromLocalFileSystem
    {
        private static readonly Logger Log = LogHelper.GetCurrentAssemblyLogger();

        public static void Restore(string originalContainer, string baseFolder, int maxDegreeOfParallelism = 1)
        {
            var searchPath = Path.Combine(baseFolder, originalContainer);

            Parallel.ForEach(
                        Directory.EnumerateFiles(searchPath, "*" + FileBackupBlobInfo.MetaFileExtension, SearchOption.AllDirectories),
                        new ParallelOptions { MaxDegreeOfParallelism = maxDegreeOfParallelism },
                        RestoreBlob);
        }

        /// <summary>
        /// Restore the blob
        /// </summary>
        /// <param name="metaFileName"></param>
        public static void RestoreBlob(string metaFileName)
        {
            using (var stream = new FileStream(metaFileName, FileMode.Open, FileAccess.Read, FileShare.None))
            {
                FileBackupBlobInfo info;

                try
                {
                    info = stream.Deserialize<FileBackupBlobInfo>();
                    if (!info.VerifyMarker())
                    {
                        Log.Warn(Resources.Text.wrong_metafile_format, info.BlobName);
                        return;
                    }
                }
                catch(Exception e)
                {
                    Log.ErrorException(string.Format(Resources.Text.fails_restore_blob_for_metafile, metaFileName), e);
                    return;
                }

                UploadBlob(info, FileBackupBlobInfo.MetaFileNameToContent(metaFileName));
            }
        }

        private static void UploadBlob(FileBackupBlobInfo info, string fileName)
        {
            using (var sourceStream = new FileStream(fileName, FileMode.Open, FileAccess.Read, FileShare.None))
            {
                CloudBlob blob;

                try
                {
                    blob = UploadBlobContent(info, sourceStream);
                }
                catch (StorageClientException e)
                {
                    if (!BlobStorageProvider.IsResourceNotFoundError(e.ErrorCode))
                        throw;

                    var idx = info.BlobName.IndexOf("/");
                    if (idx < 0)
                        throw;

                    var container = info.BlobName.Substring(0, idx);

                    AzureStorageAccountFactory.GetContainerReference(container).CreateIfNotExist();

                    blob = UploadBlobContent(info, sourceStream.ResetToBegining());
                }

                CopyProperties(info, blob);

                CopyMetadata(info, blob);
            }
        }
    
        private static CloudBlob UploadBlobContent(FileBackupBlobInfo info, Stream sourceStream)
        {
            return info.BlobType == BlobType.PageBlob
                            ? UploadPageBlob(info, sourceStream)
                            : UploadBlockBlob(info, sourceStream);
        }

        private static CloudBlob UploadBlockBlob(FileBackupBlobInfo info, Stream sourceStream)
        {
            var blob = AzureStorageAccountFactory.GetBlockBlobReference(info.BlobName);

            blob.UploadFromStream(sourceStream);

            return blob;
        }

        private static CloudBlob UploadPageBlob(FileBackupBlobInfo info, Stream sourceStream)
        {
            if (info.PageRanges == null)
                throw new InvalidOperationException(string.Format(Resources.Text.page_ranges_are_null, info.BlobName));

            var blob = AzureStorageAccountFactory.GetPageBlobReference(info.BlobName);

            blob.Create(info.BlobSize);

            foreach (var range in info.PageRanges)
            {
                // EndOffset is inclusive... so need to add 1
                var rangeSize = (int)(range.EndOffset + 1 - range.StartOffset);

                var buffer = new byte[rangeSize];
                sourceStream.Read(buffer, 0, rangeSize);

                using (var wrapperStream = new MemoryStream(buffer))
                {
                    blob.WritePages(wrapperStream, range.StartOffset);
                }
            }

            return blob;
        }

        private static void CopyProperties(FileBackupBlobInfo info, CloudBlob destinationBlob)
        {
            if (destinationBlob.Properties.CacheControl == info.CacheControl
                && destinationBlob.Properties.ContentEncoding == info.ContentEncoding
                && destinationBlob.Properties.ContentLanguage == info.ContentLanguage
                && (
                     (destinationBlob.Properties.ContentType == info.ContentType)
                   || (destinationBlob.Properties.ContentType == null && info.ContentType == "application/octet-stream")))
            {
                return;
            }

            destinationBlob.Properties.CacheControl = info.CacheControl;
            destinationBlob.Properties.ContentEncoding = info.ContentEncoding;
            destinationBlob.Properties.ContentLanguage = info.ContentLanguage;
            destinationBlob.Properties.ContentType = info.ContentType;

            destinationBlob.SetProperties();
        }

        private static void CopyMetadata(FileBackupBlobInfo info, CloudBlob destinationBlob)
        {
            if (info.Metadata == null || info.Metadata.Count == 0)
                return;

            destinationBlob.Metadata.Add(info.Metadata);

            destinationBlob.SetMetadata();
        }
    }
}
