﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using DocaAzure.Backup.Impl;
using DocaAzure.Backup.Interfaces;
using DocaAzure.Backup.Service;
using DocaAzure.Storage;
using DocaAzure.Utils;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.ServiceRuntime;

namespace DocaAzure.Backup
{
    public class DefaultBackupServiceFactory : IBackupServiceFactory
    {
        public const string BackupAccountConfigirationKey = "DefaultBackupServiceFactory.GetStorageAccount.ConnectionString";

        private static IBackupServiceFactory _default;
        public static IBackupServiceFactory Default
        {
            get { return _default ?? NestedBackupAccountProvider.NestedProvider; }
            set { _default = value; }
        }

        public static Type BufferType { get; set; }

        public static Type BackupProgressMonitorType { get; set; }

        public static Type RestoreProgressMonitorType { get; set; }

        public static Type BackupStorageFactoryType { get; set; }

        static DefaultBackupServiceFactory()
        {
            BufferType = typeof(SharpZipLibBuffer);
            BackupProgressMonitorType = typeof (BackupDefaultProgressMonitor);
            RestoreProgressMonitorType = typeof(RestoreDefaultProgressMonitor);
            BackupStorageFactoryType = typeof(AzureBackupStorageFactory);
        }

        public virtual CloudStorageAccount GetStorageAccount()
        {
            if (!RoleEnvironment.IsAvailable)
                return AzureStorageAccountFactory.Create();

            try
            {
                var backupAccount = DefaultRoleEnvironmentProvider.Default.GetConfigurationSettingValue(BackupAccountConfigirationKey);

                return string.IsNullOrWhiteSpace(backupAccount)
                    ? AzureStorageAccountFactory.Create()
                    : CloudStorageAccount.Parse(backupAccount);
            }
            catch
            {
                return AzureStorageAccountFactory.Create();
            }
        }

        public virtual IBuffer CreateBuffer()
        {
            return Activator.CreateInstance(BufferType) as IBuffer;
        }

        public virtual IBackupProgressMonitor CreateBackupProgressMonitor()
        {
            return Activator.CreateInstance(BackupProgressMonitorType) as IBackupProgressMonitor;
        }

        public virtual IRestoreProgressMonitor CreateRestoreProgressMonitor()
        {
            return Activator.CreateInstance(RestoreProgressMonitorType) as IRestoreProgressMonitor;
        }

        public virtual IBackupStorageFactory GetBackupStorageFactory()
        {
            return Activator.CreateInstance(BackupStorageFactoryType) as IBackupStorageFactory;
        }

        public static Func<BackupEntity, bool> IfNotModifiedSince(DateTime lastModifiedUtc)
        {
            return entity => entity.Timestamp < lastModifiedUtc;
        }

        public static Func<BackupEntity, bool> IfModifiedSince(DateTime lastModifiedUtc)
        {
            return entity => entity.Timestamp >= lastModifiedUtc;
        }

        public static Func<BackupEntity, bool> All()
        {
            return entity => true;
        }

        static class NestedBackupAccountProvider
        {
            internal static IBackupServiceFactory NestedProvider { get; private set; }

            static NestedBackupAccountProvider()
            {
                NestedProvider = new DefaultBackupServiceFactory();
            }
        }
    }
}
