﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using DocaAzure.Backup.Interfaces;
using DocaAzure.Storage;
using Microsoft.WindowsAzure.StorageClient;
using Microsoft.WindowsAzure.StorageClient.Protocol;

namespace DocaAzure.Backup.Impl
{
    public class AzureBackupStorageFactory : IBackupStorageFactory
    {
        public ITableBackupWriter GetTableBackupWriter(string blobName)
        {
            return new StorageTableBackupWriter(blobName);
        }

        public ITableRestoreReader GetTableRestoreReader(string blobName)
        {
            return new StorageTableRestoreReader(blobName);
        }

        public IEnumerable<string> GetTableBlobs(string backupContainerName, string tablePrefix)
        {
            return new BlobStorageProvider(
                DefaultBackupServiceFactory.Default.GetStorageAccount()).FlatBlobList(backupContainerName, tablePrefix);
        }

        public IEnumerable<string> GetTables(string backupContainerName)
        {
            return GetSubfoldersOnRoot(backupContainerName);
        }

        public IEnumerable<string> GetOriginalContainers(string backupContainerName)
        {
            return GetSubfoldersOnRoot(backupContainerName);
        }

        protected IEnumerable<string> GetSubfoldersOnRoot(string backupContainerName)
        {
            var provider = new BlobStorageProvider(DefaultBackupServiceFactory.Default.GetStorageAccount());

            var blobEndpointUrl = provider.BlobClient.BaseUri.ToString();

            var endPointPrefixLength = blobEndpointUrl.Length;
            if (!blobEndpointUrl.EndsWith("/"))
                endPointPrefixLength++;

            endPointPrefixLength += backupContainerName.Length;

            if (!backupContainerName.EndsWith("/"))
                endPointPrefixLength++;

            var tables = new List<string>();

            // ReSharper disable LoopCanBeConvertedToQuery
            foreach (var item in provider.ListBlobs(backupContainerName))
            // ReSharper restore LoopCanBeConvertedToQuery
            {
                if (!(item is CloudBlobDirectory))
                    continue;

                var uri = item.Uri.ToString();

                tables.Add(uri.EndsWith("/")
                               ? uri.Substring(endPointPrefixLength, uri.Length - endPointPrefixLength - 1)
                               : uri.Substring(endPointPrefixLength));
            }

            return tables;
        }

        public class StorageTableBackupWriter : ITableBackupWriter
        {
            public string BlobName { get; private set; }

            public CloudBlockBlob BackupBlob { get; private set; }

            public BlobRequestOptions RequestOptions { get; private set; }

            private List<string> BlockIdList { get; set; }

            public StorageTableBackupWriter(string blobName)
            {
                BlobName = blobName;

                BlockIdList = new List<string>();

                RequestOptions = new BlobRequestOptions
                {
                    RetryPolicy = RetryPolicies.RetryExponential(5, RetryPolicies.DefaultClientBackoff)
                };
            }

            public void AddBlock(string blockId, Stream dataStream)
            {
                if(BackupBlob == null)
                {
                    BackupBlob = DefaultBackupServiceFactory.Default.GetStorageAccount().CreateCloudBlobClient().GetBlockBlobReference(BlobName);

                    BlobStorageProvider.CreateContainer(BackupBlob.Container, TimeSpan.FromHours(2));
                }

                BackupBlob.PutBlock(blockId, dataStream, null, RequestOptions);

                BlockIdList.Add(blockId);
            }

            public void FinalizeBlob()
            {
                if (BackupBlob == null)
                    return;

                if (BlockIdList.Count > 0)  // commit block list
                    BackupBlob.PutBlockList(BlockIdList, RequestOptions);
            }


            public void Dispose()
            {
                // no need to do anything
            }
        }

        public class StorageTableRestoreReader : ITableRestoreReader
        {
            public CloudBlockBlob RestoreBlob { get; private set; }

            public BlobRequestOptions RequestOptions { get; private set; }

            public StorageTableRestoreReader(string blobName)
            {
                RestoreBlob = DefaultBackupServiceFactory.Default.GetStorageAccount().CreateCloudBlobClient().GetBlockBlobReference(blobName);

                RequestOptions = new BlobRequestOptions
                {
                    RetryPolicy = RetryPolicies.RetryExponential(5, RetryPolicies.DefaultClientBackoff)
                };
            }

            /// <summary>
            /// Enumerates through blocks stored in the blob
            /// </summary>
            /// <returns>
            /// Stream for each block, the stream must be closed or disposed.
            /// Failure to close the stream will cause your application to run out of connections.
            /// </returns>
            public IEnumerable<Stream> GetBlocks()
            {
                // get all blocks and for each block read it separately as it is an xml doc by itself
                var blocks = RestoreBlob.DownloadBlockList(BlockListingFilter.Committed, RequestOptions);

                long currentOffset = 0;

                foreach (var block in blocks)
                {
                    // read each block using the range
                    var request = BlobRequest.Get(RestoreBlob.Uri, 120, null, currentOffset, block.Size, null);

                    currentOffset += block.Size;

                    RestoreBlob.ServiceClient.Credentials.SignRequest(request);

                    yield return request.GetResponse().GetResponseStream();
                }
            }

            public void Dispose()
            {
                // no need to do anything
            }
        }
    }
}
