﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System.Collections.Generic;

namespace DocaAzure.Backup.Impl
{
    /// <summary>
    /// Represents a collection of entities in a single batch
    /// </summary>
    public class Batch
    {
        private const int MaxEntityCount = 100;

        // Save at most 3.0MB in a batch so that we have enough room for 
        // the xml tags that WCF Data Services adds in the OData protocol
        private const int MaxBatchSize = (int)(3.5 * 1024 * 1024);

        private readonly State _state;

        public Batch(State state)
        {
            _state = state;
        }

        /// <summary>
        /// Yield entities until we hit a condition that should terminate a batch.
        /// The conditions to terminate on are:
        /// 1. 100 entities in a batch
        /// 2. 3.5MB of data
        /// 2. Max of block size
        /// 3. We see a new partition key
        /// </summary>
        public IEnumerable<BackupEntity> Entities
        {
            get
            {
                var batchSize = 0;
                var entityCount = 0;
                BackupEntity entity;
                string lastPartitionKeySeen = null;

                while ((entity = _state.GetNextEntity()) != null)
                {
                    if (lastPartitionKeySeen == null)
                    {
                        lastPartitionKeySeen = entity.PartitionKey;
                    }

                    var approxEntitySize = entity.GetApproxEntitySize();

                    if (entityCount >= MaxEntityCount
                        || !string.Equals(entity.PartitionKey, lastPartitionKeySeen)
                        || batchSize + approxEntitySize > MaxBatchSize
                        || _state.CurrentBlockSize + approxEntitySize > _state.MaxBlockSize)
                    {
                        // set this current entity as the look ahead since it needs to be part of the next batch
                        _state.LookAheadEntity = entity;

                        yield break;
                    }

                    entityCount++;
                    batchSize += approxEntitySize;

                    yield return entity;
                }
            }
        }
    }
}
