﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.IO;
using DocaAzure.Storage;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;

namespace DocaAzure.Backup.Impl
{
    public abstract class BlobContainerBackupRestoreBase
    {
        protected Func<string, string> MakeDestinationBlobName { get; set; }

        /// <summary>
        /// Set this property to false if you don't expect any page blobs in the sourcce container 
        /// when copying to different account in order to minimize storega transaction count.
        /// There is an extra request to determine type of the blob if this property is set to true.
        /// The default value is true.
        /// </summary>
        public bool ExpectPageBlobs { get; set; }

        protected BlobContainerBackupRestoreBase()
        {
            ExpectPageBlobs = true;
        }

        protected virtual bool IsTheSameAccount()
        {
            var sourceAccount = GetSourceAccount();
            var destinationAccount = GetDestinationAccount();

            return sourceAccount.Credentials.AccountName == destinationAccount.Credentials.AccountName;
        }

        protected Action<string> GetCopyAction()
        {
            return IsTheSameAccount() ? (Action<string>) (CopyInSameAccount) : (CopyToDifferentAccount);
        }

        private void CopyInSameAccount(string sourceBlobName)
        {
            ((IBlobOperations)new BlobStorageProvider()).CopyBlob(sourceBlobName, MakeDestinationBlobName(sourceBlobName));
        }

        private void CopyToDifferentAccount(string sourceBlobName)
        {
            var sourceBlob = GetSourceAccount().CreateCloudBlobClient().GetBlobReference(sourceBlobName);

            if (ExpectPageBlobs)
                sourceBlob.FetchAttributes();

            // copy blob content
            var destinationBlob = ExpectPageBlobs && sourceBlob.Properties.BlobType == BlobType.PageBlob 
                ? CopyPageBlob(sourceBlob.ToPageBlob, MakeDestinationBlobName(sourceBlobName)) 
                : CopyBlockBlob(sourceBlob.ToBlockBlob, MakeDestinationBlobName(sourceBlobName));

            CopyProperties(sourceBlob, destinationBlob);

            CopyMetadata(sourceBlob, destinationBlob);
        }

        private CloudBlob CopyBlockBlob(CloudBlockBlob sourceBlob, string destinationBlobName)
        {
            var destinationBlob = GetDestinationAccount().CreateCloudBlobClient().GetBlockBlobReference(destinationBlobName);

            using (var destinationStream = destinationBlob.OpenWrite())
            {
                sourceBlob.DownloadToStream(destinationStream);   
            }

            return destinationBlob;
        }

        private CloudBlob CopyPageBlob(CloudPageBlob sourceBlob, string destinationBlobName)
        {
            CloudPageBlob destinationBlob = null;

            var pageRanges = sourceBlob.GetPageRanges();

            using (var sourceStream = sourceBlob.OpenRead())
            {
                foreach (var range in pageRanges)
                {
                    // EndOffset is inclusive... so need to add 1
                    var rangeSize = (int)(range.EndOffset + 1 - range.StartOffset);

                    // Seek to the correct starting offset in the page blob stream
                    sourceStream.Seek(range.StartOffset, SeekOrigin.Begin);

                    // Read the next page range into the buffer
                    var buffer = new byte[rangeSize];
                    sourceStream.Read(buffer, 0, rangeSize);

                    // Then write the buffer
                    if(destinationBlob == null)
                    {
                        // only at this stage the source size is known
                        destinationBlob = GetDestinationAccount().CreateCloudBlobClient().GetPageBlobReference(destinationBlobName);
                        destinationBlob.Create(sourceBlob.Properties.Length);
                    }

                    using (var wrapperStream = new MemoryStream(buffer))
                    {
                        destinationBlob.WritePages(wrapperStream, range.StartOffset);    
                    }
                }
            }

            return destinationBlob;
        }

        private static void CopyProperties(CloudBlob sourceBlob, CloudBlob destinationBlob)
        {
            if( destinationBlob.Properties.CacheControl == sourceBlob.Properties.CacheControl
                && destinationBlob.Properties.ContentEncoding == sourceBlob.Properties.ContentEncoding
                && destinationBlob.Properties.ContentLanguage == sourceBlob.Properties.ContentLanguage
                && (
                     (destinationBlob.Properties.ContentType == sourceBlob.Properties.ContentType) 
                   || (destinationBlob.Properties.ContentType == null && sourceBlob.Properties.ContentType == "application/octet-stream")))
            {
                return;
            }

            destinationBlob.Properties.CacheControl = sourceBlob.Properties.CacheControl;
            destinationBlob.Properties.ContentEncoding = sourceBlob.Properties.ContentEncoding;
            destinationBlob.Properties.ContentLanguage = sourceBlob.Properties.ContentLanguage;
            destinationBlob.Properties.ContentType = sourceBlob.Properties.ContentType;

            destinationBlob.SetProperties();
        }

        private static void CopyMetadata(CloudBlob sourceBlob, CloudBlob destinationBlob)
        {
            if (sourceBlob.Metadata.Count == 0)
                return;

            destinationBlob.Metadata.Add(sourceBlob.Metadata);

            destinationBlob.SetMetadata();
        }

        protected abstract CloudStorageAccount GetSourceAccount();
        protected abstract CloudStorageAccount GetDestinationAccount();
    }
}
