﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.IO;

namespace DocaAzure.Backup.Impl
{
    public class CountingZipStreamWrapper : Stream
    {
        public const double ApproxLengthSinceLastFlushSafetyFactor = 1.05;

        /// <summary>
        /// Wrapped stream
        /// </summary>
        public Stream InnerStream { get; private set; }

        /// <summary>
        /// Base number of bytes in after which to flush
        /// </summary>
        public long FlushAfterEachBytes { get; private set; }

        /// <summary>
        /// Some streams cannot flush so all extended functionality would be off
        /// </summary>
        public bool CanFlush { get; set; }

        /// <summary>
        /// Overall number of written bytes
        /// </summary>
        public long BytesIn { get; private set; }

        /// <summary>
        /// Number of bytes written since last flaush
        /// </summary>
        public long BytesInSinceLastFlush { get; private set; }

        /// <summary>
        /// Current compression ration (at the begining assumes that is 1)
        /// </summary>
        public double CurrentCompressionRatio { get; private set; }

        /// <summary>
        /// Delegate that is able to return current inner stream supposed size
        /// (DotNetZip - it's Position, for SharpZipLib it's either Length or Position)
        /// </summary>
        public Func<Stream, long> ApproxLength { get; private set; }

        /// <summary>
        /// Approximate size of compressed data since last flush
        /// </summary>
        public long ApproxLengthSinceLastFlush
        {
            get { return (long)(BytesInSinceLastFlush * CurrentCompressionRatio * ApproxLengthSinceLastFlushSafetyFactor); }
        }

        public CountingZipStreamWrapper(Stream innerStream, long flushAfterEach, Func<Stream, long> aproxLength)
        {
            InnerStream = innerStream;
            FlushAfterEachBytes = flushAfterEach;
            ApproxLength = aproxLength;
            CurrentCompressionRatio = 1.0;

            // assume that the stream can flush, it can be overriden by the caller
            CanFlush = true;
        }

        public override void Flush()
        {
            if(!CanFlush)
                return;

            try
            {
                InnerStream.Flush();

                BytesInSinceLastFlush = 0;

                CurrentCompressionRatio = ((double)ApproxLength(InnerStream)) / BytesIn;
            }
            catch
            {
                CanFlush = false;
            }
        }

        public override long Seek(long offset, SeekOrigin origin)
        {
            return InnerStream.Seek(offset, origin);
        }

        public override void SetLength(long value)
        {
            InnerStream.SetLength(value);
        }

        public override int Read(byte[] buffer, int offset, int count)
        {
            return InnerStream.Read(buffer, offset, count);
        }

        public override void Write(byte[] buffer, int offset, int count)
        {
            BytesIn += count;
            BytesInSinceLastFlush += count;

            InnerStream.Write(buffer, offset, count);

            if ((!CanFlush) || (ApproxLength == null) || (BytesIn <= 0))
                return;

            if (BytesInSinceLastFlush < FlushAfterEachBytes)
                return;

            Flush();
        }

        public override bool CanRead
        {
            get { return InnerStream.CanRead; }
        }

        public override bool CanSeek
        {
            get { return InnerStream.CanSeek; }
        }

        public override bool CanWrite
        {
            get { return InnerStream.CanWrite; }
        }

        public override long Length
        {
            get { return InnerStream.Length; }
        }

        public override long Position
        {
            get { return InnerStream.Position; }
            set { InnerStream.Position = value; }
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing && InnerStream != null)
            {
                InnerStream.Dispose();
                InnerStream = null;
            }

            base.Dispose(disposing);
        }
    }
}
