﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Specialized;
using System.IO;
using DocaAzure.Storage;
using DocaAzure.Utils.Serialization;
using Microsoft.WindowsAzure.StorageClient;

namespace DocaAzure.Backup.Impl
{
    public class FileBackupBlobInfo
    {
        public class SerializablePageRange
        {
            public long StartOffset { get; set; }
            public long EndOffset { get; set; }
        }

        public const string MetaFileExtension = ".$--";
        public const string SpecialSubfolder = ".$special--";

        public string Marker { get; set; }
        public string BlobName { get; set; }
        public BlobType BlobType { get; set; }
        public string CacheControl { get; set; }
        public string ContentEncoding { get; set; }
        public string ContentLanguage { get; set; }
        public string ContentType { get; set; }
        public NameValueCollection Metadata { get; set; }
        public long BlobSize { get; set; }
        public SerializablePageRange[] PageRanges { get; set; }

        public bool VerifyMarker()
        {
            return Marker == MetaFileExtension;
        }

        public static string MakeSpecialMetaFileNameFor(string destinationBaseFolder, string sourceContainerName)
        {
            return Path.GetFullPath(
                Path.Combine(destinationBaseFolder, sourceContainerName, SpecialSubfolder, Guid.NewGuid() + MetaFileExtension));
        }

        public static string MetaFileNameToContent(string metaFileName)
        {
            return metaFileName.Substring(0, metaFileName.Length - MetaFileExtension.Length);
        }

        public static void Save(string contentFileName, CloudBlob blob, PageRange[] pageRanges)
        {
            using (var metadataStream = new FileStream(contentFileName + MetaFileExtension, FileMode.Create, FileAccess.Write, FileShare.None))
            {
                var info = new FileBackupBlobInfo
                {
                    Marker = MetaFileExtension,
                    BlobName = BlobStorageProvider.ExtractBlobName(blob),
                    BlobType = blob.Properties.BlobType,
                    CacheControl = blob.Properties.CacheControl,
                    ContentEncoding = blob.Properties.ContentEncoding,
                    ContentLanguage = blob.Properties.ContentLanguage,
                    ContentType = blob.Properties.ContentType,
                    Metadata = blob.Metadata,
                    BlobSize = blob.Properties.Length,
                };

                if (pageRanges != null)
                {
                    info.PageRanges = new SerializablePageRange[pageRanges.Length];
                    for (var i = 0; i < pageRanges.Length; i++)
                    {
                        info.PageRanges[i] = new SerializablePageRange
                        {
                            StartOffset = pageRanges[i].StartOffset,
                            EndOffset = pageRanges[i].EndOffset
                        };
                    }
                }

                using (var stream = info.Serialize())
                {
                    stream.CopyTo(metadataStream);
                }
            }
        }
    }
}