﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using DocaAzure.Backup.Interfaces;

namespace DocaAzure.Backup.Impl
{
    public class LocalFileSystemBackupStorageFactory : IBackupStorageFactory
    {
        public const int NextBlockSignature = 0x123456;
        public const int EofSignature = 0x654321;

        public string BaseFolder { get; private set; }

        public LocalFileSystemBackupStorageFactory(string baseFolder)
        {
            BaseFolder = baseFolder;
        }

        public ITableBackupWriter GetTableBackupWriter(string blobName)
        {
            return new LocalFileSystemTableBackupWriter(Path.Combine(BaseFolder, blobName));
        }

        public ITableRestoreReader GetTableRestoreReader(string blobName)
        {
            // expects that the base folder is already included in the name by GetTableBlobs
            return new LocalFileSystemTableRestoreReader(blobName);
        }

        public IEnumerable<string> GetTableBlobs(string backupContainerName, string tablePrefix)
        {
            var fullPath = Path.Combine(BaseFolder, backupContainerName, tablePrefix);

            return Directory.Exists(fullPath) 
                ? Directory.EnumerateFiles(fullPath) 
                : new string[0];
        }

        public IEnumerable<string> GetTables(string backupContainerName)
        {
            return GetSubfoldersOnRoot(backupContainerName);
        }

        public IEnumerable<string> GetOriginalContainers(string backupContainerName)
        {
            return GetSubfoldersOnRoot(backupContainerName);
        }

        protected IEnumerable<string> GetSubfoldersOnRoot(string backupContainerName)
        {
            var fullPath = Path.Combine(BaseFolder, backupContainerName);

            if (!Directory.Exists(fullPath))
                return new string[0];

            return Directory.EnumerateDirectories(fullPath).Select(dir => dir.Substring(fullPath.Length + 1)).ToList();
        }

        public class LocalFileSystemTableBackupWriter : ITableBackupWriter
        {
            public string FileName { get; private set; }

            public BinaryWriter FileWriter { get; private set; }

            public LocalFileSystemTableBackupWriter(string fileName)
            {
                FileName = fileName;
            }

            public void AddBlock(string blockId, Stream dataStream)
            {
                try
                {
                    if (FileWriter == null)
                    {
                        // ReSharper disable AssignNullToNotNullAttribute
                        Directory.CreateDirectory(Path.GetDirectoryName(FileName));
                        // ReSharper restore AssignNullToNotNullAttribute

                        FileWriter = new BinaryWriter(new FileStream(FileName, FileMode.Create, FileAccess.ReadWrite, FileShare.None));
                    }

                    // write signature that there is next block
                    FileWriter.Write(NextBlockSignature);

                    // write block's length
                    if (dataStream.Length > int.MaxValue)
                        throw new InvalidOperationException(string.Format(Resources.Text.data_length_in_block_too_long, blockId, dataStream.Length));

                    var dataLength = (int)dataStream.Length;
                    FileWriter.Write(dataLength);
                    FileWriter.Write(~dataLength);  // reversed for safety check

                    // write block's data
                    dataStream.CopyTo(FileWriter.BaseStream);
                }
                catch
                {
                    Dispose();
                    throw;
                }
            }

            public void FinalizeBlob()
            {
                if (FileWriter == null)
                    return;

                // write EOF signature
                FileWriter.Write(EofSignature);

                FileWriter.Close();

                FileWriter = null;
            }

            public void Dispose()
            {
                try
                {
                    if (FileWriter != null)
                    {
                        FileWriter.Close();

                        FileWriter = null;

                        File.Delete(FileName);
                    }
                }
                // ReSharper disable EmptyGeneralCatchClause
                catch
                // ReSharper restore EmptyGeneralCatchClause
                {
                }
            }
        }
    
        public class LocalFileSystemTableRestoreReader : ITableRestoreReader
        {
            public string FileName { get; private set; }

            public BinaryReader FileReader { get; private set; }

            public LocalFileSystemTableRestoreReader(string fileName)
            {
                FileName = fileName;
                FileReader = new BinaryReader(new FileStream(FileName, FileMode.Open, FileAccess.Read, FileShare.None));
            }

            public IEnumerable<Stream> GetBlocks()
            {
                while(true)
                {
                    var signature = FileReader.ReadInt32();

                    if (signature == NextBlockSignature)
                    {
                        var dataLength = FileReader.ReadInt32();

                        if (dataLength != (~FileReader.ReadInt32()))
                            throw new InvalidDataException(string.Format(Resources.Text.restore_file_corrut, FileName));

                        yield return new MemoryStream(FileReader.ReadBytes(dataLength));
                    }
                    else
                    {
                        FileReader.Close();
                        FileReader = null;

                        if( signature != EofSignature)
                            throw new InvalidDataException(string.Format(Resources.Text.restore_file_corrut, FileName));

                        break;
                    }
                }
            }

            public void Dispose()
            {
                if(FileReader == null)
                    return;

                FileReader.Close();
                FileReader = null;
            }
        }
    }
}
