﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.IO;
using DocaAzure.Backup.Interfaces;
using DocaAzure.Utils;
using ICSharpCode.SharpZipLib.Zip.Compression;
using ICSharpCode.SharpZipLib.Zip.Compression.Streams;

namespace DocaAzure.Backup.Impl
{
    public class SharpZipLibBuffer : IBuffer
    {
        private const long ComressionBufferSize = 64 * 1024;

        private MemoryStream _buffer;
        private CountingZipStreamWrapper _zipStream;

        public int MaxBlockSize { get; set; }

        public long ApproxLength
        {
            get
            {
                return _buffer.Length + _zipStream.ApproxLengthSinceLastFlush;
            }
        }

        public Stream DataStream { get { return _zipStream; } }

        public IDisposable StartNewBlock()
        {
            ClearZipStream();

            if (_buffer == null)
            {
                _buffer = new MemoryStream((int) (1.1*MaxBlockSize)); // extra 10%
            }
            else
            {
                _buffer.ResetToBegining();
                _buffer.SetLength(0);
            }

            _zipStream =
                new CountingZipStreamWrapper(
                    new DeflaterOutputStream(_buffer, new Deflater(Deflater.BEST_COMPRESSION), (int)ComressionBufferSize)
                        {
                            IsStreamOwner = false
                        },
                    (int)Math.Min(ComressionBufferSize, MaxBlockSize / 2), 
                    x => _buffer.Length);

            return new ZipStreamCleaner(this);
        }

        public Stream GetFinalData()
        {
            ClearZipStream();

            if (_buffer == null)
                return null;

            _buffer.ResetToBegining();

            return _buffer;
        }

        public Stream InflateStream(Stream compressedStream)
        {
            var decomressedStream = new MemoryStream(1024 * 1024);

            using (var gZipStream = new InflaterInputStream(compressedStream) { IsStreamOwner = false })
            {
                gZipStream.CopyTo(decomressedStream);
            }

            return decomressedStream.ResetToBegining();
        }

        public void Dispose()
        {
            ClearZipStream();

            ClearBufferStream();
        }

        private void ClearZipStream()
        {
            if (_zipStream == null)
                return;

            _zipStream.Dispose();
            _zipStream = null;
        }

        private void ClearBufferStream()
        {
            if (_buffer == null)
                return;

            _buffer.Dispose();
            _buffer = null;
        }

        class ZipStreamCleaner : IDisposable
        {
            private SharpZipLibBuffer _zlibBuffer;

            public ZipStreamCleaner(SharpZipLibBuffer zlibBuffer)
            {
                _zlibBuffer = zlibBuffer;
            }

            public void Dispose()
            {
                if (_zlibBuffer == null)
                    return;

                _zlibBuffer.ClearZipStream();
                _zlibBuffer = null;
            }
        }
    }
}
