﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using DocaAzure.Backup.Interfaces;
using DocaAzure.Storage;
using DocaAzure.Utils;
using NLog;

namespace DocaAzure.Backup.Impl
{

    /// <summary>
    /// The class that maintains the global state for the iteration
    /// </summary>
    public class State : IDisposable 
    {
        protected static readonly Logger Log = LogHelper.GetCurrentAssemblyLogger();

        private IEnumerator<BackupEntity> _queryIterator;
        private readonly Func<BackupEntity, bool> _entityFilter;

        public IBuffer Buffer { get; private set; }

        public int MaxBlockSize { get; private set; }

        /// <summary>
        /// This entity is the one we may have retrieved but it does not belong to the batch
        /// So we store it here so that it can be returned on the next iteration
        /// </summary>
        public BackupEntity LookAheadEntity { get; set; }

        /// <summary>
        /// We have completed if look ahead entity is null and iterator is completed too.
        /// </summary>
        public bool HasCompleted
        {
            get
            {
                return _queryIterator == null && LookAheadEntity == null;
            }
        }

        /// <summary>
        /// Get the amount of data we have saved in the entity
        /// </summary>
        public long CurrentBlockSize
        {
            get
            {
                return Buffer.ApproxLength;
            }
        }

        public State(IEnumerable<BackupEntity> query, Func<BackupEntity, bool> entityFilter, int maxBlockSize)
            : this(query, entityFilter, maxBlockSize, DefaultBackupServiceFactory.Default.CreateBuffer())
        {
        }

        public State(IEnumerable<BackupEntity> query, Func<BackupEntity, bool> entityFilter, int maxBlockSize, IBuffer buffer)
        {
            _queryIterator = query.GetEnumerator();
            _entityFilter = entityFilter;
            MaxBlockSize = maxBlockSize;
            Buffer = buffer;
            Buffer.MaxBlockSize = maxBlockSize;
        }

        public void Dispose()
        {
            if(Buffer == null)
                return;

            Buffer.Dispose();
            Buffer = null;
        }

        /// <summary>
        /// Return the next entity - which can be either the 
        /// look ahead entity or a new one from the iterator.
        /// We return null if there are no more entities
        /// </summary>
        /// <returns></returns>
        public BackupEntity GetNextEntity()
        {
            BackupEntity entityToReturn = null;

            if (LookAheadEntity != null)
            {
                entityToReturn = LookAheadEntity;
                LookAheadEntity = null;
            }
            else if (_queryIterator != null)
            {
                if (MoveIteratorToNext())
                {
                    entityToReturn = _queryIterator.Current;
                }
                else
                {
                    _queryIterator = null;
                }
            }

            return entityToReturn;
        }

        private bool MoveIteratorToNext()
        {
            try
            {
                while (_queryIterator.MoveNext())
                {
                    if (_entityFilter(_queryIterator.Current))
                        return true;
                }
            }
            catch (Exception e)
            {
                var code = e.ExtractAtsError();

                if( code != AtsError.ResourceNotFound && code != AtsError.InvalidInput)
                {
                    Log.ErrorException(string.Empty, e);
                    throw;
                }
            }

            return false;
        }
    }
}
