﻿#region License
/*
 * Copyright ©2010-2011 Doca Labs Project
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *  
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using DocaAzure.Backup.Impl;
using DocaAzure.Backup.Interfaces;
using Newtonsoft.Json;

namespace DocaAzure.Backup
{
    [Serializable]
    public class PartitionKeyRange : ITableRange
    {
        #region predefined Guid ranges

        public readonly static string[] GuidRanges16 = new[]
                    {
                        "1", "2", "3", "4", "5", "6", "7", "8", "9", "a", "b", "c", "d", "e", "f"
                    };

        public readonly static string[] GuidRanges32 = new[]
                    {
                        "08", "10", "18", "20", "28", "30", "38", "40", "48", "50", "58", "60", "68", "70", "78", "80", 
                        "88", "90", "98", "a0", "a8", "b0", "b8", "c0", "c8", "d0", "d8", "e0", "e8", "f0", "f8"
                    };

        public readonly static string[] GuidRanges64 = new[]
                    {
                              "04", "08", "0c", 
                        "10", "14", "18", "1c", 
                        "20", "24", "28", "2c", 
                        "30", "34", "38", "3c", 
                        "40", "44", "48", "4c", 
                        "50", "54", "58", "5c", 
                        "60", "64", "68", "6c", 
                        "70", "74", "78", "7c", 
                        "80", "84", "88", "8c", 
                        "90", "94", "98", "9c", 
                        "a0", "a4", "a8", "ac", 
                        "b0", "b4", "b8", "bc", 
                        "c0", "c4", "c8", "cc", 
                        "d0", "d4", "d8", "dc", 
                        "e0", "e4", "e8", "ec",  
                        "f0", "f4", "f8", "fc"
                    };

        #endregion predefined Guid ranges

        public string Min { get; private set; }
        public string Max { get; private set; }
        public DateTime? ModifiedFromInclusive { get; private set; }
        public DateTime? ModifiedToExclusive { get; private set; }

        public PartitionKeyRange(string min, string max)
        {
            Min = min;
            Max = max;
        }

        public PartitionKeyRange(string min, string max, DateTime modifiedFromInclusive)
        {
            Min = min;
            Max = max;
            ModifiedFromInclusive = modifiedFromInclusive;
        }

        [JsonConstructor]
        public PartitionKeyRange(string min, string max, DateTime? modifiedFromInclusive, DateTime? modifiedToExclusive)
        {
            Min = min;
            Max = max;
            ModifiedFromInclusive = modifiedFromInclusive;
            ModifiedToExclusive = modifiedToExclusive;
        }

        public static IList<ITableRange> FromKeyRange(
            string[] keys, DateTime? modifiedFromInclusive = null, DateTime? modifiedToExclusive = null)
        {
            // sort the keys
            Array.Sort(keys, StringComparer.InvariantCulture);

            // split key list {A, M, X} into {[null-A), [A-M), [M-X), [X-null)}
            var ranges = new List<ITableRange>
            {
                new PartitionKeyRange(null, keys.Length > 0 ? keys[0] : null, modifiedFromInclusive, modifiedToExclusive)
            };

            for (var i = 1; i < keys.Length; i++)
                ranges.Add(new PartitionKeyRange(keys[i - 1], keys[i], modifiedFromInclusive, modifiedToExclusive));

            if (keys.Length > 0)
                ranges.Add(new PartitionKeyRange(keys[keys.Length - 1], null, modifiedFromInclusive, modifiedToExclusive));

            return ranges;
        }

        public IQueryable ApplyRange(IQueryable input)
        {
            var query = (IQueryable<BackupEntity>)input;

            if (Min != null)
                query = query.Where(entity => entity.PartitionKey.CompareTo(Min) >= 0);

            if (Max != null)
                query = query.Where(entity => entity.PartitionKey.CompareTo(Max) < 0);

            if (ModifiedFromInclusive != null)
                query = query.Where(entity => entity.Timestamp >= ModifiedFromInclusive);

            if (ModifiedToExclusive != null)
                query = query.Where(entity => entity.Timestamp < ModifiedToExclusive);

            return query;
        }

        public override string ToString()
        {
            return HttpUtility.UrlEncode((Min ?? "null") + ".to." + (Max ?? "null"));
        }
    }
}
